package de.fesenmeyer.dbnormalizer.core;

import java.util.Collection;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;

/**
 * A {@link Set} implementation for handling a set of attribute names. The set is ordered (by the
 * string which represents the attribute name) to allow fast comparison of two instances with the
 * {@link #compareTo(AttributeStringSet)} and {@link #equals(Object)} methods.
 * 
 * @author DF
 * 
 */
@SuppressWarnings("serial")
public class AttributeStringSet extends TreeSet<String> implements Comparable<AttributeStringSet> {

	/**
	 * Class constructor for creating an instance by copying the AttributeStringSet given by the
	 * attrStrSet argument.
	 * 
	 * @param attrStrSet
	 *            the AttributeStringSet to be used to create a new instance
	 */
	public AttributeStringSet(AttributeStringSet attrStrSet) {
		super(attrStrSet);
	}

	/**
	 * Class constructor for creating an instance by copying the collection given by the attrStrs
	 * argument.
	 * 
	 * @param attrStrs
	 *            the collection to be used to create a new instance
	 */
	public AttributeStringSet(Collection<String> attrStrs) {
		super(attrStrs);
	}

	/**
	 * Class constructor for creating an empty instance.
	 */
	public AttributeStringSet() {
	}

	/**
	 * Compares this instance to another instance by comparing all attribute names (represented by
	 * strings) of the two instances. This method can be used for ordering
	 * {@link AttributeStringSet} instances based on the alphabetical order of the attribute names
	 * contained in the instances. For example, <code>{A}</code> is smaller than
	 * <code>{A,B}</code>, and <code>{A,B}</code> is smaller than <code>{A,C}</code>.
	 * 
	 * @param attrStrSet
	 *            the instance to be compared with this instance
	 * @return Returns <code>0</code>, if the attribute names in both instances are the same.
	 *         Returns an integer smaller than or greater than <code>0</code> if the attribute
	 *         names in both instances are not the same.
	 */
	public int compareTo(AttributeStringSet attrStrSet) {
		if (attrStrSet == null) {
			throw new IllegalArgumentException("The object to compare with must not be null.");
		}
		Iterator<String> iter1 = this.iterator();
		Iterator<String> iter2 = attrStrSet.iterator();

		while (iter1.hasNext()) {
			String attr1 = iter1.next();
			if (!iter2.hasNext()) {
				return 1;
			}
			String attr2 = iter2.next();
			int comparison = attr1.compareTo(attr2);
			if (comparison != 0) {
				return comparison;
			}
		}

		if (iter2.hasNext()) {
			return -1;
		}

		return 0;
	}

	/**
	 * Compares this instance with the object specified by the o argument.
	 * 
	 * @param o
	 *            the object to compare this instance to.
	 * @return false, if the o argument is not an instance of {@link AttributeStringSet}, null and
	 *         the {@link #compareTo(AttributeStringSet)} method returns a value different from
	 *         <code>0</code>; true, if the {@link #compareTo(AttributeStringSet)} method returns
	 *         <code>0</code>
	 */
	@Override
	public boolean equals(Object o) {
		if (o == null || !(o instanceof AttributeStringSet)) {
			return false;
		}
		AttributeStringSet attrStrSet = (AttributeStringSet) o;
		return (compareTo(attrStrSet) == 0);
	}

	/**
	 * Returns the hashCode for this instance, which is computed based on the attribute names.
	 * 
	 * @return the computed hashCode
	 */
	@Override
	public int hashCode() {
		int hash = 7;
		for (String attr : this) {
			hash = 31 * hash + attr.hashCode();
		}
		return hash;
	}

	/**
	 * Copies this instance.
	 * 
	 * @return the copy of this instance
	 */
	public AttributeStringSet copy() {
		return new AttributeStringSet(this);
	}
}
