package de.fesenmeyer.dbnormalizer.core;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import net.sourceforge.squirrel_sql.fw.sql.SQLConnection;
import de.fesenmeyer.dbnormalizer.core.configuration.AttributeSetConfig;
import de.fesenmeyer.dbnormalizer.core.configuration.ConfigurationWriter;
import de.fesenmeyer.dbnormalizer.core.configuration.FDConfig;
import de.fesenmeyer.dbnormalizer.core.configuration.SessionConfig;
import de.fesenmeyer.dbnormalizer.core.configuration.TableConfig;
import de.fesenmeyer.dbnormalizer.core.configuration.TableContainerConfig;
import de.fesenmeyer.dbnormalizer.core.dbaccess.DBConnectionManager;
import de.fesenmeyer.dbnormalizer.core.dbaccess.DBException;
import de.fesenmeyer.dbnormalizer.core.dbaccess.DatabaseMetadataUtil;
import de.fesenmeyer.dbnormalizer.core.log.DBNormalizerLogger;

/**
 * DBNormalizerSession is a class holding all information needed and collected
 * by DBNormalizer during the analysis of one or more tables (for example, a DB Schema).
 * It provides a context, which is used by several classes from DBNormalizer
 * to retrieve or store information.
 * @author DF
 *
 */
public class DBNormalizerSession {

	/**
	 * The logger for this class.
	 */
	private static DBNormalizerLogger logger = DBNormalizerLogger
			.getLogger(DBNormalizerSession.class);

	/**
	 * The (unique) name of the session.
	 */
	private String sessionName = Long.toString(System.currentTimeMillis());

	/**
	 * The DB-configuration data for this session.
	 */
	private DBConfig dbConfig = null;

	/**
	 * A SQuirreL-SQL-Connection object which is used for
	 * retrieving database information.
	 */
	private SQLConnection sqlConnection = null;

	/**
	 * A Map containing sets of tables to be normalized. They are separated
	 * by TableContainer, the key of the map.
	 */
	private TreeMap<TableContainer, Set<Table>> tables2normalize = new TreeMap<TableContainer, Set<Table>>();

	/**
	 * Class Constructor for creating a session without DB-Configuration.
	 */
	public DBNormalizerSession() {

	}

	/**
	 * Class Constructor for creating a session with DB-Configuration.
	 * @param dbConfig the DB-Configuration object
	 * @throws DBException If an exception occurs when accessing the DB.
	 */
	public DBNormalizerSession(DBConfig dbConfig) throws DBException {
		assignFrom(dbConfig);
	}

	/**
	 * Adds tables to be normalized to this session, based on configuration data.
	 * @param tableContainerConfigs configuration data
	 */
	public void addTableInformation(Collection<TableContainerConfig> tableContainerConfigs) {
		createTables2normalizeFromConfig(tableContainerConfigs);
	}

	/**
	 * Assign a DB-Configuration to this session.
	 * @param dbConfig the DB-Configuration
	 * @throws DBException If an exception occurs when accessing the DB.
	 */
	private void assignFrom(DBConfig dbConfig) throws DBException {
		this.dbConfig = dbConfig;
		startSession();
	}

	/**
	 * Adds the tables given in the tableContainerConfigs argument (with their FDs)
	 * to the {@link #tables2normalize} map
	 * 
	 * @param tableContainerConfigs the tableContainer configuration data
	 */
	private void createTables2normalizeFromConfig(
			Collection<TableContainerConfig> tableContainerConfigs) {
		boolean loadFromDB = (this.dbConfig != null);

		for (TableContainerConfig tableContainerConfig : tableContainerConfigs) {
			TableContainer tableContainer = tableContainerConfig.getTableContainer();
			for (TableConfig tableConfig : tableContainerConfig.getTableConfigs()) {
				createAndAddTable2Normalize(tableContainer, tableConfig, loadFromDB);
			}
		}
	}

	/**
	 * Helper method for creating one table and its FDs
	 * based on configuration data and adding this information
	 * to the {@link #tables2normalize} Map.
	 * @param tableContainer the tableContainer containing the table
	 * @param tableConfig the configuration data of the table
	 * @param loadFromDB true, if table information (such as attribute names)
	 * should be loaded from a DB; false, otherwise (that means attribute names
	 * must be stored in the configuration)
	 */
	private void createAndAddTable2Normalize(TableContainer tableContainer,
			TableConfig tableConfig, boolean loadFromDB) {
		// build table just with name
		Table emptytable = new Table();
		emptytable.setName(tableConfig.getName());

		switch (tableContainer.getType()) {
		case CATALOG:
			emptytable.setCatalog(tableContainer.getName());
			break;
		case SCHEMA:
			emptytable.setSchema(tableContainer.getName());
		default:
			// do nothing
			break;
		}

		emptytable.setContainer(tableContainer);

		Table table = null;

		if (loadFromDB) {
			table = findTableFromDB(emptytable);
			if (table == null) {
				if (table == null) {
					logger.error("Table with name " + emptytable.getQualifiedName()
							+ " was not found in DB");
					return;
				}
			}
		} else {
			table = emptytable;
			// add attributes to the table
			AttributeStringSet attrs = new AttributeStringSet();
			for (String attrName : tableConfig.getAttributeSetConfig().getAttributes()) {
				attrs.add(attrName);
			}
			table.setAttributes(attrs);
		}

		// copy FDs from SessionConfig to the table
		TreeSet<FD> fds = new TreeSet<FD>();
		for (FDConfig fdConfig : tableConfig.getFdConfigs()) {
			AttributeStringSet lhs = new AttributeStringSet();
			for (String attrName : fdConfig.getLhs().getAttributes()) {
				lhs.add(attrName);
			}

			AttributeStringSet rhs = new AttributeStringSet();
			for (String attrName : fdConfig.getRhs().getAttributes()) {
				rhs.add(attrName);
			}

			FD fd = new FD();
			fd.setLHS(lhs);
			fd.setRHS(rhs);
			fds.add(fd);
		}
		table.setFDs(fds);

		addTable2Normalize(tableContainer, table);
	}

	/**
	 * Helper Method for retrieving table information from the database.
	 * @param table a table object at least containing name, catalog and schema information
	 * @return the table found in the database;<code>null</code>, if the table could not be found
	 */
	private Table findTableFromDB(Table table) {
		Table foundTable = null;
		try {
			foundTable = DatabaseMetadataUtil.getTableInfo(this, table.getCatalog(), table
					.getSchema(), table.getName(), true);
		} catch (SQLException e) {
			logger.error("Unexpected SQL Exception", e);
		}

		return foundTable;
	}

	/**
	 * Starts the session by creating a connection to the DB
	 * specified by the {@link #dbConfig} member. If there is no DB-Configuration
	 * given, the method does nothing.
	 * @throws DBException If an exception occurs when connecting to the DB.
	 */
	public void startSession() throws DBException {
		if (this.dbConfig != null) {
			DBConnectionManager dbMgr = DBConnectionManager.getInstance();
			this.sqlConnection = dbMgr.createAndAddConnection(sessionName, dbConfig);
		}
	}

	/**
	 * Closes the session by closing the connection to the DB. If there is no DB-Configuration
	 * given, the method does nothing.
	 * @throws DBException If an exception occurs when closing the connection.
	 */
	public void closeSession() throws DBException {
		if (this.dbConfig != null) {
			DBConnectionManager dbMgr = DBConnectionManager.getInstance();
			dbMgr.removeConnection(this.sessionName);
			this.sqlConnection = null;
		}
	}

	/**
	 * Gets the SQuirreL-SQL-Connection, if a connection
	 * to a DB exists.
	 * @return the SQuirrel-SQL-Connection, if a connection to a DB exists;
	 * null, otherwise.
	 */
	public Connection getConnection() {
		Connection conn = null;
		if (sqlConnection != null) {
			conn = sqlConnection.getConnection();
		}
		return conn;
	}

	/**
	 * Gets the DB-Configuration used to establish
	 * a connection to a DB.
	 * @return the DB-Configuration, if one is specified;
	 * null, otherwise.
	 */
	public DBConfig getDbConfig() {
		return dbConfig;
	}

	/**
	 * Gets all TableContainers which contain tables which are in
	 * the set of tables to be normalized.
	 * @return a set of TableContainers, which may be empty.
	 */
	public Set<TableContainer> getTableContainers2Normalize() {
		return this.tables2normalize.keySet();
	}

	/**
	 * Gets all tables which are in the set of tables to be normalized
	 * and are contained in the TableContainer specified by argument
	 * tableContainer.
	 * @param tableContainer a TableContainer
	 * @return a set of Tables, which may be empty.
	 */
	public Set<Table> getTables2Normalize(TableContainer tableContainer) {
		return this.tables2normalize.get(tableContainer);
	}

	/**
	 * Adds a TableContainer to the session. This method just adds a tableContainer
	 * to the internal data representation, which makes it possible to add tables
	 * to to be normalized to this container.
	 * @param tableContainer a TableContainer
	 * @return true, if it has been successfully added; false, otherwise (if it has already
	 * been added before this method call).
	 */
	public boolean addTableContainer2Normalize(TableContainer tableContainer) {
		// it is very important if the key already exists, otherwise tables
		// would be removed
		boolean containerNotExists = !this.tables2normalize.containsKey(tableContainer);
		if (containerNotExists) {
			this.tables2normalize.put(tableContainer, new TreeSet<Table>());
		}
		return containerNotExists;
	}

	/**
	 * Adds a table to the set of tables to be normalized. Because the tables
	 * are grouped by TableContainer internally, not only the table argument,
	 * but also the tableContainer argument has to be specified.
	 * @param tableContainer the TableContainer the table is contained in
	 * @param table the table
	 * @return true, if the table has been successfully added; false, otherwise
	 * (if it has already been added before this method call).
	 */
	public boolean addTable2Normalize(TableContainer tableContainer, Table table) {
		// add table to tables2normalize
		boolean containerNotExists = !this.tables2normalize.containsKey(tableContainer);
		if (containerNotExists) {
			this.tables2normalize.put(tableContainer, new TreeSet<Table>());
		}
		boolean tableNotExists = this.tables2normalize.get(tableContainer).add(table);
		return tableNotExists;
	}

	/**
	 * Removes the TableContainer specified by argument tableContainer 
	 * and all tables, which are in the set of tables to be normalized and
	 * are contained in this tableContainer, from the internal data representation.
	 * @param tableContainer a TableContainer
	 */
	public void removeTableContainer2Normalize(TableContainer tableContainer) {
		this.tables2normalize.remove(tableContainer);
	}

	/**
	 * Creates a session configuration from the current state of
	 * the session. This session configuration can then be saved to
	 * a file by means of the {@link ConfigurationWriter} class.
	 * @param configFileName the file where the session configuration
	 * should be saved
	 * @return the created session configuration
	 */
	public SessionConfig createSessionConfig(String configFileName) {
		SessionConfig sessionConfig = new SessionConfig(configFileName);
		sessionConfig.setDbConfig(this.dbConfig);
		Collection<TableContainerConfig> tableContainerConfigs = new ArrayList<TableContainerConfig>();
		for (TableContainer tableContainer : this.tables2normalize.keySet()) {
			TableContainerConfig tableContainerConfig = new TableContainerConfig(tableContainer);
			Collection<TableConfig> tableConfigs = new ArrayList<TableConfig>();
			for (Table table : this.tables2normalize.get(tableContainer)) {
				TableConfig tableConfig = new TableConfig();
				tableConfig.setName(table.getName());

				// if the session is not a DB-Session, the table's attributes
				// can not be retrieved as DB-Metadata and therefore must
				// be specified in the configuration file
				if (this.dbConfig == null) {
					AttributeStringSet attributes = table.getAttributes();
					AttributeSetConfig attributeSetConfig = createAttributeSetConfig(attributes);
					tableConfig.setAttributeSetConfig(attributeSetConfig);
				}
				Collection<FDConfig> fdConfigs = new ArrayList<FDConfig>();
				for (FD fd : table.getFDs()) {
					FDConfig fdConfig = new FDConfig();
					AttributeSetConfig lhs = createAttributeSetConfig(fd.getLHS());
					fdConfig.setLhs(lhs);
					AttributeSetConfig rhs = createAttributeSetConfig(fd.getRHS());
					fdConfig.setRhs(rhs);
					fdConfigs.add(fdConfig);
				}
				tableConfig.setFdConfigs(fdConfigs);

				tableConfigs.add(tableConfig);
			}
			tableContainerConfig.setTableConfigs(tableConfigs);

			tableContainerConfigs.add(tableContainerConfig);
		}
		sessionConfig.setTableContainerConfigs(tableContainerConfigs);

		return sessionConfig;

	}

	private AttributeSetConfig createAttributeSetConfig(AttributeStringSet attributes) {
		AttributeSetConfig attributeSetConfig = new AttributeSetConfig();
		for (String attribute : attributes) {
			attributeSetConfig.getAttributes().add(attribute);
		}
		return attributeSetConfig;
	}

	public SQLConnection getSqlConnection() {
		return sqlConnection;
	}

}
