package de.fesenmeyer.dbnormalizer.core;

/**
 * This class represents a Functional Dependency (FD). A FD
 * consists of a LHS (left hand side) and RHS (right hand side) of
 * attributes. Furthermore it has a state, which specifies, if the FD has
 * been checked, and if it has been checked, if the associated table satisfies
 * the FD. This state can be get and set by means of the {@link #getFdCheckState()}
 * and {@link #setFdCheckState(FDCheckState)} methods.
 * @author DF
 *
 */
public class FD extends AbstractDBItem implements Comparable<FD> {
	/**
	 * The check state of this FD.
	 */
	private FDCheckState fdCheckState = FDCheckState.UNCHECKED;

	/**
	 * The LHS of this FD.
	 */
	private AttributeStringSet LHS = new AttributeStringSet();
	/**
	 * The RHS of this FD.
	 */
	private AttributeStringSet RHS = new AttributeStringSet();

	/**
	 * Gets the LHS.
	 * @return the LHS
	 */
	public AttributeStringSet getLHS() {
		return LHS;
	}

	/**
	 * Sets the LHS.
	 * @param lhs the LHS
	 */
	public void setLHS(AttributeStringSet lhs) {
		LHS = lhs;
	}

	/**
	 * Gets the RHS.
	 * @return the RHS
	 */
	public AttributeStringSet getRHS() {
		return RHS;
	}

	/**
	 * Sets the RHS.
	 * @param rhs the RHS
	 */
	public void setRHS(AttributeStringSet rhs) {
		RHS = rhs;
	}

	/**
	 * Overridden method, mainly for debugging purposes. Creates
	 * a String such as <code>[A,B] -> [C]</code>.
	 * @return the String describing this FD.
	 */
	@Override
	public String toString() {
		StringBuffer strBuf = new StringBuffer();
		if (this.getName() != null) {
			strBuf.append(this.getName()).append(": ");
		}
		strBuf.append(LHS.toString()).append(" -> ").append(RHS.toString());
		return strBuf.toString();
	}

	/**
	 * Compares this FD to another FD, based on the
	 * {@link #toString()} method.
	 * @param fd the FD to which this FD should be compared
	 */
	public int compareTo(FD fd) {
		return this.toString().compareTo(fd.toString());
	}

	/**
	 * Makes a deep copy of this FD.
	 * @return the deep copy of this FD
	 */
	public FD copy() {
		FD fd = new FD();
		fd.setName(this.getName());
		fd.setLHS(this.LHS.copy());
		fd.setRHS(this.RHS.copy());
		fd.setFdCheckState(this.fdCheckState);
		return fd;
	}

	/**
	 * Gets the check state of this FD.
	 * @return the check state of this FD
	 */
	public FDCheckState getFdCheckState() {
		return fdCheckState;
	}

	/**
	 * Sets the check state of this FD
	 * @param fdCheckState the check state of this FD
	 */
	public void setFdCheckState(FDCheckState fdCheckState) {
		this.fdCheckState = fdCheckState;
	}
}
