package de.fesenmeyer.dbnormalizer.core;

import java.sql.DatabaseMetaData;
import java.util.ArrayList;
import java.util.List;

import de.fesenmeyer.dbnormalizer.core.util.MiscUtil;

/**
 * This class represents a Foreign Key relationship (FK) between
 * two tables. The referenced table is called the PK-Table, because it
 * contains the primary key which is referenced. The properties of this class
 * with the prefix <code>pk</code> hence describe the referenced table.
 * The referencing table is also called the FK-Table, because it defines
 * the FK. The properties of this class with the prefix <code>fk</code>
 * hence describe the referencing table.
 * @author DF
 *
 */
public class FK extends AbstractDBItem implements Comparable<FK> {
	/**
	 * The catalog of the PK-Table, may be null.
	 */
	private String pkCatalog = null;
	/**
	 * The schema of the PK-Table, may be null.
	 */
	private String pkSchema = null;
	/**
	 * The name of the PK-Table.
	 */
	private String pkTableName = null;
	/**
	 * The name of the FK-Table.
	 */
	private String fkTableName = null;
	/**
	 * The schema of the FK-Table, may be null.
	 */
	private String fkSchema = null;
	/**
	 * The catalog of the FK-Table, may be null.
	 */
	private String fkCatalog = null;
	/**
	 * The update rule.
	 */
	private int updateRule = DatabaseMetaData.importedKeyNoAction;
	/**
	 * The delete rule.
	 */
	private int deleteRule = DatabaseMetaData.importedKeyNoAction;
	/**
	 * The name of the referenced PK.
	 */
	private String pkName;
	/**
	 * The deferability.
	 */
	private int deferability;

	/**
	 * The referenced columns of the PK-Table.
	 */
	private List<String> pkColumnNames = new ArrayList<String>();
	/**
	 * The referencing columns of the FK-Table.
	 */
	private List<String> fkColumnNames = new ArrayList<String>();



	/**
	 * Gets the catalog of the PK-Table, may be null.
	 * @return the pkCatalog
	 */
	public String getPkCatalog() {
		return pkCatalog;
	}

	/**
	 * Sets the catalog of the PK-Table, may be null.
	 * @param pkCatalog the pkCatalog to set
	 */
	public void setPkCatalog(String pkCatalog) {
		this.pkCatalog = pkCatalog;
	}

	/**
	 * Gets the schema of the PK-Table, may be null.
	 * @return the pkSchema
	 */
	public String getPkSchema() {
		return pkSchema;
	}

	/**
	 * Sets the schema of the PK-Table, may be null.
	 * @param pkSchema the pkSchema to set
	 */
	public void setPkSchema(String pkSchema) {
		this.pkSchema = pkSchema;
	}

	/**
	 * Gets the name of the PK-Table.
	 * @return the pkTableName
	 */
	public String getPkTableName() {
		return pkTableName;
	}

	/**
	 * Sets the name of the PK-Table.
	 * @param pkTableName the pkTableName to set
	 */
	public void setPkTableName(String pkTableName) {
		this.pkTableName = pkTableName;
	}

	/**
	 * Gets the name of the FK-Table.
	 * @return the fkTableName
	 */
	public String getFkTableName() {
		return fkTableName;
	}

	/**
	 * Sets the name of the FK-Table.
	 * @param fkTableName the fkTableName to set
	 */
	public void setFkTableName(String fkTableName) {
		this.fkTableName = fkTableName;
	}

	/**
	 * Gets the schema of the FK-Table, may be null.
	 * @return the fkSchema
	 */
	public String getFkSchema() {
		return fkSchema;
	}

	/**
	 * Sets the schema of the FK-Table, may be null.
	 * @param fkSchema the fkSchema to set
	 */
	public void setFkSchema(String fkSchema) {
		this.fkSchema = fkSchema;
	}

	/**
	 * Gets the catalog of the FK-Table, may be null.
	 * @return the fkCatalog
	 */
	public String getFkCatalog() {
		return fkCatalog;
	}

	/**
	 * Sets the schema of the FK-Table, may be null.
	 * @param fkCatalog the fkCatalog to set
	 */
	public void setFkCatalog(String fkCatalog) {
		this.fkCatalog = fkCatalog;
	}

	/**
	 * Gets the updateRule.
	 * @return the updateRule
	 */
	public int getUpdateRule() {
		return updateRule;
	}

	/**
	 * Sets the updateRule.
	 * @param updateRule the updateRule to set
	 */
	public void setUpdateRule(int updateRule) {
		this.updateRule = updateRule;
	}

	/**
	 * Gets the deleteRule.
	 * @return the deleteRule
	 */
	public int getDeleteRule() {
		return deleteRule;
	}

	/**
	 * Sets the deleteRule.
	 * @param deleteRule the deleteRule to set
	 */
	public void setDeleteRule(int deleteRule) {
		this.deleteRule = deleteRule;
	}

	/**
	 * Gets the name of the referenced PK.
	 * @return the pkName
	 */
	public String getPkName() {
		return pkName;
	}

	/**
	 * Sets the name of the referenced PK.
	 * @param pkName the pkName to set
	 */
	public void setPkName(String pkName) {
		this.pkName = pkName;
	}

	/**
	 * Gets the deferability.
	 * @return the deferability
	 */
	public int getDeferability() {
		return deferability;
	}

	/**
	 * Sets the deferability.
	 * @param deferability the deferability to set
	 */
	public void setDeferability(int deferability) {
		this.deferability = deferability;
	}

	/**
	 * Gets the referenced columns of the PK-Table.
	 * @return the pkColumnNames
	 */
	public List<String> getPkColumnNames() {
		return pkColumnNames;
	}

	/**
	 * Sets the referenced columns of the PK-Table.
	 * @param pkColumnNames the pkColumnNames to set
	 */
	public void setPkColumnNames(List<String> pkColumnNames) {
		this.pkColumnNames = pkColumnNames;
	}

	/**
	 * Gets the referencing columns of the FK-Table.
	 * @return the fkColumnNames
	 */
	public List<String> getFkColumnNames() {
		return fkColumnNames;
	}

	/**
	 * Sets the referencing columns of the FK-Table.
	 * @param fkColumnNames the fkColumnNames to set
	 */
	public void setFkColumnNames(List<String> fkColumnNames) {
		this.fkColumnNames = fkColumnNames;
	}


	/**
	 * Overridden method, which compares this FK to an
	 * arbitrary object.
	 * @param obj the object to which this FK should be compared
	 * @return true, if the obj argument is not null, another FK
	 * and the {@link #toString()} method returns the same value
	 * for both FKs; false, otherwise.
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == null || !(obj instanceof FK)) {
			return false;
		}
		return this.toString().equals(obj.toString());
	}

	/**
	 * Overridden method which compares a hashcode
	 * based on the {@link #toString()} method.
	 * @return the computed hashCode
	 */
	@Override
	public int hashCode() {
		return this.toString().hashCode();
	}

	/**
	 * Overridden method which returns an appropriate
	 * string representation of this FK, for example:
	 * <pre>
	 * FK: Table1(A,B) --> Table2(A,B)
	 * </pre>
	 * @return the string representation of this FK
	 */
	@Override
	public String toString() {
		StringBuffer strBuf = new StringBuffer();
		if (this.getName() == null) {
			strBuf.append("FK");
		} else {
			strBuf.append("FK ").append(this.getName());
		}
		strBuf.append(": ");
		strBuf.append(MiscUtil.getQualifiedTableName(this.getFkTableName(), this.fkCatalog,
				this.fkSchema));
		strBuf.append("(");
		strBuf.append(this.getFkColumnNames().toString());
		strBuf.append(") --> ");
		strBuf.append(MiscUtil.getQualifiedTableName(this.getPkTableName(), this.pkCatalog,
				this.pkSchema));
		strBuf.append("(");
		strBuf.append(this.getPkColumnNames().toString());
		strBuf.append(")");
		return strBuf.toString();
	}

	/**
	 * Default implementation for comparing FKs. Uses the {@link #toString()} method to compare the FKs.
	 * 
	 * @param fk
	 *            the FK to be compared with this FK.
	 * @return Returns
	 * <ul>
	 * 	<li>
	 * 		an int value smaller than <code>0</code>,
	 * 		if the string representation of this FK is smaller than
	 * 		the other FK's one.
	 *  </li>
	 * 	<li>
	 * 		<code>0</code>,
	 * 		if the FKs' string representations are equal.
	 * 	</li>
	 * 	<li>an int value greater than <code>0</code>,
	 * 		if the string representation of this FK is greater than
	 * 		the other FK's one.
	 * 	</li>
	 *<ul>
	 */
	public int compareTo(FK fk) {
		return this.toString().compareTo(fk.toString());
	}

}
