package de.fesenmeyer.dbnormalizer.core;

import java.sql.DatabaseMetaData;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import de.fesenmeyer.dbnormalizer.core.algorithms.NFDeterminationResult;
import de.fesenmeyer.dbnormalizer.core.algorithms.NormalizationProposal;

/**
 * Class representing a table. If the table is a DB-Table, i.e. if it exists in a DB or should be
 * created in a DB (if it is part of a normalization proposal of a DB-Table),
 * the flag <code>dbTable</code> must be true, otherwise it must be false. This flag
 * can be get and set by means of the {@link #isDbTable()} and {@link #setDbTable(boolean)}
 * methods.
 * @author DF
 *
 */
public class Table extends AbstractDBItem implements Comparable<Table> {

	/**
	 * The catalog this table is contained in, may be null.
	 */
	private String catalog = null;
	/**
	 * The schema this table is contained in, may be null.
	 */
	private String schema = null;

	/**
	 * The {@link TableContainer} this table is contained in.
	 */
	private TableContainer container = null;

	/**
	 * The set of attributes belonging to this table.
	 */
	private AttributeStringSet attributes = new AttributeStringSet();

	/**
	 * A map containing tableColumn information for each attribute. Is
	 * empty, if the table is not a DB-Table.
	 */
	private Map<String, TableColumn> columnInfos = new TreeMap<String, TableColumn>();

	/**
	 * The PK defined in the DB. Is null, if the table is not a DB-Table.
	 */
	private PK dbPK = null;

	/**
	 * The PK selected by DBNormalizer when calculating candidate keys
	 * or by the user.
	 */
	private PK pk = null;
	
	/**
	 * The exported keys for this table.
	 * @see DatabaseMetaData#getExportedKeys(String, String, String)
	 */
	private List<FK> exportedKeys = new ArrayList<FK>();
	/**
	 * The imported keys for this table.
	 * @see DatabaseMetaData#getImportedKeys(String, String, String)
	 */
	private List<FK> importedKeys = new ArrayList<FK>();
	/**
	 * The set of functional dependencies (FDs) for this table, just as
	 * it has been specified by the user or imported from a file. May
	 * contain redundant FDs.
	 * <p>
	 * <b>NOTE: Everytime you change these FDs, make sure you recompute the
	 * minimal cover and update {@link #FDMinCover}, before determining
	 * candidate keys, normal form or creating a normalization proposal.</b>
	 * </p>
	 */
	private Set<FD> FDs = new LinkedHashSet<FD>();
	/**
	 * A minimal cover of the functional dependencies (FDs) contained in
	 * {@link #FDs}.
	 */
	private Set<FD> FDMinCover = new LinkedHashSet<FD>();

	/**
	 * The candidate keys.
	 */
	private Set<CandKey> candKeys = new LinkedHashSet<CandKey>();

	/**
	 * The result of the normal form determination for this table.
	 * May be null, if the normal form has not yet been determined.
	 */
	private NFDeterminationResult nfDeterminationResult = null;

	/**
	 * The normalization proposal for this table. May be null, if
	 * the table should not be normalized or if a normalization proposal
	 * has not yet been created.
	 */
	private NormalizationProposal normProposal = null;

	/**
	 * A flag which specifies if the table is a DB-Table or not.
	 * @see Table
	 */
	private boolean dbTable = false;

	/**
	 * Gets the minimal cover of FDs for this table.
	 * @return the minimal cover of FDs
	 * @see #FDMinCover
	 */
	public Set<FD> getFDMinCover() {
		return FDMinCover;
	}

	/**
	 * Sets the minimal cover of FDs for this table.
	 * @param minCover the minimal cover of FDs
	 * @see #FDMinCover
	 */
	public void setFDMinCover(Set<FD> minCover) {
		FDMinCover = minCover;
	}

	/**
	 * Gets the set of attributes belonging to this table.
	 * @return the attributes
	 */
	public AttributeStringSet getAttributes() {
		return attributes;
	}

	/**
	 * Sets the set of attributes belonging to this table.
	 * @param attributes the attributes
	 */
	public void setAttributes(AttributeStringSet attributes) {
		this.attributes = attributes;
	}

	/**
	 * Gets the FDs for this table.
	 * @return the FDs
	 * @see #FDs
	 */
	public Set<FD> getFDs() {
		return FDs;
	}

	/**
	 * Sets the FDs for this table.
	 * @param fds the FDs
	 * @see #FDs
	 */
	public void setFDs(Set<FD> fds) {
		FDs = fds;
	}

	/**
	 * Gets the exported keys.
	 * @return the exported keys
	 * @see DatabaseMetaData#getExportedKeys(String, String, String)
	 */
	public List<FK> getExportedKeys() {
		return exportedKeys;
	}

	/**
	 * Sets the exported keys.
	 * @param exportedKeys the exported keys
	 * @see DatabaseMetaData#getExportedKeys(String, String, String)
	 */
	public void setExportedKeys(List<FK> exportedKeys) {
		this.exportedKeys = exportedKeys;
	}

	/**
	 * Gets the PK for this table.
	 * @return the PK
	 * @see #pk
	 */
	public PK getPk() {
		return pk;
	}

	/**
	 * Sets the PK for this table.
	 * @param pk the PK
	 * @see #pk
	 */
	public void setPk(PK pk) {
		this.pk = pk;
	}

	/**
	 * Gets the imported keys.
	 * @return the imported keys
	 * @see DatabaseMetaData#getImportedKeys(String, String, String)
	 */
	public List<FK> getImportedKeys() {
		return importedKeys;
	}

	/**
	 * Sets the imported keys.
	 * @param importedKeys the imported keys
	 * @see DatabaseMetaData#getImportedKeys(String, String, String)
	 */
	public void setImportedKeys(List<FK> importedKeys) {
		this.importedKeys = importedKeys;
	}

	/**
	 * Overridden Method, tests if this table is equal
	 * to the o argument based on the qualified name.
	 * 
	 * @param o the Object to test if it is equal to this table
	 * @return true, if o is not null, is a table and the {@link #getQualifiedName()}
	 * method returns equal values for both this table and the table
	 * specified by the o argument;
	 * false, otherwise
	 */
	@Override
	public boolean equals(Object o) {
		if (o == null || !(o instanceof Table)) {
			return false;
		}
		if (this.getQualifiedName().equals(((Table) o).getQualifiedName())) {
			return true;
		}
		return false;
	}

	/**
	 * Gets the qualified name for this table. This is the table's (simple) name
	 * qualified by the catalog or schema it is contained in. If it
	 * is neither contained in a catalog nor schema, the qualified name
	 * equals to the table's (simple) name. 
	 * @return the table's qualified name
	 */
	public String getQualifiedName() {
		// TODO: not every DBS uses "." as separator
		String separator = ".";
		StringBuffer strBuf = new StringBuffer();
		if (catalog != null) {
			strBuf.append(catalog).append(separator);
		}
		if (schema != null) {
			strBuf.append(schema).append(separator);
		}
		strBuf.append(getName());
		return strBuf.toString();
	}

	/**
	 * Gets the name of the catalog this table is contained in.
	 * @return the name of the table's catalog, if it is contained in one;
	 * <code>null</code>, otherwise
	 */
	public String getCatalog() {
		return catalog;
	}

	/**
	 * <p>
	 * Sets the name of the catalog this table is contained in; must be
	 * <code>null</code>, if the table is not contained in a catalog.
	 * </p>
	 * <p>
	 * <b>NOTE: If this method is called, the {@link #setContainer(TableContainer)}
	 * method should also be called in order to avoid inconsistencies.</b>
	 * </p>
	 * @param catalog the catalog's name or <code>null</code>
	 */
	public void setCatalog(String catalog) {
		this.catalog = catalog;
	}

	/**
	 * Gets the name of the schema this table is contained in. 
	 * @return the name of the table's schema, if it is contained in one;
	 * <code>null</code>, otherwise
	 */
	public String getSchema() {
		return schema;
	}

	/**
	 * <p>
	 * Sets the name of the schema this table is contained in; must be
	 * <code>null</code>, if the table is not contained in a schema.
	 * </p>
	 * <p>
	 *  <b>NOTE: If this method is called, the {@link #setContainer(TableContainer)}
	 * method should also be called in order to avoid inconsistencies.</b>
	 * </p>
	 * @param schema the schema's name or <code>null</code>
	 */
	public void setSchema(String schema) {
		this.schema = schema;
	}

	/**
	 * Gets the TableContainer this table is contained in.
	 * @return the TableContainer this table is contained in, if
	 * it is contained in one;
	 * <code>null, otherwise</code>
	 * @see TableContainer
	 */
	public TableContainer getContainer() {
		return container;
	}

	/**
	 * <p>
	 * Sets the TableContainer this table is contained in; must be
	 * <code>null</code>, if the table is not contained in one.
	 * </p>
	 * <p>
	 * <b>NOTE: If this method is called, the {@link #setCatalog(String)} and/or
	 * {@link #setSchema(String)} method(s) should also be called in order to
	 * avoid inconsistencies.
	 * </p>
	 * @param container the table's container or <code>null</code>
	 */
	public void setContainer(TableContainer container) {
		this.container = container;
	}

	/**
	 * Gets this table's candidate keys (if they have already
	 * been set).
	 * @return this table's candidate keys, if they have already been set;
	 * an empty set, otherwise
	 */
	public Set<CandKey> getCandKeys() {
		return candKeys;
	}

	/**
	 * Sets this table's candidate keys. <b>If necessary, You should also reset
	 * this table's primary key by means of the
	 * {@link #setPk(PK)} method.</b>
	 * @param candKeys this table's candidate keys
	 */
	public void setCandKeys(Set<CandKey> candKeys) {
		this.candKeys = candKeys;
	}

	/**
	 * Gets the result of the normal form determination for this table.
	 * @return the result of the normal form determination for this table;
	 * <code>null</code>, if it has not yet been set for this table.
	 */
	public NFDeterminationResult getNfDeterminationResult() {
		return nfDeterminationResult;
	}
	
	/**
	 * Sets the result of the normal form determination for this table.
	 * @param nfDeterminationResult the result of the normal form determination
	 */
	public void setNfDeterminationResult(NFDeterminationResult nfDeterminationResult) {
		this.nfDeterminationResult = nfDeterminationResult;
	}

	/**
	 * States whether this table is a DB-Table or not.
	 * @return <code>true</code>, if it is a DB-Table; <code>false</code>, otherwise
	 * @see Table
	 */
	public boolean isDbTable() {
		return dbTable;
	}

	/**
	 * Sets whether this table is a DB-Table or not.
	 * @param dbTable must be <code>true</code>, if it is a DB-Table; <code>false</code>, otherwise
	 * @see Table
	 */
	public void setDbTable(boolean dbTable) {
		this.dbTable = dbTable;
	}

	/**
	 * Gets the primary key (PK) defined in the DB.
	 * @return the PK defined in the DB;<code>null</code>, if none has been defined in the DB
	 */
	public PK getDbPK() {
		return dbPK;
	}

	/**
	 * Sets the primary key (PK) defined in the DB.
	 * @param dbPK the PK defined in the DB; must be <code>null</code>, if
	 * none has been defined in the DB
	 */
	public void setDbPK(PK dbPK) {
		this.dbPK = dbPK;
	}

	/**
	 * Gets TableColumn information for the column (attribute) specified
	 * by the colName argument.
	 * @param colName the column's name
	 * @return a TableColumn object, if this table is a DB-Table and
	 * the column specified by the colName argument exists;
	 * <code>null</code>, otherwise.
	 */
	public TableColumn getColumnInfo(String colName) {
		return this.columnInfos.get(colName);
	}

	/**
	 * Gets TableColumn information for each of the table's columns (attributes).
	 * @return a Collection of TableColumn objects, if this table is a
	 * DB-Table; an empty set, otherwise
	 */
	public Collection<TableColumn> getColumnInfos() {
		return this.columnInfos.values();
	}

	/**
	 * Adds the TableColumn information specified by argument col
	 * for the column (attribute) specified by argument colName.
	 * <b>NOTE: The colName argument must be an element of the
	 * set returned by the {@link #getAttributes()} method.</b>
	 * @param colName the column (attribute) name
	 * @param col the TableColumn object
	 */
	public void addColumnInfo(String colName, TableColumn col) {
		this.columnInfos.put(colName, col);
	}

	/**
	 * Gets the currently set normalization proposal for this table.
	 * @return the (currently set) normalization proposal;
	 * <code>null</code>, if a normalization proposal has not yet been set
	 */
	public NormalizationProposal getNormProposal() {
		return normProposal;
	}

	/**
	 * Sets the normalization proposal for this table. Overrides
	 * a previously set normalization proposal.
	 * @param normProposal a normalization proposal
	 */
	public void setNormProposal(NormalizationProposal normProposal) {
		this.normProposal = normProposal;
	}

	/**
	 * Compares this table to another table based on 
	 * the qualified names of the tables. Can be used to order the tables alphabetically.
	 * @param table another table
	 * @return Returns
	 * <ul>
	 * 	<li>an int value smaller than <code>0</code>,
	 * 	if this table's qualified name is smaller than the other
	 * 	table's qualified name in terms of the alphabetical order.
	 * 	</li>
	 * 	<li><code>0</code>,
	 * 	if this table's qualified name exactly equals the other
	 * 	table's qualified name.
	 * 	</li>
	 * 	<li>an int value greater than <code>0</code>,
	 * 	if this table's qualified name is greater than the other
	 * 	table's qualified name in terms of the alphabetical order.
	 * 	</li>
	 * </ul>
	 */
	public int compareTo(Table table) {
		if (table == null) {
			throw new IllegalArgumentException("Table used for comparison " + "must not be null.");
		}
		return this.getQualifiedName().compareTo(table.getQualifiedName());
	}
}
