package de.fesenmeyer.dbnormalizer.core;

import java.sql.DatabaseMetaData;

/**
 * Represents a column of a table in a DB. Contains
 * most of the common properties which can be retrieved
 * by means of the JDBC-Interface {@link DatabaseMetaData}.
 * @author DF
 * @see DatabaseMetaData
 */
public class TableColumn extends AbstractDBItem implements Comparable<TableColumn> {
	/**
	 * The typeName (database specific), e.g. "VARCHAR2" (oracle-specific).
	 */
	private String typeName = null;
	/**
	 * A Flag specifying whether the column can be null or not,
	 * defaults to <code>true</code>.
	 */
	private boolean nullable = true;
	/**
	 * The data type (database-independent int representation).
	 */
	private int dataType;
	/**
	 * The decimal digits, e.g. for the column
	 * <code>someNumber NUMBER(9,2)</code> the
	 * decimal digits value is <code>2</code>.
	 */
	private int decimalDigits;
	/**
	 * The column size, e.g.
	 * <ul>
	 * <li>For the column <code>someNumber NUMBER(9,2)</code>
	 * the column size is <code>9</code>.
	 * </li>
	 * <li>For the column <code>someText VARCHAR(30)</code>
	 * the column size is <code>30</code>
	 * @see
	 * #decimalDigits
	 */
	private int columnSize;
	/**
	 * The default value.
	 */
	private String defaultValue = null;

	/**
	 * Gets the decimal digits.
	 * @return the decimal digits
	 */
	public int getDecimalDigits() {
		return decimalDigits;
	}

	/**
	 * Sets the decimal digits
	 * @param decimalDigits the decimal digits
	 */
	public void setDecimalDigits(int decimalDigits) {
		this.decimalDigits = decimalDigits;
	}

	/**
	 * Gets the column size.
	 * @return the column size
	 */
	public int getColumnSize() {
		return columnSize;
	}

	/**
	 * Sets the column size.
	 * @param columnSize the column size
	 */
	public void setColumnSize(int columnSize) {
		this.columnSize = columnSize;
	}

	/**
	 * Gets the default value.
	 * @return the default value
	 */
	public String getDefaultValue() {
		return defaultValue;
	}

	/**
	 * Sets the default value.
	 * @param defaultValue the default value
	 */
	public void setDefaultValue(String defaultValue) {
		this.defaultValue = defaultValue;
	}

	/**
	 * Gets the typeName.
	 * @return the typeName
	 */
	public String getTypeName() {
		return typeName;
	}

	/**
	 * Sets the typeName.
	 * @param type the typeName
	 */
	public void setTypeName(String type) {
		this.typeName = type;
	}

	/**
	 * Gets whether the column can be null or not.
	 * @return <code>true</code>, if the table can be null;
	 * <code>false</code>, otherwise
	 */
	public boolean isNullable() {
		return nullable;
	}

	/**
	 * Sets whether the column can be null or not.
	 * @param nullable <code>true</code>, if the table can be null;
	 * <code>false</code>, otherwise
	 */
	public void setNullable(boolean nullable) {
		this.nullable = nullable;
	}

	/**
	 * Sets the data type.
	 * @param dataType the data type
	 */
	public void setDataType(int dataType) {
		this.dataType = dataType;
	}

	/**Gets the data type.
	 * @return the data type
	 */
	public int getDataType() {
		return dataType;
	}

	/**
	 * Compares this column to another column based on
	 * their names. Can be used to sort the columns
	 * alphabetically.
	 * @param col another TableColumn
	 */
	public int compareTo(TableColumn col) {
		if (this.getName() == null) {
			throw new IllegalStateException("The name of the column to be compared is null");
		} else if (col == null || col.getName() == null) {
			throw new IllegalArgumentException("The column used for comparison"
					+ "or its name is null");
		}
		return this.getName().compareTo(col.getName());
	}

	/**
	 * Creates a copy of this TableColumn.
	 * @return the copy of this TableColumn
	 */
	public TableColumn copy() {
		TableColumn newCol = new TableColumn();
		newCol.setName(this.getName());
		newCol.setTypeName(this.typeName);
		newCol.setNullable(this.nullable);
		newCol.setDataType(this.dataType);
		newCol.setDecimalDigits(this.decimalDigits);
		newCol.setColumnSize(this.columnSize);
		newCol.setDefaultValue(this.defaultValue);
		return newCol;
	}

	/**
	 * Overridden Method, tests if this column is equal
	 * to the o argument based on the name.
	 * 
	 * @param o the Object to test if it is equal to this table
	 * @return true, if o is not null, is a TableColumn and the {@link #getName()}
	 * method returns equal values for both this TableColumn and the TableColumn
	 * specified by the o argument;
	 * false, otherwise
	 */
	@Override
	public boolean equals(Object o) {
		if (this == null || o == null || this.getName() == null || !(o instanceof TableColumn)) {
			return false;
		}
		return this.getName().equals(((TableColumn) o).getName());
	}

	/**
	 * Overridden method, returns
	 * a hash code based on the column's
	 * name.
	 * @return the hash code
	 */
	@Override
	public int hashCode() {
		if (getName() == null) {
			return super.hashCode();
		}
		return getName().hashCode();
	}
}
