package de.fesenmeyer.dbnormalizer.core;

/**
 * A TableContainer is used to group tables based on the schema or catalog they
 * are contained in. This class is mainly used by the {@link DBNormalizerSession}
 * and {@link Table} classes. The TableContainer {@link #NO_CONTAINER} can be used
 * as group for tables which are neither contained in a schema nor catalog.
 * @author DF
 * @see TableContainerType
 */
public class TableContainer extends AbstractDBItem implements Comparable<TableContainer> {
	/**
	 * The type.
	 */
	private TableContainerType type = null;

	/**
	 * The name of the container {@link #NO_CONTAINER}.
	 */
	private static final String NO_CONTAINER_NAME = "NO CONTAINER";

	/**
	 * The TableContainer to be used for tables neither contained in a catalog
	 * nor schema.
	 */
	public static final TableContainer NO_CONTAINER = new TableContainer(NO_CONTAINER_NAME,
			TableContainerType.NONE);

	/**
	 * Class Constructor specifying name and type.
	 * <p>
	 * <b>NOTE</b>: For specifying a TableContainer
	 * of type {@link TableContainerType#NONE} the constant {@link #NO_CONTAINER} must
	 * be used instead of this constructor, because a name does not make sense
	 * for this type. Otherwise an {@link IllegalArgumentException} will be thrown.
	 * </p>
	 * @param name the name
	 * @param type the type
	 * @throws IllegalArgumentException if the name argument or type
	 * argument are null, or if type is {@link TableContainerType#NONE}
	 */
	public TableContainer(String name, TableContainerType type) {
		if (name == null || type == null) {
			throw new IllegalArgumentException(
					"Both the name and type argument must not be null");
		} else if (type.equals(TableContainerType.NONE)
				&& !(NO_CONTAINER_NAME.equals(name))) {
			throw new IllegalArgumentException("You must not " +
					"specify a name for TableContainer of type NONE. " +
					"Please use the constant NO_CONTAINER instead.");
		}
		setName(name);
		this.type = type;
	}

	/**
	 * Gets the type.
	 * @return the type
	 */
	public TableContainerType getType() {
		return type;
	}

	/**
	 * Overridden method, tests if the TableContainer
	 * specified by argument o is equal to this TableContainer.
	 * @param o an Object, which should be another TableContainer
	 * @return <code>true</code>, if the o argument is not <code>null</code>,
	 * a TableContainer and its type and name is equal to this
	 * TableContainer; <code>false</code>, otherwise
	 */
	@Override
	public boolean equals(Object o) {
		if (o == null || !(o instanceof TableContainer)) {
			return false;
		}
		TableContainer container = (TableContainer) o;
		if (this.getName().equals(container.getName())
				&& this.getType().equals(container.getType())) {
			return true;
		}
		return false;
	}


	/**
	 * Overridden method, returns an appropriate
	 * string representation of this TableContainer.
	 * @return the string representation of this TableContainer.
	 */
	@Override
	public String toString() {
		StringBuffer strBuf = new StringBuffer();
		String name = getName();
		if (name != null && !NO_CONTAINER.equals(name)) {
			strBuf.append(name).append(" ");
		}
		if (type != null) {
			strBuf.append("(").append(type.toString()).append(")");
		}
		return strBuf.toString();
	}

	/**
	 * Compares this TableContainer to another by comparing
	 * type and name.
	 * @param tableContainer another TableContainer.
	 * @return Returns
	 * <ul>
	 * 	<li>an int value smaller than <code>0</code>,
	 * 	if
	 * 		<ul>
	 * 			<li> this one's type is smaller than the
	 * other one's type or
	 * 			</li>
	 * 			<li>if the types are equal and this one's name
	 * is smaller than the other one's name.
	 * 			</li>
	 * 		</ul>
	 * 	<li><code>0</code>,
	 * 	if the tables' type and name are both equal.
	 * 	</li>
	 * 	<li>an int value greater than <code>0</code>,
	 * if
	 * 		<ul>
	 * 			<li> this one's type is greater than the
	 * other one's type or
	 * 			</li>
	 * 			<li>if the types are equal and this one's name
	 * is greater than the other one's name.
	 * 			</li>
	 * 		</ul>
	 * 	</li>
	 * </ul>
	 */
	public int compareTo(TableContainer tableContainer) {
		if (tableContainer == null) {
			throw new IllegalArgumentException("TableContainer used for "
					+ "comparison must not be null");
		}
		if (this.getType() == null || tableContainer.getType() == null) {
			throw new IllegalStateException("TableContainer must have a type " + "to be compared");
		}
		int comp = this.getType().compareTo(tableContainer.getType());
		if (comp == 0) {
			comp = this.getName().compareTo(tableContainer.getName());
		}
		return comp;
	}

}
