package de.fesenmeyer.dbnormalizer.core.algorithms;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

import de.fesenmeyer.dbnormalizer.core.AttributeStringSet;
import de.fesenmeyer.dbnormalizer.core.CandKey;
import de.fesenmeyer.dbnormalizer.core.FD;
import de.fesenmeyer.dbnormalizer.core.PK;
import de.fesenmeyer.dbnormalizer.core.Table;
import de.fesenmeyer.dbnormalizer.core.log.DBNormalizerLogger;
import de.fesenmeyer.dbnormalizer.core.util.SetOperationsUtil;

/**
 * Class containing several algorithms performing operations on FDs (Functional Dependencies).
 * 
 * @author DF
 *
 */
public class FDAlgorithms {

	private static DBNormalizerLogger logger = DBNormalizerLogger.getLogger(FDAlgorithms.class);

	/**
	 * The Membership algorithm checks if the FD specified by argument fd is member of the set of FDs specified by argument fds,
	 * that means if it can be derived from this set. The membership algorithm is very similar to the Attribute-Closure algorithm.
	 * Hence this implementation is based on the Attribute-Closure algorithm {@link #calculateAttributeClosure(Collection, AttributeStringSet)}.
	 * Descriptions of the membership algorithm are given in [The theory of relational databases, Maier, 1983] and [Database analysis and design,
	 * Hawryszkiewycz, 1991].
	 * 
	 * @param fd a FD
	 * @param fds a set of FDs
	 * @return <code>true</code>, if the FD specified by argument fd is member of the set of FDs specified by argumemt fds;
	 * <code>false</code>, otherwise 
	 */
	public static boolean isMember(FD fd, Set<FD> fds) {
		AttributeStringSet lhsAttributes = new AttributeStringSet(fd.getLHS());

		AttributeStringSet attributeClosure = FDAlgorithms.calculateAttributeClosure(fds,
				lhsAttributes);

		AttributeStringSet rhsAttributes = fd.getRHS();
		return attributeClosure.containsAll(rhsAttributes);
	}

	/**
	 * This algorithm checks if an FD is redundant by using the Membership algorithm {@link #isMember(FD, Set)}. The FD specified by
	 * argument fd should be contained in the set of FDs specified by argument fds. If this FDs is still a member of the given set of FDs,
	 * if it is removed from this set, then this FD is redundant regarding the given set.
	 * @param fd a FD
	 * @param fds a set of FDs
	 * @return <code>true</code>, if the FD specified by argument fd is redundant regarding the set of FDs specified by argumemt fds;
	 * <code>false</code>, otherwise
	 */
	public static boolean isRedundant(FD fd, Set<FD> fds) {
		Set<FD> fdsWithoutFdToCheck = new TreeSet<FD>(fds);
		fdsWithoutFdToCheck.remove(fd);
		boolean isMember = isMember(fd, fdsWithoutFdToCheck);
		return isMember;
	}

	/**
	 * <p>
	 * Calculates the Attribute-Closure for the set of attributes given by argument attributes
	 * regarding the set of FDs given by argument fds. The Attribute-Closure contains all attributes which are functionally dependent on the
	 * the given set of attributes regarding the given set of FDs. The Attribute-Closure contains at least the attributes given by
	 * the attributes argument, because each attribute functionally determines itself due to the reflexivity inference rule.
	 * </p>
	 * <p>
	 * This algorithm has been implemented based on the remarks in
	 * [Fundamentals of database systems, 5th Ed., Elmasri/Navathe, 2007], where the algorithm is called "Closure of F under X".
	 * </p>
	 * 
	 * @param fds a collection (may be a set) of fds
	 * @param attributes a set of attributes
	 * @return the set of attributes which form the Attribute-Closure. Is empty, if and only if the
	 * specified set of attributes is empty. 
	 */
	public static AttributeStringSet calculateAttributeClosure(Collection<FD> fds,
			AttributeStringSet attributes) {
		//initialize attributeClosure with the given attributes
		AttributeStringSet attributeClosure = new AttributeStringSet(attributes);
		
		boolean changes = true;
		while (changes) {
			changes = false;
			for (FD fd : fds) {
				AttributeStringSet lhs = fd.getLHS();
				AttributeStringSet rhs = fd.getRHS();
				if (lhs.size() > 0) { // if lhs is empty, do nothing!
					if (attributeClosure.containsAll(lhs)) {
						if (!attributeClosure.containsAll(rhs)) {
							attributeClosure.addAll(rhs);
							changes = true;
						}
					}
				}
			}
		}
		
		return attributeClosure;
	}

	/**
	 * <p>
	 * Calculates the minimal cover of the set of FDs specified by
	 * the fds argument. The minimal cover is a set of FDs without redundant FDs and contains
	 * no extraneous attributes on both sides and only one attribute on the RHS of each FD.
	 * </p>
	 * <p>
	 * This algorithm is based on [The theory of relational databases, Maier, 1983].
	 * </p>
	 * 
	 * @param fds a set of FDs
	 * @return the minimal cover for the given set of FDs
	 */
	public static SortedSet<FD> calculateMinCover(Set<FD> fds) {
		//calculate a reduced (non-redundant) cover
		ArrayList<FD> fdsWork = calculateReducedCover(fds);

		//split FDs by using decomposition rule, e.g. {a -> bc} => {a -> b, a -> c}
		return splitFDs(fdsWork);
	}

	/**
	 * <p>
	 * Merges FDs using the union inference rule, e.g.:
	 * <pre>{a -> b, a -> c} => {a -> bc}</pre>
	 * One can also say that the FDs are grouped by equal LHS.
	 * </p>
	 * <p>
	 * If the set of of FDs specified by argument fds is
	 * a minimal cover, the returned set of FDs will be
	 * a grouped minimal cover.
	 * </p>
	 * @param fds a set of FDs
	 * @return a set of FDs as described above
	 */
	public static SortedSet<FD> mergeFDs(Collection<FD> fds) {
		//a temporary map which holds LHS attributes as key and RHS attributes as values
		TreeMap<AttributeStringSet, AttributeStringSet> temp = new TreeMap<AttributeStringSet, AttributeStringSet>();

		//fill the temporary map
		for (FD fd : fds) {
			AttributeStringSet lhs = fd.getLHS();
			//look if an entry for the LHS of the current
			//FD already exists
			AttributeStringSet newRhs = temp.get(lhs);
			//if the LHS is not yet in the map, create an entry
			if (newRhs == null) {
				newRhs = new AttributeStringSet();
				temp.put(lhs, newRhs);
			}
			//add the RHS attribute to the appropriate map entry
			newRhs.addAll(fd.getRHS());
		}

		//create a set from the temporary map
		TreeSet<FD> fdsResult = new TreeSet<FD>();
		for (AttributeStringSet lhs : temp.keySet()) {
			AttributeStringSet rhs = temp.get(lhs);
			FD fd = new FD();
			fd.setLHS(lhs);
			fd.setRHS(rhs);
			fdsResult.add(fd);
		}
		
		return fdsResult;
	}

	/**
	 * Calculates a reduced cover for the set of FDs specified by
	 * the fds argument. This algorithm first removes all extraneous
	 * attributes on the LHS of each FD (Left reduction). As a second step,
	 * it removes all extraneous attributes on the RHS of each FD, 
	 * removing the whole FD, if the RHS gets empty during this step.
	 * <p>
	 * The reduced cover is the return value of this method. The FDs contained
	 * in the fds argument are not changed by the algorithm.
	 * </p>
	 * <p>
	 * This algorithm is based on [The theory of relational databases, Maier, 1983].
	 * </p>
	 * 
	 * @param fds a set of FDs
	 * @return a list representing the reduced cover for the given set of FDs
	 */
	private static ArrayList<FD> calculateReducedCover(Set<FD> fds) {
		//We use a list for the FDs for easier implementation (we can get
		//FDs by their index) and better performance (java.util.Set operations
		//are slower because they must assure uniqueness)
		ArrayList<FD> fdsWork = new ArrayList<FD>(fds.size());

		for (FD fd : fds) {
			fdsWork.add(fd.copy());
		}

		// 1. Left reduction (removes all extraneous attributes from LHS)
		for (int i = 0; i < fdsWork.size(); i++) {
			FD fd = fdsWork.get(i);
			AttributeStringSet lhs = fd.getLHS();
			AttributeStringSet rhs = fd.getRHS();
			String[] lhsAttrs = new String[lhs.size()];
			lhs.toArray(lhsAttrs);
			for (int j = 0; j < lhsAttrs.length; j++) {
				String attr = lhsAttrs[j];
				AttributeStringSet lhsWithoutAttr = new AttributeStringSet(lhs);
				lhsWithoutAttr.remove(attr);

				AttributeStringSet attributeClosure = FDAlgorithms.calculateAttributeClosure(
						fdsWork, lhsWithoutAttr);
				//
				if (attributeClosure.containsAll(rhs)) {
					// if the attribute is extraneous,
					// remove it from lhs
					lhs.remove(attr);
				}
			}
		}

		// 2. Right reduction (removes all extraneous attributes from RHS)
		int deletedFDs = 0;
		int initialSize = fdsWork.size();

		for (int i = 0; i < initialSize; i++) {
			int currentPos = i - deletedFDs;
			FD fd = fdsWork.get(currentPos);
			AttributeStringSet lhs = fd.getLHS();
			AttributeStringSet rhs = fd.getRHS();
			String[] rhsAttrs = new String[rhs.size()];
			rhs.toArray(rhsAttrs);
			for (int j = 0; j < rhsAttrs.length; j++) {
				String attr = rhsAttrs[j];
				// we assume that the attribute
				// is extraneous and remove it from rhs
				rhs.remove(attr);

				// fdsWork contains current fd without attribute
				// attr on the rhs
				AttributeStringSet attributeClosure = FDAlgorithms.calculateAttributeClosure(
						fdsWork, lhs);

				if (attributeClosure.contains(attr)) {
					// if attr is extraneous and the rhs
					// is empty, delete the whole fd
					if (rhs.size() == 0) {
						fdsWork.remove(currentPos);
						deletedFDs++;
					}
				} else {
					// if attr is not extraneous, re-add
					// it to rhs
					rhs.add(attr);
				}
			}
		}
		
		return fdsWork;
	}

	/**
	 * Splits the FDs specified by argument fds by using the
	 * augmentation inference rule, e.g.:
	 * <pre>
	 * {a -> bc} => {a -> b, a -> c}
	 * </pre>
	 * 
	 * @param fds a set of FDs
	 * @return a "split" representation of the FDs as described above
	 */
	public static SortedSet<FD> splitFDs(Collection<FD> fds) {
		//variable for the result
		SortedSet<FD> fdsResult = new TreeSet<FD>();
		
		//for each given FD..
		for (FD fd : fds) {
			AttributeStringSet lhs = fd.getLHS();
			AttributeStringSet rhs = fd.getRHS();
			//create a new FD for each attribute on the RHS
			//and add it to the result set
			for (String rhsAttr : rhs) {
				FD newFD = new FD();
				AttributeStringSet singleRHS = new AttributeStringSet();
				singleRHS.add(rhsAttr);
				// create a copy of the lhs
				AttributeStringSet lhsCopy = lhs.copy();
				newFD.setLHS(lhsCopy);
				newFD.setRHS(singleRHS);
				fdsResult.add(newFD);
			}
		}
		
		return fdsResult;
	}

	/**
	 * <p>
	 * Algorithm for determination of candidate keys from
	 * <a href="http://en.wikipedia.org/w/index.php?title=Database_normalization&oldid=616875">
	 * http://en.wikipedia.org/w/index.php?title=Database_normalization&oldid=616875</a>.
	 * </p>
	 * <p>
	 * This algorithm should not be used directly, because it is quite slow. Use the {@link #determineCandidateKeys(AttributeStringSet, Set)}
	 * method. It calls this method, if the candidate keys cannot be found in an easier and faster way.
	 * </p>
	 * <p>
	 * <b>NOTE: </b> The argument minCover must be a minimal cover. If you provide an arbitrary
	 * set of FDs, the result may be wrong!
	 * </p>
	 * @param allAttributes
	 *            all attributes of the table for which the candidate keys should be determined
	 * @param minCover
	 *             a <b>minimal cover</b> of the FDs holding in the table 
	 *             for which the candidate keys should be determined
	 * @return a set containing all candidate keys for the table
	 */
	public static SortedSet<AttributeStringSet> determineCandidateKeysGeneral(
			AttributeStringSet allAttributes, Set<FD> minCover) {
		Set<FD> fds = splitFDs(minCover);

		SortedSet<AttributeStringSet> c = new TreeSet<AttributeStringSet>();
		SortedSet<AttributeStringSet> q = new TreeSet<AttributeStringSet>();
		q.add(new AttributeStringSet(allAttributes));

		while (!q.isEmpty()) {
			AttributeStringSet k = q.first();
			q.remove(k);
			boolean minimal = true;
			for (FD fd : fds) {
				AttributeStringSet lhs = fd.getLHS();
				AttributeStringSet rhs = fd.getRHS();
				AttributeStringSet kDash = new AttributeStringSet();
				kDash.addAll(k);
				kDash.removeAll(rhs);
				kDash.addAll(lhs);
				// if k' is a proper subset of k
				if (SetOperationsUtil.isProperSubset(kDash, k)) {
					minimal = false;
					q.add(kDash);
				}
			}
			boolean properSubsetOfKinC = false;
			for (AttributeStringSet cEntry : c) {
				// if cEntry is a proper subset of k
				if (SetOperationsUtil.isProperSubset(cEntry, k)) {
					properSubsetOfKinC = true;
					break;
				}
			}
			if (minimal && !properSubsetOfKinC) {
				AttributeStringSet[] cCopy = new AttributeStringSet[c.size()];
				c.toArray(cCopy);
				for (AttributeStringSet cEntry : cCopy) {
					if (cEntry.containsAll(k)) {
						c.remove(cEntry);
					}
				}
				c.add(k);
			}

		}

		return c;
	}

	/**
	 * <p>
	 * Algorithm for determination of candidate keys for a table. The allAttributes
	 * argument specifies the table's attributes, the minCover argument specifies
	 * a minimal cover of the FDs holding in the Table. This algorithm first tries
	 * to find the candidate keys by analyzing the structure of the minimal cover and
	 * calls the {@link #determineCandidateKeysGeneral(AttributeStringSet, Set)} method
	 * if the candidate keys cannot be determined this easy (and fast) way.
	 * </p>
	 * <p>
	 * This algorithm is in the most cases faster than the general algorithm
	 * {@link #determineCandidateKeysGeneral(AttributeStringSet, Set)} and
	 * calls the general algorithm only if it cannot find the candidate keys
	 * by analyzing the structure of the minimal cover.
	 * </p>
	 * <p>
	 * This algorithm is based on the Article [An Efficient Algorithm to Compute the Candidate Keys of a Relational Database Schema,
	 * Saiedian/Spencer, The Computer Journal, 1996].
	 * </p>
	 * <p>
	 * <b>NOTE: </b> The argument minCover must be a minimal cover. If you provide an arbitrary
	 * set of FDs, the result may be wrong!
	 * </p>
	 * @param allAttributes
	 *            all attributes of the table for which the candidate keys should be determined
	 * @param minCover
	 *             a <b>minimal cover</b> of the FDs holding in the table 
	 *             for which the candidate keys should be determined
	 * @return a set containing all candidate keys for the table
	 */
	public static SortedSet<AttributeStringSet> determineCandidateKeys(
			AttributeStringSet allAttributes, Set<FD> minCover) {
		//Step 0: initialize variables
		
		TreeSet<AttributeStringSet> candidateKeys = new TreeSet<AttributeStringSet>();
		
		// attributes only contained in LHS of fds
		AttributeStringSet l = new AttributeStringSet();
		// attributes only contained in RHS of fds
		AttributeStringSet r = new AttributeStringSet();
		// attributes contained in both LHS and RHS
		AttributeStringSet b = new AttributeStringSet();
		// attributes contained neither in LHS nor RHS
		AttributeStringSet n = new AttributeStringSet(); 

		// attributes contained in LHS
		AttributeStringSet lTemp = new AttributeStringSet();
		// attributes contained in RHS
		AttributeStringSet rTemp = new AttributeStringSet();

		// Step 1: compute the sets l, r, b and n

		for (FD fd : minCover) {
			AttributeStringSet lhs = fd.getLHS();
			AttributeStringSet rhs = fd.getRHS();

			lTemp.addAll(lhs);
			rTemp.addAll(rhs);
		}

		// compute l as difference between lTemp and rTemp
		l = new AttributeStringSet(lTemp);
		l.removeAll(rTemp);

		// compute r as difference between rTemp and lTemp
		r = new AttributeStringSet(rTemp);
		r.removeAll(lTemp);

		// compute b as intersection between lTemp and rTemp
		b = new AttributeStringSet(lTemp);
		b.retainAll(rTemp);

		// compute n as difference between allAttributes and the union of
		// lTemp and rTemp
		n = new AttributeStringSet(allAttributes);
		n.removeAll(lTemp);
		n.removeAll(rTemp);

		// Step 2: Check if the union of the sets l and n contains all attributes

		// ln is the union of l and n
		AttributeStringSet ln = new AttributeStringSet(l);
		ln.addAll(n);

		// calculate the attribute closure lnPlus for ln
		AttributeStringSet lnPlus = calculateAttributeClosure(minCover, l);

		// if lnPlus = allAttributes, ln is the only candidate key
		if (lnPlus.containsAll(allAttributes)) {
			candidateKeys.add(ln);
			return candidateKeys;
		} else {
			// Step 3

			// compute the initial superkey as the difference between
			// allAttributes and r
			AttributeStringSet initialSuperkey = new AttributeStringSet(allAttributes);
			initialSuperkey.removeAll(r);

			// call general algorithm
			return determineCandidateKeysGeneral(initialSuperkey, minCover);
		}

	}

	/**
	 * <p>
	 * Computes and sets the candidate keys for a table and sets the PK (and where required, DB-PK)
	 * which is considered to be the best. For selecting the PK currently a trivial and
	 * in some cases insufficient approach has been taken, as you can see in the
	 * source code of this method.
	 * </p>
	 * <p>
	 * <b>NOTE: </b> The minimal cover for this table must have been computed before and
	 * must have been set with the {@link Table#setFDMinCover(Set)} method.
	 * </p> 
	 * @param table
	 *            the table to set the keys for
	 * @param oldTable
	 *            the table from which the table was extracted by means of normalization, can be
	 *            <code>null</code>
	 */
	public static void setKeysForTable(Table table, Table oldTable) {
		// 1. set the candidate keys for the table

		Set<AttributeStringSet> keyAttrsSet = FDAlgorithms.determineCandidateKeys(table
				.getAttributes(), table.getFDMinCover());
		LinkedHashSet<CandKey> candKeySet = new LinkedHashSet<CandKey>();
		for (AttributeStringSet attrs : keyAttrsSet) {
			CandKey candKey = new CandKey();
			candKey.setAttributes(new ArrayList<String>(attrs));
			candKeySet.add(candKey);
		}

		table.setCandKeys(candKeySet);

		// 2. set a primary key for the table

		// variable which states if the PK has still to be set
		boolean pkToSet = true;

		// if the table has already a pk defined, it
		// remains the pk if it is contained in the candidate keys
		if (table.getPk() != null && table.getPk().getAttributes().size() > 0) {
			if (containsKey(candKeySet, table.getPk().getAttributes())) {
				pkToSet = false;
			}
		}

		// if the table is part of a normalization proposal (oldTable exists),
		// set the PK of this table if it is contained in the candidate keys
		if (pkToSet && oldTable != null && oldTable.getPk() != null
				&& oldTable.getPk().getAttributes().size() > 0) {
			if (containsKey(candKeySet, oldTable.getPk().getAttributes())) {
				PK pk = new PK();
				pk.setAttributes(new ArrayList<String>(oldTable.getPk().getAttributes()));
				table.setPk(pk);
				pkToSet = false;
			}
		}

		// if the table definition contains a PK (only is possible if the
		// table exists in a DB), this DB-PK is set as PK if it is contained
		// in the candidate keys
		if (pkToSet && table.getDbPK() != null && table.getDbPK().getAttributes().size() > 0) {
			if (containsKey(candKeySet, table.getDbPK().getAttributes())) {
				PK pk = new PK();
				pk.setAttributes(new ArrayList<String>(table.getDbPK().getAttributes()));
				table.setPk(pk);
				pkToSet = false;
			}
		}

		// if no PK could be set in the previous steps,
		// select a minimal candidate key as PK
		if (pkToSet) {
			ArrayList<String> pkAttrs = null;
			for (CandKey candKey : candKeySet) {
				List<String> candKeyAttrs = candKey.getAttributes();
				// if this key is the first or it is smaller than
				// the hitherto smallest key
				if (pkAttrs == null || candKeyAttrs.size() < pkAttrs.size()) {
					pkAttrs = new ArrayList<String>(candKeyAttrs);
				}
				PK pk = new PK();
				pk.setAttributes(pkAttrs);
				table.setPk(pk);
			}

		}

		// if the table is part of a normalization proposal
		// and it is a DB-Table (we want to generate a sql script),
		// set the pk as dbPK
		if (oldTable != null && oldTable.isDbTable()) {
			PK dbPk = new PK();
			dbPk.setAttributes(new ArrayList<String>(table.getPk().getAttributes()));
			table.setDbPK(dbPk);
		}

	}

	/**
	 * Helper method which determines if a key consisting of the attributes specified
	 * by attributes argument is contained in the candidate keys specified by the candidateKeys argument
	 * @param candidateKeys a collection of candidate keys
	 * @param attributes the attributes of a potential key
	 * @return <code>true</code>, if such a key is contained in the candidate keys;
	 * <code>false</code>, otherwise
	 */
	private static boolean containsKey(Collection<CandKey> candidateKeys,
			Collection<String> attributes) {
		boolean containsPk = false;
		for (CandKey candKey : candidateKeys) {
			List<String> candKeyAttrs = candKey.getAttributes();
			if (candKeyAttrs.containsAll(attributes)
					&& candKeyAttrs.size() == attributes.size()) {
				containsPk = true;
				break;
			}
		}
		return containsPk;
	}
}
