package de.fesenmeyer.dbnormalizer.core.algorithms;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import de.fesenmeyer.dbnormalizer.core.FD;
import de.fesenmeyer.dbnormalizer.core.FDCheckState;
import de.fesenmeyer.dbnormalizer.core.dbaccess.DBException;
import de.fesenmeyer.dbnormalizer.core.dbaccess.DataSet;
import de.fesenmeyer.dbnormalizer.core.log.DBNormalizerLogger;
import de.fesenmeyer.dbnormalizer.core.util.MiscUtil;

/**
 * <p>
 * Class which gives the possibility to check if a FD
 * is satisfied by the current content of a Table (method {@link #checkFD(String, FD)}).
 * This is done by constructing an appropriate SQL query based on the FD and
 * running this query against the DB.
 * </p>
 * @author DF
 *
 */
public class FDChecker {

	/**
	 * The logger for this class.
	 */
	private static DBNormalizerLogger logger = DBNormalizerLogger.getLogger(FDChecker.class);

	/**
	 * The SQL-Connection to be used to access the DB.
	 */
	private Connection conn = null;

	/**
	 * Constant for the name of the derived table constructed by the SQL Statement.
	 */
	private static final String DERIVED_TABLE_ALIAS = "distinctrows";

	/**
	 * Class constructor specifying the SQL-Connection to be used.
	 * @param conn a SQL-Connection
	 */
	public FDChecker(Connection conn) {
		this.conn = conn;
	}

	/**
	 * <p>
	 * Checks if the FD specified by argument fd is satisfied in the table specified
	 * by argument qualifiedTableName. As a first step, the method checks if the
	 * table contains data. If it does not contain any data, it is not possible
	 * to check the FD.
	 * <p>
	 * To check if the FD
	 * <pre>
	 * {lectureno, matrno} -> {name, semester}
	 * </pre>
	 * holds in the table with the qualified name <code>university.students</code>,
	 * the following query is run:
	 * <pre>
	 * SELECT lectureno, matrno, COUNT(*) AS rhscount
	 * FROM	(SELECT DISTINCT lectureno, matrno, name, semester
	 * 	FROM university.students) distinctrows
	 * GROUP BY lectureno, matrno HAVING COUNT(*) > 1
	 * </pre>
	 * If the query does not return any row, the FD is satisfied. Otherwise, it is
	 * not satisfied.
	 * <p>
	 * @param qualifiedTableName the qualified name of a table
	 * @param fd a FD possibly holding in the table specified by argument qualifiedTableName
	 * @return Returns
	 * <ul>
	 * 	<li>{@link FDCheckState#NO_DATA}, if the table does not contain any row of data.</li>
	 * 	<li>{@link FDCheckState#SATISFIED}, if the query described above does not return any row.</li>
	 *  <li>{@link FDCheckState#NOT_SATISFIED}, if the query described above returns at least one row.</li>
	 * </ul>
	 * @throws DBException if an SQL Exception occurs, e.g. because the table or the
	 * attributes of the FD do not exist, or because the DBS does not support the
	 * syntax used in the created SQL query
	 */
	public FDCheckState checkFD(String qualifiedTableName, FD fd) throws DBException {
		FDCheckState fdCheckState = null;

		try {
			Statement stmt = conn.createStatement();
			ResultSet rs = null;

			// check at first if there are any rows in the table
			String strStmtCheckNoData = getCheckNoDataStatement(qualifiedTableName);
			if (logger.isDebugLoggable()) {
				logger.debug("Performing No-Data-Check for Table '" + qualifiedTableName
						+ "', FD '" + fd.toString() + "', SQL Statement: '" + strStmtCheckNoData
						+ "'");
			}
			rs = stmt.executeQuery(strStmtCheckNoData);
			// the resultset should return the rowCount as scalar result
			// --> we should get just one row and column
			if (rs.next()) {
				int rowCount = rs.getInt(1);
				if (rowCount == 0) {
					fdCheckState = FDCheckState.NO_DATA;
				}
			} else {
				logger.error("Unexpected Exception: SQL Statement '" + strStmtCheckNoData
						+ "' does not return a row");
			}
			rs.close();

			// if fdCheckState is still null (--> that means if data exist in the table),
			// check if these data satisfy the FD
			if (fdCheckState == null) {
				String strStmtCheckFD = getCheckFDStatement(qualifiedTableName, fd);
				if (logger.isDebugLoggable()) {
					logger.debug("Performing FD Check for Table '" + qualifiedTableName + "', FD '"
							+ fd.toString() + "', SQL Statement: '" + strStmtCheckFD + "'");
				}
				stmt.executeQuery(strStmtCheckFD);
				rs = stmt.getResultSet();
				if (rs.next()) {
					fdCheckState = FDCheckState.NOT_SATISFIED;
				} else {
					fdCheckState = FDCheckState.SATISFIED;
				}

				rs.close();
			}

			stmt.close();
		} catch (SQLException e) {
			throw new DBException(e);
		}
		return fdCheckState;
	}

	/**
	 * Returns the tuples returned by the SQL query used for checking FDs.
	 * These tuples are returned in a {@link DataSet} object.
	 * @param qualifiedTableName a qualified table name
	 * @param fd a FD
	 * @return a DataSet object, which may contain no data, if the FD is satisfied or
	 * if the table does not contain any data
	 * @throws DBException if an SQL Exception occurs, e.g. because the table or the
	 * attributes of the FD do not exist, or because the DBS does not support the
	 * syntax used in the created SQL query
	 * @see #checkFD(String, FD)
	 */
	public DataSet getNotSatisfiedTupels(String qualifiedTableName, FD fd) throws DBException {
		DataSet dataSet = null;
		try {
			Statement stmt = conn.createStatement();
			String strStmtCheckFD = getCheckFDStatement(qualifiedTableName, fd);
			if (logger.isDebugLoggable()) {
				logger.debug("Retrieving not satisfied tupels for Table '" + qualifiedTableName
						+ "', FD '" + fd.toString() + "', SQL Statement: '" + strStmtCheckFD + "'");
			}
			stmt.executeQuery(strStmtCheckFD);
			ResultSet rs = stmt.getResultSet();
			dataSet = new DataSet(rs);
			stmt.close();
			rs.close();
		} catch (SQLException e) {
			throw new DBException(e);
		}
		return dataSet;
	}

	/**
	 * Creates an SQL query to check if the table specified by argument
	 * qualifiedTableName contains any data. This SQL query returns a
	 * single value, which is the number of rows in the table (the table's
	 * row count).
	 * @param qualifiedTableName
	 * @return an SQL query as described above
	 */
	private String getCheckNoDataStatement(String qualifiedTableName) {
		StringBuffer sb = new StringBuffer();
		sb.append("SELECT COUNT(*) AS rowCount ");
		sb.append("FROM ");
		sb.append(qualifiedTableName);
		return sb.toString();
	}

	/**
	 * Creates an SQL Statement, which can be used to check if the FD specified by
	 * the fd argument is satisfied by the Table specified by the qualifiedTableName
	 * argument.
	 * @param qualifiedTableName a qualified table name
	 * @param fd a FD
	 * @return an SQL statement for checking the FD
	 * @see #checkFD(String, FD)
	 */
	private String getCheckFDStatement(String qualifiedTableName, FD fd) {
		StringBuffer sb = new StringBuffer();
		sb.append("SELECT ");
		StringBuffer lhsAttrList = MiscUtil.getCollectionAsStringBuffer(fd.getLHS());
		sb.append(lhsAttrList);
		sb.append(", COUNT(*) AS rhscount ");
		sb.append("FROM (SELECT DISTINCT ");
		sb.append(lhsAttrList);
		sb.append(", ");
		StringBuffer rhsAttrList = MiscUtil.getCollectionAsStringBuffer(fd.getRHS());
		sb.append(rhsAttrList);
		sb.append(" FROM ");
		sb.append(qualifiedTableName);
		sb.append(") ");
		sb.append(DERIVED_TABLE_ALIAS);
		sb.append(" GROUP BY ");
		sb.append(lhsAttrList);
		sb.append(" HAVING COUNT(*) > 1");
		return sb.toString();
	}
}
