package de.fesenmeyer.dbnormalizer.core.algorithms;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import de.fesenmeyer.dbnormalizer.core.AttributeStringSet;
import de.fesenmeyer.dbnormalizer.core.CandKey;
import de.fesenmeyer.dbnormalizer.core.DBNormalizerException;
import de.fesenmeyer.dbnormalizer.core.FD;
import de.fesenmeyer.dbnormalizer.core.NormalForm;
import de.fesenmeyer.dbnormalizer.core.Table;
import de.fesenmeyer.dbnormalizer.core.log.DBNormalizerLogger;
import de.fesenmeyer.dbnormalizer.core.util.SetOperationsUtil;

/**
 * This class implements an algorithm for the determination of the normal form
 * of a table (from 1NF to BCNF). The algorithms are based on the definitions 
 * in [Fundamentals of database systems, 5th Ed., Elmasri/Navathe, 2007].
 * @author DF
 */
public class NFDetermination {

	/**
	 * The logger for this class.
	 */
	private static DBNormalizerLogger logger = DBNormalizerLogger.getLogger(NFDetermination.class);

	/**
	 * The candidate keys of the table.
	 */
	private Set<AttributeStringSet> candidateKeys = null;

	/**
	 * The table for which the normal form should be determined.
	 */
	private Table table = null;

	/**
	 * The FDs of the table.
	 */
	private Set<FD> fds = null;

	/**
	 * flag, which is <code>true</code>, if the table only contains
	 * "simple" keys (keys consisting of one attribute)
	 */
	private boolean onlySimpleKeys = false;

	
	/**
	 * Map, which holds for each key (normal form)
	 * the FDs which violate this normal form
	 */
	private Map<NormalForm, Set<FD>> violatingFDs = null;

	/**
	 * Class Constructor specifying the table for which the
	 * normal form should be determined.
	 * @param table a Table object
	 */
	public NFDetermination(Table table) {
		this.table = table;
	}

	/**
	 * Helper method for initialization of the member variables
	 * of this class.
	 * @param table the table
	 * @throws DBNormalizerException
	 */
	private void initialize(Table table) throws DBNormalizerException {
		// throw an exception if the candidate keys have not yet been computed
		if (table.getCandKeys() == null || table.getCandKeys().size() == 0) {
			throw new DBNormalizerException(
					"The candidate keys for the table '"
							+ table.getQualifiedName()
							+ "' have not yet been computed. Without the candidate keys, the normal form cannot be determined.");
		}

		this.candidateKeys = new TreeSet<AttributeStringSet>();

		// create an AttributeStringSet of the attributes of each
		// candidate key and add it to the candidate key set
		this.candidateKeys = new TreeSet<AttributeStringSet>();
		for (CandKey candKey : table.getCandKeys()) {
			this.candidateKeys.add(new AttributeStringSet(candKey.getAttributes()));
		}

		// use the minimal cover of the table to determine the NF
		this.fds = table.getFDMinCover();

		// initialize the violating FDs
		violatingFDs = new HashMap<NormalForm, Set<FD>>();

		// check, if there are only candidate keys
		// containing just one attributes ("simple keys"):
		// if true, 2NF cannot be violated, which makes it possible
		// to improve performance of the 2NF check
		this.onlySimpleKeys = true;
		for (AttributeStringSet key : candidateKeys) {
			if (key.size() > 1) {
				onlySimpleKeys = false;
				break;
			}
		}
	}

	/**
	 * <p>
	 * The normal form determination algorithm, which returns the normal form
	 * and the FDs violating one of the normal forms 2NF, 3NF or BCNF.
	 * </p>
	 * <p>
	 * <b>NOTE</b>: Before applying this method to a table, make sure that it contains
	 * all the necessary data such as minimal cover ({@link Table#setFDMinCover(java.util.Set)})
	 * and candidate keys ({@link Table#setCandKeys(java.util.Set)}). The methods for calculating
	 * this data can be found in the {@link FDAlgorithms} class.
	 * </p>
	 * @return the result of the normal form determination
	 * @throws DBNormalizerException
	 * @see NFDeterminationResult
	 */
	public NFDeterminationResult determine() throws DBNormalizerException {
		initialize(this.table);

		//find all FDs which violate 2NF, 3NF and BCNF (other normal forms
		//are not considered)
		for (FD fd : fds) {
			boolean violates = violates2NF(fd);
			if (violates) {
				addToViolatingFDs(NormalForm.NF2, fd);
			} else {
				violates = violates3NF(fd);
				if (violates) {
					addToViolatingFDs(NormalForm.NF3, fd);
				} else {
					violates = violatesBCNF(fd);
					if (violates) {
						addToViolatingFDs(NormalForm.BCNF, fd);
					}
				}
			}
		}

		//determine the normal form based
		//on the normal form violations
		NormalForm nf = NormalForm.BCNF;
		if (violatingFDs.containsKey(NormalForm.BCNF)) {
			nf = NormalForm.NF3;
		}
		if (violatingFDs.containsKey(NormalForm.NF3)) {
			nf = NormalForm.NF2;
		}
		if (violatingFDs.containsKey(NormalForm.NF2)) {
			nf = NormalForm.NF1;
		}

		NFDeterminationResult result = new NFDeterminationResult(candidateKeys, table, fds, nf,
				violatingFDs);

		return result;
	}

	/**
	 * Checks if a FD violates 2NF.
	 * @param fd the FD to be checked
	 * @return <code>true</code>, if the FD violates 2NF;
	 * <code>false</code>, otherwise
	 */
	private boolean violates2NF(FD fd) {
		boolean violates = false;

		// if there are keys with more than one attribute
		if (!onlySimpleKeys) {
			AttributeStringSet lhs = fd.getLHS();
			AttributeStringSet rhs = fd.getRHS();

			for (AttributeStringSet key : candidateKeys) {
				// if lhs is a proper subset of the key
				if (SetOperationsUtil.isProperSubset(lhs, key)) {
					for (String rhsAttr : rhs) {
						// if the attribute is non prime
						if (!isKeyAttribute(rhsAttr)) {
							violates = true;
							break;
						}
					}
				}
				if (violates) {
					break;
				}
			}
		}

		return violates;
	}

	/**
	 * Checks if a FD violates 3NF.
	 * @param fd the FD to be checked
	 * @return <code>true</code>, if the FD violates 3NF;
	 * <code>false</code>, otherwise
	 */
	private boolean violates3NF(FD fd) {
		boolean violates = false;
		AttributeStringSet lhs = fd.getLHS();
		AttributeStringSet rhs = fd.getRHS();

		boolean lhsIsSuperkey = false;
		// because the set of fds is minimal,
		// the lhs of a fd must be a key, if it is a superkey
		// --> we just need to check if the lhs is a key, we
		// need not to compute and check all superkeys
		if (candidateKeys.contains(lhs)) {
			lhsIsSuperkey = true;
		}

		// only if lhs is not superkey, it is possible
		// that the fd violates 3NF
		if (!lhsIsSuperkey) {
			// if there are any non prime attributes on rhs,
			// then the fd violates 3NF
			for (String rhsAttr : rhs) {
				if (!isKeyAttribute(rhsAttr)) {
					violates = true;
					break;
				}
			}
		}

		return violates;
	}

	/**
	 * Checks if a FD violates BCNF.
	 * @param fd the FD to be checked
	 * @return <code>true</code>, if the FD violates BCNF;
	 * <code>false</code>, otherwise
	 */
	private boolean violatesBCNF(FD fd) {
		AttributeStringSet lhs = fd.getLHS();

		boolean lhsIsSuperkey = false;
		// because the set of fds is minimal,
		// the lhs of a fd must be a key, if it is a superkey
		// --> we just need to check if the lhs is a key, we
		// need not to compute and check all superkeys
		if (candidateKeys.contains(lhs)) {
			lhsIsSuperkey = true;
		}

		return !lhsIsSuperkey;
	}

	/**
	 * Helper method which looks if the attribute specified by argument attr
	 * is part of a candidate key (a prime attribute).
	 * @param attr an attribute name
	 * @return <code>true</code>, if the attribute is part of a candidate key;
	 * <code>false</code>, otherwise
	 */
	private boolean isKeyAttribute(String attr) {
		boolean isKeyAttr = false;
		for (AttributeStringSet key : candidateKeys) {
			if (key.contains(attr)) {
				isKeyAttr = true;
				break;
			}
		}
		return isKeyAttr;
	}

	/**
	 * Helper method which adds a FD to the map
	 * of violating FDs.
	 * @param nf the normal form which is violated by the FD
	 * @param fd the FD
	 */
	private void addToViolatingFDs(NormalForm nf, FD fd) {
		Set<FD> fds = violatingFDs.get(nf);
		if (fds == null) {
			fds = new TreeSet<FD>();
			violatingFDs.put(nf, fds);
		}
		fds.add(fd);
	}

}
