package de.fesenmeyer.dbnormalizer.core.algorithms;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;

import de.fesenmeyer.dbnormalizer.core.AttributeStringSet;
import de.fesenmeyer.dbnormalizer.core.FK;
import de.fesenmeyer.dbnormalizer.core.Key;
import de.fesenmeyer.dbnormalizer.core.Table;
import de.fesenmeyer.dbnormalizer.core.log.DBNormalizerLogger;
import de.fesenmeyer.dbnormalizer.core.util.MiscUtil;

/**
 * <p>
 * Class which can be used for determination of the foreign keys (FKs) which
 * should be defined for the tables of a normalization proposal. The implementation
 * is based on the assumption that foreign keys cannot reference arbitrary
 * candidate keys, but only primary keys.
 * </p>
 * <p>
 * <b>NOTE: </b>The current implementation only leads to an optimal solution, if
 * each of the created tables of the normalization proposal just contains
 * one candidate key. For a better solution, the selection of the primary keys for
 * the created tables could possibly be made dependent on which foreign keys can be
 * defined between the tables.
 * </p>
 * @author DF
 * @see NormalizationFKsProposal
 */
public class NormalizationFKsDetermination {

	/**
	 * The logger for this class.
	 */
	private static DBNormalizerLogger logger = DBNormalizerLogger
			.getLogger(NormalizationFKsDetermination.class);

	/**
	 * The normalization proposal for which FKs should be proposed.
	 */
	private NormalizationProposal normProposal = null;

	/**
	 * Class constructor specifying the normalization proposal for
	 * which FKs should be proposed.
	 * @param normProposal the normalization proposal
	 */
	public NormalizationFKsDetermination(NormalizationProposal normProposal) {
		this.normProposal = normProposal;
	}

	/**
	 * Computes a proposal of FKs to be defined and sets
	 * these FKs for the tables which are returned by the
	 * {@link NormalizationProposal#getCreatedTables()} method.
	 * @return the FKs proposal
	 * @see NormalizationFKsProposal
	 */
	public NormalizationFKsProposal getNormalizationFKsProposal() {
		NormalizationFKsProposal fksProp = new NormalizationFKsProposal();

		ArrayList<FK> innerFKs = computeInnerFKs();
		fksProp.setInnerFKS(innerFKs);

		ArrayList<FK> importedFKs = computeImportedFKs();
		fksProp.setImportedFKs(importedFKs);

		ArrayList<FK> exportedFKs = computeExportedFKs();
		fksProp.setExportedFKs(exportedFKs);

		ArrayList<FK> hierarchicalFKs = computeHierarchicalFKs();
		fksProp.setHierarchicalFKs(hierarchicalFKs);

		// assure that the imported and exported keys for each table
		// are empty at the beginning
		for (Table table : this.normProposal.getCreatedTables()) {
			table.getImportedKeys().clear();
			table.getExportedKeys().clear();
		}

		// use the computed inner FKs, exported FKs and hierarchical FKs
		// to add exported keys to the tables of the normalization proposal
		ArrayList<FK> exportedFKList = new ArrayList<FK>();
		exportedFKList.addAll(innerFKs);
		exportedFKList.addAll(exportedFKs);
		exportedFKList.addAll(hierarchicalFKs);
		addExportedKeysToTables(exportedFKList, this.normProposal.getCreatedTables());

		// use the computed inner FKs, imported FKs and hierarchical FKs
		// to add exported keys to the tables of the normalization proposal
		ArrayList<FK> importedFKList = new ArrayList<FK>();
		importedFKList.addAll(innerFKs);
		importedFKList.addAll(importedFKs);
		importedFKList.addAll(hierarchicalFKs);
		addImportedKeysToTables(importedFKList, this.normProposal.getCreatedTables());

		return fksProp;
	}

	/**
	 * Helper method which adds each FK given by the fks argument as exported key
	 * to the appropriate table contained in the collection given by
	 * the tables argument, if possible.
	 * @param fks a collection of FKs
	 * @param tables a collection of tables
	 */
	private void addExportedKeysToTables(Collection<FK> fks, Collection<Table> tables) {
		// a map for the tables (key is qualified name)
		HashMap<String, Table> tablesMap = new HashMap<String, Table>(tables.size());

		for (Table table : tables) {
			tablesMap.put(table.getQualifiedName(), table);
		}
		for (FK fk : fks) {
			String pkTableQName = MiscUtil.getQualifiedTableName(fk.getPkTableName(), fk
					.getPkCatalog(), fk.getPkSchema());
			Table pkTable = tablesMap.get(pkTableQName);
			if (pkTable == null) {
				logger.debug("Unexpected Exception: PK-Table for FK '" + fk + "' was not found "
						+ "when adding computed FKs to tables");
			} else {
				pkTable.getExportedKeys().add(fk);
			}
		}
	}

	/**
	 * Helper method which adds each FK given by the fks argument as imported key
	 * to the appropriate table contained in the collection given by
	 * the tables argument, if possible.
	 * @param fks a collection of FKs
	 * @param tables a collection of tables
	 */
	private void addImportedKeysToTables(Collection<FK> fks, Collection<Table> tables) {
		// a map for the tables (key is qualified name)
		HashMap<String, Table> tablesMap = new HashMap<String, Table>(tables.size());

		for (Table table : tables) {
			tablesMap.put(table.getQualifiedName(), table);
		}
		for (FK fk : fks) {
			String fkTableQName = MiscUtil.getQualifiedTableName(fk.getFkTableName(), fk
					.getFkCatalog(), fk.getFkSchema());
			Table fkTable = tablesMap.get(fkTableQName);
			if (fkTable == null) {
				logger.debug("Unexpected Exception: FK-Table for FK '" + fk + "' was not found "
						+ "when adding computed FKs to tables");
			} else {
				fkTable.getImportedKeys().add(fk);
			}
		}
	}

	/**
	 * Helper method which computes FKs to be defined
	 * due to the old table's exported keys.
	 * @return the list of FKs
	 */
	private ArrayList<FK> computeExportedFKs() {
		ArrayList<FK> newFKs = new ArrayList<FK>();

		Table oldTable = normProposal.getOldTable();
		Collection<FK> oldFKs = oldTable.getExportedKeys();

		// if the old table has exported keys
		if (oldFKs.size() > 0) {

			Key oldTblDBPK = oldTable.getDbPK();
			Key oldTblPK = oldTable.getPk();
			if (oldTblDBPK == null) {
				logger.error("There must be an error in the program code: The table '"
						+ oldTable.getQualifiedName() + "' has exported keys, "
						+ "but no pk is defined for it in the db.");
			} else if (oldTblPK == null) {
				logger.warn("There has no PK been selected for table '"
						+ oldTable.getQualifiedName()
						+ "'. Exported keys of the table will get lost!");
			} else {
				List<String> oldDBPkAttrs = oldTblDBPK.getAttributes();
				List<String> oldPkAttrs = oldTblPK.getAttributes();

				// check if the attributes of the selected PK and the DB-defined PK are equal,
				// if not, the exported keys will get lost!
				if (!(oldDBPkAttrs.containsAll(oldPkAttrs) && oldDBPkAttrs.size() == oldPkAttrs
						.size())) {
					logger.warn("The selected PK and the DB-defined PK for table '"
							+ oldTable.getQualifiedName()
							+ "' do not match. Exported keys will get lost!");
				} else {
					for (FK oldFK : oldFKs) {
						Collection<Table> createdTables = normProposal.getCreatedTables();
						for (Table table : createdTables) {
							List<String> pkAttrs = table.getPk().getAttributes();
							// if the current table has the same pk as the original table
							if (oldPkAttrs.containsAll(pkAttrs)
									&& oldPkAttrs.size() == pkAttrs.size()) {

								// assure that no hierarchical FKs are created here
								boolean isHierarchicalFK = MiscUtil.checkTableEquality(oldTable
										.getName(), oldTable.getCatalog(), oldTable.getSchema(),
										oldFK.getFkTableName(), oldFK.getFkCatalog(), oldFK
												.getFkSchema());
								if (!isHierarchicalFK) {
									FK fk = new FK();
									fk.setFkCatalog(oldFK.getFkCatalog());
									fk.setFkSchema(oldFK.getFkSchema());
									fk.setFkTableName(oldFK.getFkTableName());

									fk.setPkCatalog(table.getCatalog());
									fk.setPkSchema(table.getSchema());
									fk.setPkTableName(table.getName());

									fk.setFkColumnNames(oldFK.getFkColumnNames());
									fk.setPkColumnNames(oldPkAttrs);
									
									//copy the update and delete rule
									fk.setUpdateRule(oldFK.getUpdateRule());
									fk.setDeleteRule(oldFK.getDeleteRule());
									
									newFKs.add(fk);

									// the pk of the original table should only
									// be the pk of one of the created tables
									// (at least if standard normalization algorithms
									// are used), therefore we break the iteration here
									break;
								}
							}
						}
					}
				}
			}
		}

		return newFKs;
	}

	/**
	 * Helper method which computes FKs to be defined
	 * due to the old table's imported keys.
	 * @return the list of FKs
	 */
	private ArrayList<FK> computeImportedFKs() {
		ArrayList<FK> newFKs = new ArrayList<FK>();

		Table oldTable = normProposal.getOldTable();
		Collection<FK> oldFKs = oldTable.getImportedKeys();
		for (FK oldFK : oldFKs) {
			List<String> fkAttrs = oldFK.getFkColumnNames();

			Collection<Table> createdTables = normProposal.getCreatedTables();
			// look in each table, if it contains all of the old
			// fk's attributes
			for (Table table : createdTables) {
				Collection<String> tblAttrs = table.getAttributes();
				if (tblAttrs.containsAll(fkAttrs)) {
					// assure that no hierarchical FKs are created here
					boolean isHierarchicalFK = MiscUtil.checkTableEquality(oldTable.getName(),
							oldTable.getCatalog(), oldTable.getSchema(), oldFK.getPkTableName(),
							oldFK.getPkCatalog(), oldFK.getPkSchema());
					if (!isHierarchicalFK) {
						FK fk = new FK();
						fk.setFkCatalog(table.getCatalog());
						fk.setFkSchema(table.getSchema());
						fk.setFkTableName(table.getName());

						fk.setPkCatalog(oldFK.getPkCatalog());
						fk.setPkSchema(oldFK.getPkSchema());
						fk.setPkTableName(oldFK.getPkTableName());

						fk.setFkColumnNames(fkAttrs);
						fk.setPkColumnNames(oldFK.getPkColumnNames());
						
						//copy the update and delete rule
						fk.setUpdateRule(oldFK.getUpdateRule());
						fk.setDeleteRule(oldFK.getDeleteRule());

						newFKs.add(fk);
					}
				}
			}
		}

		return newFKs;
	}

	/**
	 * Helper method which computes FKs to be defined
	 * between the created tables of the normalization proposal.
	 * @return the list of FKs
	 */
	private ArrayList<FK> computeInnerFKs() {
		ArrayList<FK> innerFKs = new ArrayList<FK>();
		// for each pk of a table, create a fk to it for each
		// table which contains the pk's attributes
		Collection<Table> createdTables = normProposal.getCreatedTables();
		for (Table pkTable : createdTables) {
			Key pk = pkTable.getPk();
			for (Table table : createdTables) {
				AttributeStringSet tblAttrs = table.getAttributes();
				// check if table is the same as pkTable to avoid a cyclic FK
				boolean isSameTable = table.equals(pkTable);
				if (tblAttrs.containsAll(pk.getAttributes()) && !isSameTable) {
					FK fk = new FK();
					fk.setFkCatalog(table.getCatalog());
					fk.setFkSchema(table.getSchema());
					fk.setFkTableName(table.getName());

					fk.setPkCatalog(pkTable.getCatalog());
					fk.setPkSchema(pkTable.getSchema());
					fk.setPkTableName(pkTable.getName());

					// for inner fks, fk and pk attributes must have the same names
					List<String> columnNames = new ArrayList<String>(pk.getAttributes());
					fk.setFkColumnNames(columnNames);
					fk.setPkColumnNames(columnNames);

					innerFKs.add(fk);
				}
			}
		}

		return innerFKs;
	}

	/**
	 * Helper method which computes FKs to be defined
	 * due to FKs between the old table's attributes ("hierarchical FKs").
	 * @return the list of FKs
	 */
	private ArrayList<FK> computeHierarchicalFKs() {
		ArrayList<FK> newFKs = new ArrayList<FK>();

		Table oldTable = normProposal.getOldTable();
		Collection<FK> oldFKs = oldTable.getExportedKeys();

		for (FK oldFK : oldFKs) {
			// assure that no hierarchical FKs are created here
			boolean isHierarchicalFK = MiscUtil.checkTableEquality(oldTable.getName(), oldTable
					.getCatalog(), oldTable.getSchema(), oldFK.getFkTableName(), oldFK
					.getFkCatalog(), oldFK.getFkSchema());
			if (isHierarchicalFK) {
				// get the table which contains the PK of the FK
				Table pkTable = null;
				List<String> pkAttrs = oldFK.getPkColumnNames();
				for (Table table : normProposal.getCreatedTables()) {
					if (table.getPk().getAttributes().containsAll(pkAttrs)
							&& table.getPk().getAttributes().size() == pkAttrs.size()) {
						pkTable = table;
						break;
					}
				}

				if (pkTable == null) {
					logger.warn("The hierarchical FK '" + oldFK + "' will get lost, because "
							+ "there is no new table containing the FK's pk columns.");
				} else {
					// set an FK to all tables which contain the old fk's columns
					List<String> fkAttrs = oldFK.getFkColumnNames();
					for (Table table : normProposal.getCreatedTables()) {
						if (table.getAttributes().containsAll(fkAttrs)) {
							FK fk = new FK();
							fk.setFkCatalog(table.getCatalog());
							fk.setFkSchema(table.getSchema());
							fk.setFkTableName(table.getName());

							fk.setPkCatalog(pkTable.getCatalog());
							fk.setPkSchema(pkTable.getSchema());
							fk.setPkTableName(pkTable.getName());

							fk.setFkColumnNames(fkAttrs);
							fk.setPkColumnNames(pkAttrs);
							
							//copy the update and delete rule
							fk.setUpdateRule(oldFK.getUpdateRule());
							fk.setDeleteRule(oldFK.getDeleteRule());

							newFKs.add(fk);
						}
					}
				}
			}
		}

		return newFKs;
	}

}
