package de.fesenmeyer.dbnormalizer.core.algorithms;

import java.util.SortedSet;
import java.util.TreeSet;

import de.fesenmeyer.dbnormalizer.core.NormalForm;
import de.fesenmeyer.dbnormalizer.core.Table;
import de.fesenmeyer.dbnormalizer.core.log.DBNormalizerLogger;

/**
 * Class representing a normalization proposal.
 * @author DF
 * @see AbstractNormalizationAlgorithm#normalize(Table)
 */
public class NormalizationProposal {
	/**
	 * The logger for this class.
	 */
	private static DBNormalizerLogger logger = DBNormalizerLogger
			.getLogger(NormalizationProposal.class);
	/**
	 * The old table (the table for which the normalization
	 * proposal has been generated).
	 */
	private Table oldTable;
	/**
	 * The tables to be created due to the normalization proposal.
	 */
	private SortedSet<Table> createdTables = new TreeSet<Table>();
	/**
	 * The foreign keys (FKs) to be created due to the normalization
	 * proposal.
	 */
	private NormalizationFKsProposal fksProposal = null;
	/**
	 * The SQL script generated for normalization. Is <code>null</code>,
	 * if the old table does not exist in a DB.
	 */
	private String sqlScript = null;

	/**
	 * Class Constructor specifying the old table (the table for which the normalization
	 * proposal has been generated).
	 * @param oldTable the old table
	 */
	public NormalizationProposal(Table oldTable) {
		this.oldTable = oldTable;
	}

	/**
	 * Gets the old table (the table for which the normalization
	 * proposal has been generated).
	 * @return the old table
	 */
	public Table getOldTable() {
		return oldTable;
	}

	/**
	 * Adds a table to the tables to be created due to the normalization proposal.
	 * @param table a table
	 */
	public void addCreatedTable(Table table) {
		this.createdTables.add(table);
	}

	/**
	 * Gets the tables to be created due to the normalization proposal.
	 * @return a set of tables
	 */
	public SortedSet<Table> getCreatedTables() {
		return createdTables;
	}

	/**
	 * Gets the foreign keys (FKs) to be created due to the normalization
	 * proposal.
	 * @return a {@link NormalizationFKsProposal} object containing the
	 * FKs to be created
	 */
	public NormalizationFKsProposal getFksProposal() {
		return fksProposal;
	}

	/**
	 * Sets the foreign keys (FKs) to be created due to the normalization
	 * proposal.
	 * @param fksProposal a {@link NormalizationFKsProposal} object containing the
	 * FKs to be created
	 */
	public void setFksProposal(NormalizationFKsProposal fksProposal) {
		this.fksProposal = fksProposal;
	}

	/**
	 * Gets normal form of the normalization proposal (the lowest
	 * normal form of the created tables).
	 * @return the normal form; may be <code>null</code>, if there have not
	 * yet any tables been added to the normalization proposal
	 */
	public NormalForm getNf() {
		NormalForm nf = null;
		if (this.createdTables != null) {
			for (Table table : this.createdTables) {
				NFDeterminationResult result = table.getNfDeterminationResult();
				if (result == null) {
					logger.error("For the created table '" + table + "' the normal " +
							"form has not yet been determined.");
				} else {
					NormalForm nfOfTable = result.getNf();
					if (nf == null || nfOfTable.compareTo(nf) < 0) {
						nf = nfOfTable;
					}
				}
			}
		}
		return nf;
	}

	/**
	 * Gets the SQL script generated for normalization.
	 * @return the sql script; <code>null</code>, if the old table does not exist in a DB.
	 */
	public String getSqlScript() {
		return sqlScript;
	}

	/**
	 * Sets the SQL script generated for normalization.
	 * @param sqlScript the sql script
	 * @see NormalizationScriptBuilder
	 */
	public void setSqlScript(String sqlScript) {
		this.sqlScript = sqlScript;
	}
}
