package de.fesenmeyer.dbnormalizer.core.algorithms;

import java.sql.DatabaseMetaData;
import java.util.Collection;
import java.util.List;
import java.util.SortedSet;

import de.fesenmeyer.dbnormalizer.core.AttributeStringSet;
import de.fesenmeyer.dbnormalizer.core.CandKey;
import de.fesenmeyer.dbnormalizer.core.FK;
import de.fesenmeyer.dbnormalizer.core.PK;
import de.fesenmeyer.dbnormalizer.core.Table;
import de.fesenmeyer.dbnormalizer.core.TableColumn;
import de.fesenmeyer.dbnormalizer.core.util.MiscUtil;

/**
 * <p>
 * This class creates a SQL script which can be used to apply
 * the normalization proposed by a {@link NormalizationProposal} object
 * to a DB.
 * </p>
 * <p>
 * <b>NOTE: </b>The current implementation has some limitations:
 * <ul>
 * 	<li>It is ORACLE-specific.</li>
 * 	<li>It does not take into account multiple candidate keys. It just creates
 * 		one key, the PK, for each table.</li>
 * 	<li>Check-Constraints are also ignored.</li>
 *  <li>The foreign keys which are defined depend on which primary key has been
 *  	selected for each table in the method {@link FDAlgorithms#setKeysForTable(Table, Table)}.
 *  	Hence, in some cases, no foreign keys will be defined at all.</li>
 * </ul>
 * </p>
 * <p>
 * <b>Because of this limitations, first check the generated SQL script before
 * applying it to the DB.</b>
 * </p>
 * @author DF
 *
 */
public class NormalizationScriptBuilder {

	/**
	 * The normalization proposal for which the sql script
	 * should be generated.
	 */
	private NormalizationProposal normProposal = null;

	/**
	 * Class constructor specifying the normalization proposal for which the sql script
	 * should be generated.
	 * @param normProposal a normalization proposal
	 */
	public NormalizationScriptBuilder(NormalizationProposal normProposal) {
		this.normProposal = normProposal;
	}

	/**
	 * Creates a SQL script based on the normalization proposal.
	 * @return the SQL script as a string
	 * @see NormalizationScriptBuilder
	 */
	public String createScript() {
		StringBuffer strBuf = new StringBuffer();

		SortedSet<Table> createdTables = normProposal.getCreatedTables();
		Table oldTable = normProposal.getOldTable();

		strBuf.append("-- create new tables (without FK-Constraints)\n");
		StringBuffer createTablesScript = buildCreateTablesScript(createdTables);
		strBuf.append(createTablesScript);
		strBuf.append("\n\n");

		strBuf.append("-- data migration (copy table contents)\n");
		StringBuffer dmScript = buildDataMigrationScript(createdTables, oldTable);
		strBuf.append(dmScript);
		strBuf.append("\n\n");

		strBuf.append("-- drop old Table (with constraints)\n");
		StringBuffer dropOldTableScript = buildDropOldTableScript(oldTable);
		strBuf.append(dropOldTableScript);
		strBuf.append("\n\n");

		NormalizationFKsProposal fksProposal = normProposal.getFksProposal();
		Collection<FK> innerFKs = fksProposal.getInnerFKS();
		if (innerFKs.size() > 0) {
			strBuf.append("-- create fks between new tables\n");
			strBuf.append(buildAddFKConstraints(innerFKs));
			strBuf.append("\n");
		}
		Collection<FK> importedFKs = fksProposal.getImportedFKs();
		if (importedFKs.size() > 0) {
			strBuf.append("-- create fks due to the old table's imported keys\n");
			strBuf.append(buildAddFKConstraints(importedFKs));
			strBuf.append("\n");
		}
		Collection<FK> exportedFKs = fksProposal.getExportedFKs();
		if (exportedFKs.size() > 0) {
			strBuf.append("-- create fks due to the old table's exported keys\n");
			strBuf.append(buildAddFKConstraints(exportedFKs));
			strBuf.append("\n");
		}
		Collection<FK> hierarchicalFKs = fksProposal.getHierarchicalFKs();
		if (hierarchicalFKs.size() > 0) {
			strBuf.append("-- create fks due to fks between the old table's attributes\n");
			strBuf.append(buildAddFKConstraints(hierarchicalFKs));
			strBuf.append("\n");
		}
		strBuf.append("\n");

		strBuf.append("-- create view as substitute for the old table\n");
		StringBuffer createViewScript = buildCreateView(createdTables, oldTable);
		strBuf.append(createViewScript);
		strBuf.append("\n");

		return strBuf.toString();
	}

	/**
	 * Builds SQL statements to create the tables (without foreign keys)
	 * given by the createdTables argument.
	 * @param createdTables the tables for which CREATE TABLE-statements should
	 * be build
	 * @return a StringBuffer containing SQL statements
	 */
	private StringBuffer buildCreateTablesScript(Collection<Table> createdTables) {
		StringBuffer strBuf = new StringBuffer();
		int i = 0;
		for (Table table : createdTables) {
			StringBuffer tableScript = buildCreateTableScript(table);
			strBuf.append(tableScript);
			if (i != (createdTables.size() - 1)) {
				strBuf.append("\n\n");
			}
			i++;
		}

		return strBuf;
	}

	/**
	 * Builds a SQL statement to create the table (without foreign keys)
	 * given by the table argument
	 * @param table the table for which a CREATE TABLE-statement should
	 * be build
	 * @return a StringBuffer containing the SQL statement
	 */
	private StringBuffer buildCreateTableScript(Table table) {
		StringBuffer strBuf = new StringBuffer();

		strBuf.append("CREATE TABLE ");
		strBuf.append(table.getQualifiedName());
		strBuf.append(" (");

		int i = 0;
		AttributeStringSet attributes = table.getAttributes();
		for (String attrName : attributes) {
			strBuf.append("\n\t");
			TableColumn col = table.getColumnInfo(attrName);
			StringBuffer tableColDef = buildTableColumnDef(table, col);
			strBuf.append(tableColDef);
			if (i < (attributes.size() - 1)) {
				strBuf.append(",");
			}
			i++;
		}
		PK pk = table.getPk();
		// if size == 1, the pk has already been defined for that column
		if (pk != null && pk.getAttributes().size() > 1) {
			strBuf.append(",\n\t");
			StringBuffer pkStrBuf = buildPrimaryKeyDef(pk);
			strBuf.append(pkStrBuf);
		}
		
		Collection<CandKey> candKeys = table.getCandKeys();
		for (CandKey candKey : candKeys) {
			//if the candidate key has not already been defined as PK
			if (!(candKey.getAttributes().containsAll(pk.getAttributes())
					&& candKey.getAttributes().size() == pk.getAttributes().size())) {
				strBuf.append(",\n\t");
				StringBuffer uniqueStrBuf = buildUniqueConstraintDef(candKey.getAttributes());
				strBuf.append(uniqueStrBuf);
			}
		}
		
		strBuf.append("\n);");
		return strBuf;
	}

	/**
	 * Builds a primary key (PK) definition for the PK
	 * specified by argument pk
	 * @param pk a PK
	 * @return a StringBuffer containing the PK definition
	 */
	private StringBuffer buildPrimaryKeyDef(PK pk) {
		StringBuffer strBuf = new StringBuffer();
		strBuf.append("PRIMARY KEY (");
		strBuf.append(MiscUtil.getCollectionAsStringBuffer(pk.getAttributes()));
		strBuf.append(")");
		return strBuf;
	}
	
	/**
	 * Builds a unique constraint definition for the
	 * specified attributes
	 * @param attributes a collection containing the attribute names
	 * @return a StringBuffer containing the unique constraint definition
	 */
	private StringBuffer buildUniqueConstraintDef(Collection<String> attributes) {
		StringBuffer strBuf = new StringBuffer();
		strBuf.append("UNIQUE (");
		strBuf.append(MiscUtil.getCollectionAsStringBuffer(attributes));
		strBuf.append(")");
		return strBuf;
	}

	/**
	 * Builds a table column definition for the column
	 * given by argument col
	 * @param table the table which contains the column
	 * @param col the column
	 * @return a StringBuffer containing the table column definition
	 */
	private StringBuffer buildTableColumnDef(Table table, TableColumn col) {
		StringBuffer strBuf = new StringBuffer();

		int decimalDigits = col.getDecimalDigits();
		String columnName = col.getName();
		int columnSize = col.getColumnSize();

		String typeName = col.getTypeName();
		String strLowerTypeName = typeName.toLowerCase();

		String strDecimalDigits = "";
		if (decimalDigits != 0) {
			strDecimalDigits = "," + decimalDigits;
		}

		// add the column name to the string
		strBuf.append(columnName);
		strBuf.append(" ");
		// add the type name to the string
		strBuf.append(typeName);

		// if the type needs to be constrained by column size (and decimal digits)
		if (strLowerTypeName.indexOf("char") != -1 || strLowerTypeName.equals("numeric")
				|| strLowerTypeName.equals("number") || strLowerTypeName.equals("decimal")) {
			strBuf.append("(");
			strBuf.append(columnSize);
			strBuf.append(strDecimalDigits);
			strBuf.append(")");
		}

		boolean declaredAsPK = false;
		PK pk = table.getPk();
		if (pk != null) {
			List<String> pkAttrs = pk.getAttributes();
			// if the column is the whole PK of the table
			if (pkAttrs.size() == 1 && pkAttrs.contains(columnName)) {
				strBuf.append(" PRIMARY KEY");
				declaredAsPK = true;
			}
		}

		if (col.isNullable() == false) {
			// if column has been declared as pk, it is implicitly "NOT NULL"
			if (!declaredAsPK) {
				strBuf.append(" NOT NULL");
			}
		}

		return strBuf;
	}

	/**
	 * Builds a data migration script which copies the data from the old table
	 * to the new tables.
	 * @param createdTables the created (new) tables
	 * @param oldTable the old table
	 * @return a StringBuffer containing the data migration script
	 */
	private StringBuffer buildDataMigrationScript(Collection<Table> createdTables, Table oldTable) {
		StringBuffer strBuf = new StringBuffer();
		int i = 0;
		for (Table table : createdTables) {
			StringBuffer tableDMScript = buildDataMigrationScript(table, oldTable);
			strBuf.append(tableDMScript);
			if (i != (createdTables.size() - 1)) {
				strBuf.append("\n");
			}
			i++;
		}
		return strBuf;
	}

	/**
	 * Builds a data migration script which copies the appropriate data from the old table
	 * to one of the new (created) tables.
	 * @param createdTable one of the new (created) tables
	 * @param oldTable the old table
	 * @return a StringBuffer containing the data migration script
	 */
	private StringBuffer buildDataMigrationScript(Table createdTable, Table oldTable) {
		StringBuffer strBuf = new StringBuffer();
		String attributesString = MiscUtil.getCollectionAsString(createdTable.getAttributes());

		strBuf.append("INSERT INTO ");
		strBuf.append(createdTable.getQualifiedName());
		strBuf.append(" (");
		strBuf.append(attributesString);
		strBuf.append(" )\n");
		strBuf.append("\tSELECT DISTINCT ");
		strBuf.append(attributesString);
		strBuf.append("\n\tFROM ");
		strBuf.append(oldTable.getQualifiedName());
		strBuf.append(";");

		return strBuf;
	}

	/**
	 * Builds a statement which drops the old table
	 * with the <code>CASCADE CONSTRAINTS</code> option.
	 * @param oldTable the old table
	 * @return a StringBuffer containing the <code>DROP TABLE</code>-statement
	 */
	private StringBuffer buildDropOldTableScript(Table oldTable) {
		StringBuffer strBuf = new StringBuffer();
		strBuf.append("DROP TABLE ");
		strBuf.append(oldTable.getQualifiedName());
		strBuf.append(" CASCADE CONSTRAINTS");
		strBuf.append(";");

		return strBuf;
	}

	/**
	 * Builds SQL statements which create the foreign key (FK) constraints
	 * specified by the fks argument
	 * @param fks the FKs to be defined
	 * @return a StringBuffer containing the SQL statements
	 */
	private StringBuffer buildAddFKConstraints(Collection<FK> fks) {
		StringBuffer strBuf = new StringBuffer();
		int i = 0;
		for (FK fk : fks) {
			StringBuffer strAddFKConstraint = buildAddFKConstraint(fk);
			strBuf.append(strAddFKConstraint);
			if (i != (fks.size() - 1)) {
				strBuf.append("\n");
			}
			i++;
		}
		return strBuf;
	}

	/**
	 * Builds a SQL statement which creates the foreign key (FK) constraint
	 * specified by the fk argument
	 * @param fk the FK to be defined
	 * @return a StringBuffer containing the SQL statement
	 */
	private StringBuffer buildAddFKConstraint(FK fk) {
		StringBuffer strBuf = new StringBuffer();

		String fkName = fk.getName();
		String fkTableName = MiscUtil.getQualifiedTableName(fk.getFkTableName(), fk.getFkCatalog(),
				fk.getFkSchema());
		String pkTableName = MiscUtil.getQualifiedTableName(fk.getPkTableName(), fk.getPkCatalog(),
				fk.getPkSchema());
		String strFKCols = MiscUtil.getCollectionAsString(fk.getFkColumnNames());
		String strPKCols = MiscUtil.getCollectionAsString(fk.getPkColumnNames());

		strBuf.append("ALTER TABLE ");
		strBuf.append(fkTableName);
		strBuf.append("\n");
		strBuf.append("\tADD");
		if (fkName != null) {
			strBuf.append(" CONSTRAINT ");
			strBuf.append(fkName);
		}
		strBuf.append(" FOREIGN KEY (");
		strBuf.append(strFKCols);
		strBuf.append(") REFERENCES ");
		strBuf.append(pkTableName);
		strBuf.append(" (");
		strBuf.append(strPKCols);
		strBuf.append(")");

		// add the delete rule
		switch (fk.getDeleteRule()) {
		case DatabaseMetaData.importedKeyCascade:
			strBuf.append(" ON DELETE CASCADE");
			break;
		case DatabaseMetaData.importedKeySetNull:
			strBuf.append(" ON DELETE SET NULL");
			break;
		case DatabaseMetaData.importedKeySetDefault:
			strBuf.append(" ON DELETE SET DEFAULT");
			break;
		case DatabaseMetaData.importedKeyRestrict:
		case DatabaseMetaData.importedKeyNoAction:
		default:
			// Oracle does neither support the "ON DELETE RESTRICT"
			// nor "ON DELETE NO ACTION" syntax --> if no rule is specified,
			// "ON DELETE NO ACTION" is used as the default
			break;
		}

		// add the update rule
		switch (fk.getUpdateRule()) {
		case DatabaseMetaData.importedKeyCascade:
			strBuf.append(" ON UPDATE CASCADE");
			break;
		case DatabaseMetaData.importedKeySetNull:
			strBuf.append(" ON UPDATE SET NULL");
			break;
		case DatabaseMetaData.importedKeySetDefault:
			strBuf.append(" ON UPDATE SET DEFAULT");
			break;
		case DatabaseMetaData.importedKeyRestrict:
		case DatabaseMetaData.importedKeyNoAction:
		default:
			// Oracle does neither support the "ON UPDATE RESTRICT"
			// nor "ON UPDATE NO ACTION" --> if no rule is specified,
			// "ON UPDATE NO ACTION" is used as the default
			break;
		}

		strBuf.append(";");

		return strBuf;
	}

	/**
	 * Builds a <code>CREATE VIEW</code> statement for creating a view
	 * which constructs the old table's data from the new (created) tables.
	 * For this purpose, the <code>NATURAL JOIN</code> operator has been used.
	 * @param createdTables the new (created) tables
	 * @param oldTable the old table
	 * @return a StringBuffer containing the <code>CREATE VIEW</code> statement
	 */
	private StringBuffer buildCreateView(Collection<Table> createdTables, Table oldTable) {
		StringBuffer strBuf = new StringBuffer();

		strBuf.append("CREATE VIEW ");
		strBuf.append(oldTable.getQualifiedName());
		strBuf.append(" AS\n");
		strBuf.append("\tSELECT * FROM ");

		StringBuffer strBufJoin = new StringBuffer();
		int i = 0;
		for (Table table : createdTables) {
			if (i != 0) {
				strBufJoin.append(" NATURAL JOIN ");
			}
			strBufJoin.append(table.getQualifiedName());
			i++;
		}

		strBuf.append(strBufJoin);
		strBuf.append(";");

		return strBuf;
	}

}
