package de.fesenmeyer.dbnormalizer.core.algorithms;

import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import de.fesenmeyer.dbnormalizer.core.AttributeStringSet;
import de.fesenmeyer.dbnormalizer.core.CandKey;
import de.fesenmeyer.dbnormalizer.core.DBNormalizerException;
import de.fesenmeyer.dbnormalizer.core.FD;
import de.fesenmeyer.dbnormalizer.core.Table;
import de.fesenmeyer.dbnormalizer.core.TableColumn;
import de.fesenmeyer.dbnormalizer.core.TableContainer;
import de.fesenmeyer.dbnormalizer.core.log.DBNormalizerLogger;
import de.fesenmeyer.dbnormalizer.core.util.MiscUtil;
import de.fesenmeyer.dbnormalizer.core.util.SetOperationsUtil;

/**
 * Implementation of the synthesis normalization algorithm described in
 * [Datenbanksysteme, Kemper/Eickler, 5th Ed., 2004, p. 181].
 * All tables of the generated decomposition are at least in 3NF. The algorithm has the
 * lossless join property and is dependency-preserving. Unfortunately, in some cases
 * the decomposition can create global redundancies.
 * 
 * @author DF
 *
 */
public class ThreeNFSynthesisAlgorithmKemper extends AbstractNormalizationAlgorithm {

	/**
	 * The logger for this class.
	 */
	private static DBNormalizerLogger logger = DBNormalizerLogger
			.getLogger(ThreeNFSynthesisAlgorithmKemper.class);

	/**
	 * The actual normalization algorithm based on 
	 * [Datenbanksysteme, Kemper/Eickler, 5th Ed., 2004, p. 181].
	 * 
	 * @param oldTable
	 *            the old table, the table to be normalized
	 * @return the tables in which the table has been split
	 * @throws DBNormalizerException if an exception occurs
	 * @see AbstractNormalizationAlgorithm#normalize(Table)
	 */
	protected NormalizationProposal doNormalize(Table oldTable) throws DBNormalizerException {
		NormalizationProposal normProposal = new NormalizationProposal(oldTable);

		// we use a list for creating the new tables, not a set, because we do not have names
		// for the new tables and therefore cannot compare them by their name
		List<Table> tables = new LinkedList<Table>();

		// 1.) calculate grouped minimal cover (minimal cover grouped by equal lhs)

		// minimal cover must have already been computed
		Set<FD> minCover = oldTable.getFDMinCover();
		// calculate a grouped minimal cover from the minimal cover
		Set<FD> groupedMinCover = FDAlgorithms.mergeFDs(minCover);

		// create an Array of the grouped minimal cover (for easier iteration)
		FD[] canonicalCoverageArray = new FD[groupedMinCover.size()];
		groupedMinCover.toArray(canonicalCoverageArray);

		// variable which holds the candidate keys of the old table as AttributeStringSets
		SortedSet<AttributeStringSet> candidateKeys = new TreeSet<AttributeStringSet>();

		// if candidate keys have not yet been computed, throw an exception
		if (oldTable.getCandKeys() == null || oldTable.getCandKeys().size() == 0) {
			throw new DBNormalizerException(
					"The candidate keys for the table '"
							+ oldTable.getQualifiedName()
							+ "' have not yet been computed. Without the candidate keys, no normalization proposal can be given.");
		}

		// convert the already computed candidate keys to AttributeStringSets and
		// add it to the candidateKeys variable
		for (CandKey candKey : oldTable.getCandKeys()) {
			candidateKeys.add(new AttributeStringSet(candKey.getAttributes()));
		}

		// if a primary key has not yet been defined, throw an exception
		if (oldTable.getPk() == null || oldTable.getPk().getAttributes().size() == 0) {
			throw new DBNormalizerException("No PK has been defined for the table '"
					+ oldTable.getQualifiedName()
					+ "'. Without PK, no normalization proposal can be given.");
		}

		// variable which holds the old table's primary key attributes
		AttributeStringSet pkAttributes = new AttributeStringSet(oldTable.getPk().getAttributes());

		// variable candidateKeyInCreatedRelations is needed for 3.)
		boolean candidateKeyInCreatedRelations = false;

		// 2. for each fd ..
		for (int i = 0; i < canonicalCoverageArray.length; i++) {
			FD fd = canonicalCoverageArray[i];

			// 2.1 create a new table
			Table newTable = new Table();
			//Here a temporary table name is set. This name
			//may be changed later, if some tables are removed in Step 4
			String tempTableName = oldTable.getName() + "_" + i;
			newTable.setName(tempTableName);
			newTable.setContainer(oldTable.getContainer());
			newTable.setSchema(oldTable.getSchema());
			newTable.setCatalog(oldTable.getCatalog());

			AttributeStringSet newTableAttributes = newTable.getAttributes();
			newTableAttributes.addAll(fd.getLHS());
			newTableAttributes.addAll(fd.getRHS());

			if (candidateKeyInCreatedRelations == false) {
				for (AttributeStringSet candidate : candidateKeys) {
					if (SetOperationsUtil.isSubset(candidate, newTableAttributes)) {
						candidateKeyInCreatedRelations = true;
						break;
					}
				}
			}

			// 2.2 add all fds between the new table's attributes
			for (int j = 0; j < canonicalCoverageArray.length; j++) {
				FD fd1 = canonicalCoverageArray[j];

				if (SetOperationsUtil.isSubset(fd1.getLHS(), newTableAttributes)
						&& SetOperationsUtil.isSubset(fd1.getRHS(), newTableAttributes)) {
					FD fdCopy = fd1.copy();
					newTable.getFDs().add(fdCopy);
				}
			}

			tables.add(newTable);
		}

		// 3. if the generated relations contain no candidate key,
		// build a relation containing just one key (the given primary key)
		if (!candidateKeyInCreatedRelations) {
			Table newTable = new Table();
			//Here a temporary table name is set. This name
			//may be changed later, if some tables are removed in Step 4
			String tempTableName = oldTable.getName() + "_" + (tables.size() + 1);
			newTable.setName(tempTableName);
			TableContainer container = MiscUtil.createTableContainer(oldTable.getCatalog(),
					oldTable.getSchema());
			newTable.setContainer(container);
			newTable.setSchema(oldTable.getSchema());
			newTable.setCatalog(oldTable.getCatalog());

			newTable.getAttributes().addAll(pkAttributes);
			tables.add(newTable);
		}

		// 4. Delete schemas contained in other schemas

		// temporary array of tables
		Table[] tableArr = new Table[tables.size()];
		tables.toArray(tableArr);

		for (int i = 0; i < tableArr.length; i++) {
			Table tableToCheck = tableArr[i];
			if (tableToCheck != null) {
				for (int j = 0; j < tableArr.length; j++) {
					Table tableToCompare = tableArr[j];
					if (i != j && tableToCompare != null) {
						AttributeStringSet attrsToCheck = tableToCheck.getAttributes();
						AttributeStringSet attrsToCompare = tableToCompare.getAttributes();
						if (attrsToCompare.containsAll(attrsToCheck)) {
							tableArr[i] = null;
							tableToCompare.getFDs().addAll(tableToCheck.getFDs());
							tables.remove(tableToCheck);
						}
					}
				}
			}
		}

		//for each created table...
		for (int i = 0; i < tables.size(); i++) {
			Table table = tables.get(i);
			//set a new name for the table
			String newTableName = oldTable.getName() + "_" + (i + 1);
			table.setName(newTableName);
			
			//add the table to the normalization proposal
			normProposal.addCreatedTable(table);
		}

		return normProposal;
	}

	/**
	 * <p>
	 * Method called after the actual normalization algorithm, to
	 * do necessary work not included in the normalization algorithm,
	 * such as calculating minimal cover and candidate keys for the
	 * new (created) tables, determining foreign keys (FKs) etc.
	 * </p>
	 * <p>
	 * Extends the normalization proposal by adding further information.
	 * </p>
	 * @param normProposal the normalization proposal to be extended
	 * @return the extend normalization proposal
	 * @throws DBNormalizerException
	 */
	protected NormalizationProposal postNormalize(NormalizationProposal normProposal)
			throws DBNormalizerException {
		// for each new table (table of the normalization proposal)
		for (Table table : normProposal.getCreatedTables()) {

			// compute and set the minimal cover
			Set<FD> newTblMinCover = FDAlgorithms.calculateMinCover(table.getFDs());
			table.setFDMinCover(newTblMinCover);

			// compute and set candidate keys and a primary key
			FDAlgorithms.setKeysForTable(table, normProposal.getOldTable());

			// determine the NF of the new tables
			NFDetermination nfDetermination = new NFDetermination(table);
			NFDeterminationResult result = nfDetermination.determine();
			table.setNfDeterminationResult(result);
		}

		// copy the attribute information from the old table to the new tables
		copyTableColumnInformation(normProposal);

		// determine which FKs should be created for and to the new tables
		NormalizationFKsDetermination fkDetermination = new NormalizationFKsDetermination(
				normProposal);
		NormalizationFKsProposal fksProposal = fkDetermination
				.getNormalizationFKsProposal();
		normProposal.setFksProposal(fksProposal);

		// create normalization script, if the original table is
		// a DB-Table
		if (normProposal.getOldTable().isDbTable()) {
			NormalizationScriptBuilder scriptBuilder = new NormalizationScriptBuilder(normProposal);
			String sqlScript = scriptBuilder.createScript();
			normProposal.setSqlScript(sqlScript);
		}

		return normProposal;
	}
	
	/**
	 * Copies the table column information from
	 * the old table to the new (created) tables
	 * of the given normalization proposal.
	 * @param normProposal the normalization proposal
	 */
	protected void copyTableColumnInformation(NormalizationProposal normProposal) {
		// tableColumn information can only be copied if it exists
		// (it exists only if the original table is a DB-Table)
		if (normProposal.getOldTable().isDbTable()) {
			for (Table table : normProposal.getCreatedTables()) {
				for (String attrName : table.getAttributes()) {
					TableColumn oldCol = normProposal.getOldTable().getColumnInfo(attrName);
					TableColumn newCol = oldCol.copy();
					table.addColumnInfo(attrName, newCol);
				}
			}
		}
	}

	/* (non-Javadoc)
	 * @see de.fesenmeyer.dbnormalizer.core.algorithms.AbstractNormalizationAlgorithm#normalize(de.fesenmeyer.dbnormalizer.core.Table)
	 */
	public NormalizationProposal normalize(Table oldTable) throws DBNormalizerException {
		//calls the actual implementation of the Kemper/Eickler normalization algorithm
		NormalizationProposal normProposal = doNormalize(oldTable);
		
		//do some work which is not included in the Kemper/Eickler algorithm
		postNormalize(normProposal);
		
		return normProposal;
	}
}