package de.fesenmeyer.dbnormalizer.core.configuration;

import java.io.File;
import java.net.URI;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.TreeSet;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.TransformerException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import de.fesenmeyer.dbnormalizer.core.DBConfig;
import de.fesenmeyer.dbnormalizer.core.TableContainer;
import de.fesenmeyer.dbnormalizer.core.TableContainerType;
import de.fesenmeyer.dbnormalizer.core.configuration.xml.DBNormalizerXMLPathUtil;
import de.fesenmeyer.dbnormalizer.core.configuration.xml.DOMTransformer;
import de.fesenmeyer.dbnormalizer.core.configuration.xml.DOMUtil;
import de.fesenmeyer.dbnormalizer.core.configuration.xml.DOMValidator;
import de.fesenmeyer.dbnormalizer.core.configuration.xml.SimpleErrorHandler;
import de.fesenmeyer.dbnormalizer.core.configuration.xml.XMLConstants;
import de.fesenmeyer.dbnormalizer.core.log.DBNormalizerLogger;
import de.fesenmeyer.dbnormalizer.core.util.MiscUtil;

/**
 * Class which reads a session configuration from a XML file. Such a file contains FDs grouped by schema, catalog and table.
 * Either a DB-Configuration for reading table information from a DB or attributes for each table are also contained in the file.
 * A standard DB-Normalizer-file can be read with the {@link #read()} method. To import FDs from Tobias Rafreiders tool "Tane-java",
 * which uses a slightly different format, you can use the method {@link #readTaneJavaConfig()}.
 * @author DF
 * @see ConfigurationWriter
 * @see SessionConfig
 */
public class ConfigurationReader {
	/**
	 * The logger for this class.
	 */
	private static DBNormalizerLogger logger = DBNormalizerLogger.getLogger(ConfigurationReader.class);

	/**
	 * The name (and path) of the file to be read.
	 */
	private String fileName = null;

	/**
	 * Flag which states whether the file contains a DB-Configuration or not.
	 */
	private boolean hasDBConfig = false;

	/**
	 * Class constructor specifying the file to be read.
	 * @param fileName the name (and) path of the file to be read
	 */
	public ConfigurationReader(String fileName) {
		// the following code is a workaround to the
		// java sun bug 6506304 (http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6506304)
		File f = new File(fileName);
		URI uri = f.toURI();
		this.fileName = uri.toString();
	}

	/**
	 * Reads the configuration file into a session configuration object. Before creating the session
	 * configuration object, the file is validated against the XML-Schema <em>xsd/configuration.xsd</em>.
	 * @return a session configuration object
	 * @throws ConfigurationException if an exception occurs while reading
	 * the file, e.g. if it is not valid
	 */
	public SessionConfig read() throws ConfigurationException {
		SessionConfig sessionConfig = null;

		DocumentBuilderFactory docFactory = DocumentBuilderFactory.newInstance();
		docFactory.setNamespaceAware(true);
		docFactory.setValidating(true);
		docFactory.setAttribute(XMLConstants.JAXP_SCHEMA_LANGUAGE, XMLConstants.W3C_XML_SCHEMA);
		docFactory.setAttribute(XMLConstants.JAXP_SCHEMA_SOURCE,
				DBNormalizerXMLPathUtil.getDBNormalizerSchemaFilename());
		DocumentBuilder docBuilder;
		try {
			docBuilder = docFactory.newDocumentBuilder();
			docBuilder.setErrorHandler(new SimpleErrorHandler());
			Document doc = docBuilder.parse(fileName);
			Element configNode = doc.getDocumentElement();
			sessionConfig = retrieveSessionConfig(configNode);
		} catch (ConfigurationException e) {
			throw e;
		} catch (Exception e) {
			throw new ConfigurationException(e);
		}

		return sessionConfig;
	}
	
	/**
	 * Can be used to import FDs from Tobias Rafreider's tool "TANE-java". First performs
	 * an XSL-Transformation on the file (by using the XSL-file <em>xsl/transform_tanejava2dbn.xsl</em>),
	 * then validates the result against the XML-Schema <em>xsd/configuration.xsd</em> and
	 * creates a session configuration object.
	 * @return a session configuration object
	 * @throws ConfigurationException if an exception occurs during transformation or validation
	 * of the file
	 */
	public SessionConfig readTaneJavaConfig() throws ConfigurationException {
		SessionConfig sessionConfig = null;

		Document taneDoc = null;
		
		DocumentBuilderFactory docFactory = DocumentBuilderFactory.newInstance();
		docFactory.setNamespaceAware(true);
		docFactory.setValidating(false);
		try {
			DocumentBuilder docBuilder = docFactory.newDocumentBuilder();
			docBuilder = docFactory.newDocumentBuilder();
			docBuilder.setErrorHandler(new SimpleErrorHandler());
			taneDoc = docBuilder.parse(fileName);
		} catch (Exception e) {
			throw new ConfigurationException(e);
		}
		
		Document dbNormDoc = null;
		
		String xsltFile = DBNormalizerXMLPathUtil.getTaneJava2DBNormalizerXSLFilename();
		DOMTransformer domTransformer = new DOMTransformer(taneDoc, xsltFile);
		try {
			dbNormDoc = domTransformer.transform();
		} catch (TransformerException e) {
			throw new ConfigurationException(e);
		}
		
		String schemaFile = DBNormalizerXMLPathUtil.getDBNormalizerSchemaFilename();
		DOMValidator domValidator = new DOMValidator(dbNormDoc, schemaFile);
		try {
			domValidator.validate();
		} catch (Exception e) {
			throw new ConfigurationException(e);
		}
		
		Element configNode = dbNormDoc.getDocumentElement();
		sessionConfig = retrieveSessionConfig(configNode);
		return sessionConfig;
	}

	/**
	 * Reads the whole file into a session configuration object.
	 * @param configNode the root node of the file
	 * @return the session configuration object
	 * @throws ConfigurationException
	 */
	private SessionConfig retrieveSessionConfig(Element configNode) throws ConfigurationException {
		SessionConfig sessionConfig = new SessionConfig(this.fileName);

		Element dbConfigElement = DOMUtil.getFirstElement(configNode, "dbConfig");

		if (dbConfigElement != null) {
			this.hasDBConfig = true;
			DBConfig dbConfig = retrieveDBConfig(dbConfigElement);
			sessionConfig.setDbConfig(dbConfig);
		}

		Collection<TableContainerConfig> tableContainerConfigs = new ArrayList<TableContainerConfig>();

		Collection<TableContainerConfig> schemaConfigs = retrieveSchemaConfigs(configNode);
		tableContainerConfigs.addAll(schemaConfigs);

		Collection<TableContainerConfig> catalogConfigs = retrieveCatalogConfigs(configNode);
		tableContainerConfigs.addAll(catalogConfigs);

		Element tableInfoNode = DOMUtil.getFirstChildElement(configNode, "tableInfo");
		if (tableInfoNode != null) {
			Collection<TableConfig> tableConfigs = retrieveTableConfigs(tableInfoNode, null, null);
			if (tableConfigs.size() > 0) {
				TableContainer noContainer = TableContainer.NO_CONTAINER;
				TableContainerConfig noContainerConfig = new TableContainerConfig(noContainer);
				noContainerConfig.setTableConfigs(tableConfigs);

				tableContainerConfigs.add(noContainerConfig);
			}
		}

		sessionConfig.setTableContainerConfigs(tableContainerConfigs);

		return sessionConfig;
	}

	/**
	 * Reads the <code>dbConfig</code> node into a DB-Configuration object.
	 * @param dbConfigElement the <code>dbConfig</code> node
	 * @return a DB-Configuration object
	 */
	private DBConfig retrieveDBConfig(Element dbConfigElement) {
		String dbUrl = DOMUtil.getSimpleElementText(dbConfigElement, "dbUrl");
		// username and password are optional, that means they can be null
		String username = DOMUtil.getSimpleElementText(dbConfigElement, "username");
		String password = DOMUtil.getSimpleElementText(dbConfigElement, "password");

		String driverClassName = DOMUtil.getSimpleElementText(dbConfigElement, "driverClassName");
		String driverFileName = DOMUtil.getSimpleElementText(dbConfigElement, "driverFileName");

		DBConfig dbConfig = new DBConfig(dbUrl, username, password, driverClassName, driverFileName);
		return dbConfig;
	}

	/**
	 * Reads the data from a <code>schema</code> node.
	 * @param parentNode the parent node of the <code>schema</code> node
	 * @return a collection of {@link TableContainerConfig} objects
	 * @throws ConfigurationException
	 */
	private Collection<TableContainerConfig> retrieveSchemaConfigs(Element parentNode)
			throws ConfigurationException {
		LinkedHashSet<TableContainerConfig> schemaConfigs = new LinkedHashSet<TableContainerConfig>();
		NodeList schemaNodes = parentNode.getElementsByTagName("schema");
		for (int i = 0; i < schemaNodes.getLength(); i++) {
			Element schemaNode = (Element) schemaNodes.item(i);
			String name = schemaNode.getAttribute("name");

			TableContainer schema = new TableContainer(name, TableContainerType.SCHEMA);

			TableContainerConfig schemaConfig = new TableContainerConfig(schema);

			Element tableInfoNode = DOMUtil.getFirstElement(schemaNode, "tableInfo");
			if (tableInfoNode != null) {
				Collection<TableConfig> tableConfigs = retrieveTableConfigs(tableInfoNode, null,
						name);
				schemaConfig.setTableConfigs(tableConfigs);
			}

			schemaConfigs.add(schemaConfig);
		}
		return schemaConfigs;
	}

	/**
	 * Reads the data from a <code>catalog</code> node.
	 * @param parentNode the parent node of the <code>catalog</code> node
	 * @return a collection of {@link TableContainerConfig} objects
	 * @throws ConfigurationException
	 */
	private Collection<TableContainerConfig> retrieveCatalogConfigs(Element parentNode)
			throws ConfigurationException {
		LinkedHashSet<TableContainerConfig> catalogConfigs = new LinkedHashSet<TableContainerConfig>();
		NodeList catalogNodes = parentNode.getElementsByTagName("catalog");
		for (int i = 0; i < catalogNodes.getLength(); i++) {
			Element catalogNode = (Element) catalogNodes.item(i);
			String name = catalogNode.getAttribute("name");

			TableContainer catalog = new TableContainer(name, TableContainerType.SCHEMA);

			TableContainerConfig catalogConfig = new TableContainerConfig(catalog);

			Element tableInfoNode = DOMUtil.getFirstElement(catalogNode, "tableInfo");
			if (tableInfoNode != null) {
				Collection<TableConfig> tableConfigs = retrieveTableConfigs(tableInfoNode, name,
						null);
				catalogConfig.setTableConfigs(tableConfigs);
			}

			catalogConfigs.add(catalogConfig);
		}
		return catalogConfigs;
	}

	/**
	 * Reads the (tables') data from a <code>tableInfo</code> node.
	 * @param tableInfoNode the <code>tableInfo</code> node
	 * @param catalogName the name of the catalog containing the tables, may be <code>null</code>
	 * @param schemaName the name of the schema containing the tables, may be <code>null</code>
	 * @return a collection of {@link TableConfig} objects
	 * @throws ConfigurationException
	 */
	private Collection<TableConfig> retrieveTableConfigs(Element tableInfoNode, String catalogName,
			String schemaName) throws ConfigurationException {
		LinkedHashSet<TableConfig> tableConfigs = new LinkedHashSet<TableConfig>();
		NodeList tableNodes = tableInfoNode.getElementsByTagName("table");

		for (int i = 0; i < tableNodes.getLength(); i++) {
			TableConfig tableConfig = new TableConfig();

			Element tableNode = (Element) tableNodes.item(i);
			String name = tableNode.getAttribute("name");
			tableConfig.setName(name);

			Element attrsNode = DOMUtil.getFirstElement(tableNode, "attributes");

			if (attrsNode != null) {
				String qTableName = MiscUtil.getQualifiedTableName(name, catalogName, schemaName);
				if (this.hasDBConfig) {
					throw new ConfigurationException(
							"The XML document contains a 'dbConfig' element. "
									+ "Therefore it must not contain attribute information. "
									+ "Remove the 'attributes' tag for table '" + qTableName
									+ "' or remove " + "the 'dbConfig' element.");
				} else {
					AttributeSetConfig attrs = retrieveAttributes(attrsNode);
					tableConfig.setAttributeSetConfig(attrs);
				}
			}

			Element fdsNode = DOMUtil.getFirstElement(tableNode, "fds");
			if (fdsNode != null) {
				Collection<FDConfig> fdConfigs = retrieveFDConfigs(fdsNode);
				tableConfig.setFdConfigs(fdConfigs);
			}

			tableConfigs.add(tableConfig);
		}
		return tableConfigs;
	}

	/**
	 * Reads the FDs from a <code>fds</code> node.
	 * @param fdsNode the <code>fds</code> node
	 * @return a collection of {@link FDConfig} objects
	 */
	private Collection<FDConfig> retrieveFDConfigs(Element fdsNode) {
		LinkedHashSet<FDConfig> fdConfigs = new LinkedHashSet<FDConfig>();
		NodeList fdNodes = fdsNode.getElementsByTagName("fd");
		for (int i = 0; i < fdNodes.getLength(); i++) {
			Element fdNode = (Element) fdNodes.item(i);
			FDConfig fdConfig = new FDConfig();

			Element lhsNode = DOMUtil.getFirstElement(fdNode, "lhs");
			AttributeSetConfig lhs = retrieveAttributes(lhsNode);
			fdConfig.setLhs(lhs);
			Element rhsNode = DOMUtil.getFirstElement(fdNode, "rhs");
			AttributeSetConfig rhs = retrieveAttributes(rhsNode);
			fdConfig.setRhs(rhs);

			fdConfigs.add(fdConfig);
		}
		return fdConfigs;
	}

	/**
	 * Reads the (table) attributes from a node containing (table) attributes.
	 * @param attributeSetNode the node containing (table) attributes
	 * @return an {@link AttributeSetConfig} object
	 */
	private AttributeSetConfig retrieveAttributes(Element attributeSetNode) {
		TreeSet<String> attributes = new TreeSet<String>();
		NodeList attributeNodes = attributeSetNode.getElementsByTagName("attribute");
		for (int i = 0; i < attributeNodes.getLength(); i++) {
			Element attributeNode = (Element) attributeNodes.item(i);
			String attribute = attributeNode.getTextContent();
			attributes.add(attribute);
		}
		AttributeSetConfig attributeSetConfig = new AttributeSetConfig();
		attributeSetConfig.setAttributes(attributes);
		return attributeSetConfig;
	}

}
