package de.fesenmeyer.dbnormalizer.core.configuration;

import java.io.IOException;
import java.util.Collection;
import java.util.Set;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import de.fesenmeyer.dbnormalizer.core.DBConfig;
import de.fesenmeyer.dbnormalizer.core.TableContainer;
import de.fesenmeyer.dbnormalizer.core.TableContainerType;
import de.fesenmeyer.dbnormalizer.core.configuration.xml.DBNormalizerXMLPathUtil;
import de.fesenmeyer.dbnormalizer.core.configuration.xml.DOMDocumentWriter;
import de.fesenmeyer.dbnormalizer.core.configuration.xml.XMLConstants;

/**
 * Class which writes a session configuration to a XML file.
 * @author DF
 * @see ConfigurationReader
 * @see SessionConfig
 */
public class ConfigurationWriter {

	/**
	 * The session configuration to be written to file.
	 */
	private SessionConfig sessionConfig = null;

	/**
	 * The DOM-document used for creating nodes etc.
	 */
	private Document doc = null;

	/**
	 * Class constructor specifying the session configuration to be written.
	 * @param sessionConfig the session configuration object
	 */
	public ConfigurationWriter(SessionConfig sessionConfig) {
		this.sessionConfig = sessionConfig;
	}

	/**
	 * Writes the session configuration to a XML-file.
	 * @throws ConfigurationException if an exception occurs, such that
	 * the file can not be successfully written (e.g. an {@link IOException})
	 */
	public void write() throws ConfigurationException {
		DocumentBuilderFactory docFactory = DocumentBuilderFactory.newInstance();
		docFactory.setNamespaceAware(true);
		docFactory.setValidating(true);
		docFactory.setAttribute(XMLConstants.JAXP_SCHEMA_LANGUAGE, XMLConstants.W3C_XML_SCHEMA);
		docFactory.setAttribute(XMLConstants.JAXP_SCHEMA_SOURCE,
				DBNormalizerXMLPathUtil.getDBNormalizerSchemaFilename());
		DocumentBuilder docBuilder;
		try {
			docBuilder = docFactory.newDocumentBuilder();
			this.doc = docBuilder.newDocument();
			Node configNode = getConfigurationNode(this.sessionConfig);
			this.doc.appendChild(configNode);

			DOMDocumentWriter docWriter = new DOMDocumentWriter(this.doc);
			docWriter.write(sessionConfig.getConfigFileName());
		} catch (Exception e) {
			throw new ConfigurationException(e);
		}
	}

	/**
	 * Creates a <code>configuration</code> node from the session
	 * configuration object.
	 * @param sessionConfig the session configuration object
	 * @return the <code>configuration</code> node
	 */
	private Node getConfigurationNode(SessionConfig sessionConfig) {
		Element configurationNode = this.doc.createElement("configuration");

		DBConfig dbConfig = sessionConfig.getDbConfig();
		if (dbConfig != null) {
			Element dbConfigNode = getDBConfigNode(dbConfig);
			configurationNode.appendChild(dbConfigNode);
		}

		Collection<TableContainerConfig> tableContainerConfigs = sessionConfig
				.getTableContainerConfigs();
		for (TableContainerConfig tableContainerConfig : tableContainerConfigs) {
			Element tableContainerConfigNode = getTableContainerConfigNode(tableContainerConfig);
			configurationNode.appendChild(tableContainerConfigNode);
		}

		return configurationNode;
	}

	/**
	 * Creates a <code>dbConfig</code> node.
	 * @param dbConfig the DB-Config object from which the node should be created
	 * @return the <code>dbConfig</code> node
	 */
	private Element getDBConfigNode(DBConfig dbConfig) {
		Element dbConfigNode = this.doc.createElement("dbConfig");

		Element childElement = null;

		childElement = this.doc.createElement("dbUrl");
		childElement.setTextContent(dbConfig.getDbURL());
		dbConfigNode.appendChild(childElement);

		String username = dbConfig.getUsername();
		if (username != null) {
			childElement = this.doc.createElement("username");
			childElement.setTextContent(username);
			dbConfigNode.appendChild(childElement);
		}

		String password = dbConfig.getPassword();
		if (password != null) {
			childElement = this.doc.createElement("password");
			childElement.setTextContent(password);
			dbConfigNode.appendChild(childElement);
		}

		childElement = this.doc.createElement("driverClassName");
		childElement.setTextContent(dbConfig.getDriverClassName());
		dbConfigNode.appendChild(childElement);

		childElement = this.doc.createElement("driverFileName");
		childElement.setTextContent(dbConfig.getDriverFileName());
		dbConfigNode.appendChild(childElement);

		return dbConfigNode;
	}

	/**
	 * Creates a node for a TableContainerConfig object.
	 * @param tableContainerConfig the TableContainerConfig object
	 * @return an appropriate node for the TableContainerConfig object
	 */
	private Element getTableContainerConfigNode(TableContainerConfig tableContainerConfig) {
		TableContainer tableContainer = tableContainerConfig.getTableContainer();
		TableContainerType type = tableContainer.getType();

		String tableContainerNodeName = null;

		switch (type) {
		case SCHEMA:
			tableContainerNodeName = "schema";
			break;
		case CATALOG:
			tableContainerNodeName = "catalog";
			break;
		case NONE:
			tableContainerNodeName = null;
			break;
		}

		Element tableContainerNode = null;

		if (tableContainerNodeName == null) {
			tableContainerNode = getTableInfoNode(tableContainerConfig.getTableConfigs());
		} else {
			tableContainerNode = this.doc.createElement(tableContainerNodeName);
			tableContainerNode.setAttribute("name", tableContainer.getName());
			Element tableInfoNode = getTableInfoNode(tableContainerConfig.getTableConfigs());
			tableContainerNode.appendChild(tableInfoNode);
		}

		return tableContainerNode;
	}

	/**
	 * Creates a <code>tableInfo</code> node from the TableConfig objects
	 * specified by the tableConfigs argument.
	 * @param tableConfigs a collection of TableConfig objects
	 * @return the <code>tableInfo</code> node
	 */
	private Element getTableInfoNode(Collection<TableConfig> tableConfigs) {
		Element tableInfoNode = this.doc.createElement("tableInfo");
		for (TableConfig tableConfig : tableConfigs) {
			Element tableNode = getTableNode(tableConfig);
			tableInfoNode.appendChild(tableNode);
		}
		return tableInfoNode;
	}

	/**
	 * Creates a <code>table</code> node from
	 * the TableConfig object specified by argument tableConfig.
	 * @param tableConfig the tableConfig object
	 * @return the <code>table</code> node
	 */
	private Element getTableNode(TableConfig tableConfig) {
		Element tableNode = this.doc.createElement("table");
		tableNode.setAttribute("name", tableConfig.getName());

		Set<String> attributes = tableConfig.getAttributeSetConfig().getAttributes();
		if (attributes.size() > 0) {
			Element attributesNode = getAttributesNode("attributes", attributes);
			tableNode.appendChild(attributesNode);
		}

		Collection<FDConfig> fdConfigs = tableConfig.getFdConfigs();
		if (fdConfigs.size() > 0) {
			Element fdsNode = getFDsNode(fdConfigs);
			tableNode.appendChild(fdsNode);
		}
		return tableNode;
	}

	/**
	 * Creates an attributes node, which gets its name from the
	 * specified argument nodeName.
	 * @param nodeName the node's name to be set
	 * @param attributes a set of attribute names
	 * @return an attributes node as described above
	 */
	private Element getAttributesNode(String nodeName, Set<String> attributes) {
		Element attributesNode = this.doc.createElement(nodeName);
		for (String attribute : attributes) {
			Element attributeNode = this.doc.createElement("attribute");
			attributeNode.setTextContent(attribute);
			attributesNode.appendChild(attributeNode);
		}
		return attributesNode;
	}

	/**
	 * Creates a <code>fds</code> node from FDConfig objects.
	 * @param fdConfigs a collection of FDConfig objects
	 * @return the <code>fds</code> node
	 */
	private Element getFDsNode(Collection<FDConfig> fdConfigs) {
		Element fdsNode = this.doc.createElement("fds");
		for (FDConfig fdConfig : fdConfigs) {
			Element fdNode = this.doc.createElement("fd");
			Element lhsNode = getAttributesNode("lhs", fdConfig.getLhs().getAttributes());
			Element rhsNode = getAttributesNode("rhs", fdConfig.getRhs().getAttributes());
			fdNode.appendChild(lhsNode);
			fdNode.appendChild(rhsNode);
			fdsNode.appendChild(fdNode);
		}
		return fdsNode;
	}

}
