package de.fesenmeyer.dbnormalizer.core.dbaccess;

import java.io.File;
import java.sql.SQLException;
import java.util.HashMap;

import net.sourceforge.squirrel_sql.client.gui.db.SQLAlias;
import net.sourceforge.squirrel_sql.client.util.IdentifierFactory;
import net.sourceforge.squirrel_sql.fw.id.IIdentifierFactory;
import net.sourceforge.squirrel_sql.fw.sql.ISQLAlias;
import net.sourceforge.squirrel_sql.fw.sql.ISQLDriver;
import net.sourceforge.squirrel_sql.fw.sql.SQLConnection;
import net.sourceforge.squirrel_sql.fw.sql.SQLDriver;
import net.sourceforge.squirrel_sql.fw.sql.SQLDriverManager;
import de.fesenmeyer.dbnormalizer.core.DBConfig;

/**
 * Singleton class to create, get and remove SQuirreL-DB connections. Provides
 * a single access for DB-Connections and the possibility to work with more than
 * one session at once.
 * @author DF
 *
 */
public final class DBConnectionManager {

	/**
	 * HashMap holding the connections (keyed by a session name).
	 */
	private HashMap<String, SQLConnection> connections = new HashMap<String, SQLConnection>();

	/**
	 * The singleton instance.
	 */
	private static DBConnectionManager INSTANCE;

	/**
	 * Makes the constructor inaccessible.
	 */
	private DBConnectionManager() {

	}

	/**
	 * Gets the single instance of the class.
	 * @return the DBConnectionManager instance
	 */
	public static DBConnectionManager getInstance() {
		if (INSTANCE == null) {
			INSTANCE = new DBConnectionManager();
		}
		return INSTANCE;
	}

	/**
	 * Creates, adds and returns a connection for the session with
	 * the specified name and DB configuration.
	 * @param sessionName the session name
	 * @param dbConfig the DB-Configuration
	 * @return a SQuirreL-SQLConnection object
	 * @throws DBException if an Exception occurs creating the connection.
	 * Can have many reasons, e.g.:
	 * <ul>
	 * 	<li>The DB-URL is incorrect or the specified DB does not exist.</li>
	 * 	<li>The path specified for the JDBC-Driver is incorrect.</li>
	 * 	<li>The driver class can not be loaded, e.g. because the specified
	 * class name is incorrect.</li>
	 * 	<li>The specified username or password for the DB is incorrect.</li>
	 * </ul>
	 */
	public SQLConnection createAndAddConnection(String sessionName, DBConfig dbConfig)
			throws DBException {
		SQLConnection sqlConnection = null;

		File file = new File(dbConfig.getDriverFileName());
		if (!file.exists()) {
			throw new DBException("The driver file name '" + dbConfig.getDriverFileName()
					+ "' does not exist.");
		}

		IIdentifierFactory idFactory = IdentifierFactory.getInstance();
		ISQLAlias sqlAlias = new SQLAlias(idFactory.createIdentifier());
		sqlAlias.setDriverProperties(null); // SQLDriverProperties are not
		// needed
		try {

			sqlAlias.setName(sessionName);
			sqlAlias.setUrl(dbConfig.getDbURL());
			sqlAlias.setUserName(dbConfig.getUsername());
			sqlAlias.setPassword(dbConfig.getPassword());

			ISQLDriver sqlDriver = new SQLDriver(idFactory.createIdentifier());
			sqlDriver.setJarFileNames(new String[] { dbConfig.getDriverFileName() });
			sqlDriver.setDriverClassName(dbConfig.getDriverClassName());
			// we set the same as DriverClassName as name of the driver
			sqlDriver.setName(dbConfig.getDriverClassName());

			SQLDriverManager sqlDriverManager = new SQLDriverManager();
			sqlDriverManager.registerSQLDriver(sqlDriver);

			sqlConnection = sqlDriverManager.getConnection(sqlDriver, sqlAlias, dbConfig
					.getUsername(), dbConfig.getPassword());

		} catch (Exception e) {
			throw new DBException(e);
		}

		// add SQLConnection to internal hashmap
		connections.put(sessionName, sqlConnection);

		return sqlConnection;
	}

	/**
	 * Returns the connection for the session specified by argument
	 * sessionName.
	 * @param sessionName the session name
	 * @return a SQuirreL-SQLConnection, if a connection has already been added
	 * for the specified session;
	 * <code>null</code>, otherwise
	 */
	public SQLConnection getConnection(String sessionName) {
		return connections.get(sessionName);
	}

	/**
	 * Removes and closes the connection for the session specified by argument
	 * sessionName.
	 * @param sessionName the session name
	 * @throws DBException If an exception occurs when closing the connection
	 */
	public void removeConnection(String sessionName) throws DBException {
		SQLConnection connection = connections.remove(sessionName);
		if (connection != null) {
			try {
				connection.close();
			} catch (SQLException e) {
				throw new DBException(e);
			}
		}
	}
}
