package de.fesenmeyer.dbnormalizer.core.dbaccess;

import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import net.sourceforge.squirrel_sql.fw.sql.ITableInfo;
import net.sourceforge.squirrel_sql.fw.sql.PrimaryKeyInfo;
import net.sourceforge.squirrel_sql.fw.sql.SQLDatabaseMetaData;
import net.sourceforge.squirrel_sql.fw.sql.TableColumnInfo;
import de.fesenmeyer.dbnormalizer.core.DBNormalizerSession;
import de.fesenmeyer.dbnormalizer.core.FK;
import de.fesenmeyer.dbnormalizer.core.PK;
import de.fesenmeyer.dbnormalizer.core.Table;
import de.fesenmeyer.dbnormalizer.core.TableColumn;
import de.fesenmeyer.dbnormalizer.core.TableContainer;
import de.fesenmeyer.dbnormalizer.core.TableContainerType;
import de.fesenmeyer.dbnormalizer.core.log.DBNormalizerLogger;
import de.fesenmeyer.dbnormalizer.core.util.MiscUtil;

/**
 * Utility class for retrieving database information. Acts as a facade
 * to SQuirreL and {@link DatabaseMetaData} code.
 * @author DF
 * @see DatabaseMetaData
 */
public final class DatabaseMetadataUtil {

	/**
	 * The logger for this class.
	 */
	private static DBNormalizerLogger logger = DBNormalizerLogger
			.getLogger(DatabaseMetadataUtil.class);

	/**
	 * Gets all tables from the specified catalog or schema.
	 * <p>
	 * <b>NOTE: </b>The returned table objects are not completely filled due
	 * to performance issues, but only contain the table's name, catalog, schema
	 * and TableContainer. To get all available information, use the
	 * {@link #getTableInfos(DBNormalizerSession, String, String, boolean)} method
	 * or get further information by other methods of this class (e.g.
	 * {@link #getTableColumnInfos(DBNormalizerSession, String, String, String)}.
	 * </p>
	 * @param session the DBNormalizer session which is connected to the DB
	 * @param catalogName the catalog's name, may be <code>null</code>
	 * @param schemaName the schema's name, may be <code>null</code>
	 * @return a collection of Table objects
	 * @throws SQLException
	 */
	public static Collection<Table> getTableInfos(DBNormalizerSession session, String catalogName,
			String schemaName) throws SQLException {
		return getTableInfos(session, catalogName, schemaName, false);
	}

	/**
	 * Gets all tables from the specified catalog or schema.
	 * <p>
	 * <b>NOTE: </b>If you set the argument <code>complete</code> to <code>false</code>,
	 * the returned table objects will not be completely filled,
	 * but only contain the table's name, catalog, schema
	 * and TableContainer. To get all available information,
	 * which is not high-performance,
	 * set the argument <code>complete</code> to <code>true</code>.
	 * </p>
	 * @param session the DBNormalizer session which is connected to the DB
	 * @param catalogName the catalog's name, may be <code>null</code>
	 * @param schemaName the schema's name, may be <code>null</code>
	 * @param complete <code>true</code>, to read all available information;
	 * <code>false</code>, otherwise
	 * @return a collection of Table objects
	 * @throws SQLException
	 */
	public static Collection<Table> getTableInfos(DBNormalizerSession session, String catalogName,
			String schemaName, boolean complete) throws SQLException {
		SQLDatabaseMetaData md = session.getSqlConnection().getSQLMetaData();
		ITableInfo[] tableInfos = md.getTables(catalogName, schemaName, null,
				new String[] { "TABLE" }, null);

		ArrayList<Table> tables = new ArrayList<Table>(tableInfos.length);

		for (ITableInfo tableInfo : tableInfos) {
			Table table = createTableFromTableInfo(tableInfo);
			if (complete) {
				retrieveCompleteTableInfo(session, table);
			}
			tables.add(table);
		}

		return tables;
	}

	/**
	 * Gets all tables from the specified TableContainer.
	 * <p>
	 * <b>NOTE: </b>If you set the argument <code>complete</code> to <code>false</code>,
	 * the returned table objects will not be completely filled,
	 * but only contain the table's name, catalog, schema
	 * and TableContainer. To get all available information,
	 * which is not high-performance,
	 * set the argument <code>complete</code> to <code>true</code>.
	 * </p>
	 * @param session the DBNormalizer session which is connected to the DB
	 * @param tableContainer the TableContainer
	 * @param complete <code>true</code>, to read all available information;
	 * <code>false</code>, otherwise
	 * @return a collection of Table objects
	 * @throws SQLException
	 */
	public static Collection<Table> getTableInfos(DBNormalizerSession session,
			TableContainer tableContainer, boolean complete) throws SQLException {
		String catalogName = null;
		String schemaName = null;
		switch (tableContainer.getType()) {
		case CATALOG:
			catalogName = tableContainer.getName();
			break;
		case SCHEMA:
			schemaName = tableContainer.getName();
			break;
		default:
			break;
		}

		return getTableInfos(session, catalogName, schemaName, complete);
	}

	/**
	 * Helper method which converts from SQuirreL's ITableInfo object
	 * to DBNormalizer's Table object. 
	 * @param tableInfo an ITableInfo object
	 * @return a Table object
	 */
	private static Table createTableFromTableInfo(ITableInfo tableInfo) {
		Table table = new Table();
		table.setDbTable(true);

		table.setName(tableInfo.getSimpleName());
		table.setCatalog(tableInfo.getCatalogName());
		table.setSchema(tableInfo.getSchemaName());

		TableContainer tableContainer = MiscUtil.createTableContainer(tableInfo.getCatalogName(),
				tableInfo.getSchemaName());
		table.setContainer(tableContainer);

		return table;
	}

	/**
	 * Gets all TableContainers (catalogs or schemas) for the
	 * specified DBNormalizer session.
	 * @param session the DBNormalizer-Session
	 * @return a collection of TableContainers
	 * @throws SQLException
	 */
	public static Collection<TableContainer> getTableContainers(DBNormalizerSession session)
			throws SQLException {
		Collection<String> catalogs = getCatalogs(session);
		Collection<String> schemas = getSchemas(session);
		ArrayList<TableContainer> containers = new ArrayList<TableContainer>(catalogs.size()
				+ schemas.size());
		for (String catalog : catalogs) {
			containers.add(new TableContainer(catalog, TableContainerType.CATALOG));
		}
		for (String schema : schemas) {
			containers.add(new TableContainer(schema, TableContainerType.SCHEMA));
		}
		return containers;
	}

	/**
	 * Gets the table with the specified name, catalog and schema.
	 * <p>
	 * <b>NOTE: </b>If you set the argument <code>complete</code> to <code>false</code>,
	 * the returned table object will not be completely filled,
	 * but only contain the table's name, catalog, schema
	 * and TableContainer. To get all available information,
	 * which is not high-performance,
	 * set the argument <code>complete</code> to <code>true</code>.
	 * </p>
	 * @param session the DBNormalizer session which is connected to the DB
	 * @param catalogName the catalog's name, may be <code>null</code>
	 * @param schemaName the schema's name, may be <code>null</code>
	 * @param tableName the table's (simple) name
	 * @param complete <code>true</code>, to read all available information;
	 * <code>false</code>, otherwise
	 * @return a table object, if such a table could be found in the DB;
	 * <code>null</code>, otherwise
	 * @throws SQLException
	 */
	public static Table getTableInfo(DBNormalizerSession session, String catalogName,
			String schemaName, String tableName, boolean complete) throws SQLException {
		Table table = null;

		SQLDatabaseMetaData md = session.getSqlConnection().getSQLMetaData();
		ITableInfo[] tableInfos = md.getTables(catalogName, schemaName, tableName,
				new String[] { "TABLE" }, null);

		if (tableInfos.length > 0) {

			if (tableInfos.length > 1) {
				logger.warn("More than one table found with qualified name '"
						+ MiscUtil.getQualifiedTableName(tableName, catalogName, schemaName) + "'");
			}

			ITableInfo tableInfo = tableInfos[0];
			table = createTableFromTableInfo(tableInfo);

			if (complete) {
				retrieveCompleteTableInfo(session, table);
			}

		}

		return table;
	}

	/**
	 * Helper method which fills the given table object with
	 * further information (Column, PK and FK information).
	 * @param session the DBNormalizer session which is connected to the DB
	 * @param table a table object
	 * @throws SQLException
	 */
	private static void retrieveCompleteTableInfo(DBNormalizerSession session, Table table)
			throws SQLException {
		SQLDatabaseMetaData md = session.getSqlConnection().getSQLMetaData();

		String tableName = table.getName();
		String catalogName = table.getCatalog();
		String schemaName = table.getSchema();

		// retrieve tableColumn information
		Collection<TableColumn> colInfos = DatabaseMetadataUtil.getTableColumnInfos(session,
				catalogName, schemaName, tableName);

		for (TableColumn colInfo : colInfos) {
			table.getAttributes().add(colInfo.getName());
			table.addColumnInfo(colInfo.getName(), colInfo);
		}

		PK dbPK = DatabaseMetadataUtil.getPK(session, catalogName, schemaName, tableName);
		table.setDbPK(dbPK);

		Collection<FK> importedKeys = DatabaseMetadataUtil.getImportedKeys(md.getJDBCMetaData(),
				catalogName, schemaName, tableName);
		table.getImportedKeys().addAll(importedKeys);

		Collection<FK> exportedKeys = DatabaseMetadataUtil.getExportedKeys(md.getJDBCMetaData(),
				catalogName, schemaName, tableName);
		table.getExportedKeys().addAll(exportedKeys);
	}

	/**
	 * Gets the catalogs for the specified DBNormalizer session.
	 * @param session the DBNormalizer session which is connected to the DB
	 * @return the catalogs as a collection of strings
	 * @throws SQLException
	 */
	public static Collection<String> getCatalogs(DBNormalizerSession session) throws SQLException {
		SQLDatabaseMetaData md = session.getSqlConnection().getSQLMetaData();
		String[] containerArray = md.getCatalogs();
		ArrayList<String> containerList = new ArrayList<String>(containerArray.length);
		for (String container : containerArray) {
			containerList.add(container);
		}
		return containerList;
	}

	/**
	 * Gets the schemas for the specified DBNormalizer session.
	 * @param session the DBNormalizer session which is connected to the DB
	 * @return the schemas as a collection of strings
	 * @throws SQLException
	 */
	public static Collection<String> getSchemas(DBNormalizerSession session) throws SQLException {
		SQLDatabaseMetaData md = session.getSqlConnection().getSQLMetaData();
		String[] containerArray = md.getSchemas();
		ArrayList<String> containerList = new ArrayList<String>(containerArray.length);
		for (String container : containerArray) {
			containerList.add(container);
		}
		return containerList;
	}

	/**
	 * Gets the imported keys for a table specified by catalog, schema and name.
	 * @param dbmd a SQL-DatabaseMetaData-object
	 * @param catalog the table's catalog, may be <code>null</code>
	 * @param schema the table's schema, may be <code>null</code>
	 * @param tableName the table's (simple) name
	 * @return  a collection of FKs
	 * @see DatabaseMetaData#getImportedKeys(String, String, String)
	 * @throws SQLException
	 */
	public static Collection<FK> getImportedKeys(DatabaseMetaData dbmd, String catalog,
			String schema, String tableName) throws SQLException {
		ResultSet rs = dbmd.getImportedKeys(catalog, schema, tableName);
		Collection<FK> fks = buildFKsFromResultSet(rs);
		return fks;
	}

	/**
	 * Gets the exported keys for a table specified by catalog, schema and name.
	 * @param dbmd a SQL-DatabaseMetaData-object
	 * @param catalog the table's catalog, may be <code>null</code>
	 * @param schema the table's schema, may be <code>null</code>
	 * @param tableName the table's (simple) name
	 * @return  a collection of FKs
	 * @see DatabaseMetaData#getExportedKeys(String, String, String)
	 * @throws SQLException
	 */
	public static Collection<FK> getExportedKeys(DatabaseMetaData dbmd, String catalog,
			String schema, String tableName) throws SQLException {
		ResultSet rs = dbmd.getExportedKeys(catalog, schema, tableName);
		Collection<FK> fks = buildFKsFromResultSet(rs);
		return fks;
	}

	/**
	 * Helper method creating a collection of FKs from the ResultSet returned
	 * from the {@link DatabaseMetaData#getImportedKeys(String, String, String)}
	 * or {@link DatabaseMetaData#getExportedKeys(String, String, String)} method.
	 * @param rs the ResultSet
	 * @return a collection of FKs
	 * @throws SQLException
	 */
	private static Collection<FK> buildFKsFromResultSet(ResultSet rs) throws SQLException {
		Map<String, FK> fkMap = new LinkedHashMap<String, FK>();

		while (rs.next()) {
			FK tempFK = new FK();

			tempFK.setPkCatalog(rs.getString(1));
			tempFK.setPkSchema(rs.getString(2));
			tempFK.setPkTableName(rs.getString(3));

			tempFK.setFkCatalog(rs.getString(5));
			tempFK.setFkSchema(rs.getString(6));
			tempFK.setFkTableName(rs.getString(7));
			
			// it's important to set the delete and update rule
			// only if the resultset value is not null (means DB-Default)
			// --> otherwise the rule would be set to 0,
			// because rs.getInt returns 0 in this case.
			// And 0 means "CASCADE"!
			int updateRule = rs.getInt(10);
			if (!rs.wasNull()) {
				tempFK.setUpdateRule(updateRule);
			}
			int deleteRule = rs.getInt(11);
			if (!rs.wasNull()) {
				tempFK.setDeleteRule(deleteRule);
			}
			
			tempFK.setName(rs.getString(12));
			tempFK.setPkName(rs.getString(13));
			tempFK.setDeferability(rs.getInt(14));

			String pkColumnName = rs.getString(4);
			String fkColumnName = rs.getString(8);
			int keySeq = rs.getInt(9);

			String key = createFKKey(tempFK);

			// it is important to check if there is already an fk
			// for this key in the map, because we could overwrite
			// the fkColums and pkColumns for a already existing fk otherwise
			if (!fkMap.containsKey(key)) {
				fkMap.put(key, tempFK);
			}

			FK fk = fkMap.get(key);

			List<String> pkColumnNames = fk.getPkColumnNames();
			pkColumnNames.add(keySeq - 1, pkColumnName);
			List<String> fkColumnNames = fk.getFkColumnNames();
			fkColumnNames.add(keySeq - 1, fkColumnName);

		}

		rs.close();

		return fkMap.values();
	}

	/**
	 * Helper method which creates a string which can be used
	 * as a FK's key in a map.
	 * @param fk a FK
	 * @return the string key
	 */
	private static String createFKKey(FK fk) {
		StringBuffer strBuf = new StringBuffer();
		strBuf.append(fk.getFkCatalog()).append(fk.getFkSchema()).append(fk.getFkTableName())
				.append(fk.getName()).append(fk.getPkCatalog()).append(fk.getPkSchema()).append(
						fk.getPkTableName()).append(fk.getPkName());
		return strBuf.toString();
	}

	/**
	 * Gets the PK for a table specified by catalog, schema and name.
	 * @param session the DBNormalizer session which is connected to the DB
	 * @param catalog the table's catalog, may be <code>null</code>
	 * @param schema the table's schema, may be <code>null</code>
	 * @param tableName the table's (simple) name
	 * @return the table's PK, if such a table exists and it contains
	 * a PK;<code>null</code>, otherwise
	 * @throws SQLException
	 */
	public static PK getPK(DBNormalizerSession session, String catalog, String schema,
			String tableName) throws SQLException {
		SQLDatabaseMetaData md = session.getSqlConnection().getSQLMetaData();
		PrimaryKeyInfo[] primaryKeyInfos = md.getPrimaryKey(catalog, schema, tableName);
		PK pk = null;
		if (primaryKeyInfos != null && primaryKeyInfos.length > 0) {
			pk = new PK();
			for (int i = 0; i < primaryKeyInfos.length; i++) {
				PrimaryKeyInfo pkInfo = primaryKeyInfos[i];
				// assign the name only once
				if (i == 0) {
					pk.setName(pkInfo.getSimpleName());
				}
				String pkAttrName = pkInfo.getColumnName();
				pk.getAttributes().add(pkAttrName);
			}
		}
		return pk;
	}

	/**
	 * Gets TableColumn objects for a table specified by catalog, schema and name.
	 * @param session the DBNormalizer session which is connected to the DB
	 * @param catalogName the table's catalog, may be <code>null</code>
	 * @param schemaName the table's schema, may be <code>null</code>
	 * @param tableName the table's (simple) name
	 * @return a collection of TableColumn objects
	 * @throws SQLException
	 */
	public static Collection<TableColumn> getTableColumnInfos(DBNormalizerSession session,
			String catalogName, String schemaName, String tableName) throws SQLException {
		SQLDatabaseMetaData md = session.getSqlConnection().getSQLMetaData();
		TableColumnInfo[] tableColumnInfos = md.getColumnInfo(catalogName, schemaName, tableName);

		ArrayList<TableColumn> cols = new ArrayList<TableColumn>();

		for (TableColumnInfo tableColumnInfo : tableColumnInfos) {
			TableColumn col = new TableColumn();
			col.setName(tableColumnInfo.getColumnName());
			col.setTypeName(tableColumnInfo.getTypeName());
			col.setDataType(tableColumnInfo.getDataType());
			col.setNullable("YES".equals(tableColumnInfo.isNullable()));
			col.setColumnSize(tableColumnInfo.getColumnSize());
			col.setDecimalDigits(tableColumnInfo.getDecimalDigits());
			col.setDefaultValue(tableColumnInfo.getDefaultValue());
			cols.add(col);
		}
		return cols;
	}
}
