package de.fesenmeyer.dbnormalizer.core.log;

import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;

/**
 * Application-specific logging wrapper class based on <a
 * href="http://codeelixir.com/2006/06/19/java-logging-made-easy.aspx">
 * http://codeelixir.com/2006/06/19/java-logging-made-easy.aspx</a>.
 * Currently wraps log4j-Logging.
 * 
 * @author DF
 * 
 */
public class DBNormalizerLogger {
	/**
	 * A log4j-Logger.
	 */
	private Logger logger = null;

	/**
	 * Static HashMap for all loggers.
	 */
	@SuppressWarnings("unchecked")
	private static final Map loggers = new HashMap();

	/**
	 * Private Constructor for creating a logger.
	 * @param c
	 */
	@SuppressWarnings("unchecked")
	private DBNormalizerLogger(Class c) {
		this.logger = Logger.getLogger(c);
	}

	/**
	 * Gets the logger object for the class specified by
	 * argument c. Creates the logger object, if it does not
	 * yet exist.
	 * @param c the class
	 * @return the logger object
	 */
	@SuppressWarnings("unchecked")
	public static DBNormalizerLogger getLogger(Class c) {
		String className = c.getName();
		DBNormalizerLogger logger = (DBNormalizerLogger) loggers.get(className);
		if (logger == null) {
			synchronized (DBNormalizerLogger.class) {
				logger = (DBNormalizerLogger) loggers.get(className);
				if (logger == null) {
					logger = new DBNormalizerLogger(c);
					loggers.put(className, logger);
				}
			}
		}
		return logger;
	}

	/**
	 * Returns if fatal error messages are loggable.
	 * @return <code>true</code>, if they are loggable;
	 * <code>false</code>, otherwise
	 */
	public boolean isFatalLoggable() {
		return logger.isEnabledFor(Level.FATAL);
	}

	/**
	 * Returns if error messages are loggable.
	 * @return <code>true</code>, if they are loggable;
	 * <code>false</code>, otherwise
	 */
	public boolean isErrorLoggable() {
		return logger.isEnabledFor(Level.ERROR);
	}

	/**
	 * Returns if warn messages are loggable.
	 * @return <code>true</code>, if they are loggable;
	 * <code>false</code>, otherwise
	 */
	public boolean isWarnLoggable() {
		return logger.isEnabledFor(Level.WARN);
	}

	/**
	 * Returns if info messages are loggable.
	 * @return <code>true</code>, if they are loggable;
	 * <code>false</code>, otherwise
	 */
	public boolean isInfoLoggable() {
		return logger.isEnabledFor(Level.INFO);
	}

	/**
	 * Returns if debug messages are loggable.
	 * @return <code>true</code>, if they are loggable;
	 * <code>false</code>, otherwise
	 */
	public boolean isDebugLoggable() {
		return logger.isEnabledFor(Level.DEBUG);
	}

	/**
	 * Logs a fatal error with the specified message.
	 * @param message the message
	 */
	public void fatal(String message) {
		logger.fatal(message);
	}

	/**
	 * Logs a fatal error message with the specified message and
	 * error.
	 * @param message the message
	 * @param t the error as a Throwable object
	 */
	public void fatal(String message, Throwable t) {
		logger.fatal(mergeThrowableWithMessage(message, t));
	}

	/**
	 * Logs an error with the specified message.
	 * @param message the message
	 */
	public void error(String message) {
		logger.error(message);
	}

	/**
	 * Logs an error message with the specified message and
	 * error.
	 * @param message the message
	 * @param t the error as a Throwable object
	 */
	public void error(String message, Throwable t) {
		logger.error(mergeThrowableWithMessage(message, t));
	}

	/**
	 * Logs a warning with the specified message.
	 * @param message the message
	 */
	public void warn(String message) {
		logger.warn(message);
	}

	/**
	 * Logs a warning with the specified message and
	 * error.
	 * @param message the message
	 * @param t the error as a Throwable object
	 */
	public void warn(String message, Throwable t) {
		logger.warn(mergeThrowableWithMessage(message, t));
	}

	/**
	 * Logs an info with the specified message.
	 * @param message the message
	 */
	public void info(String message) {
		logger.info(message);
	}

	/**
	 * Logs a debug message.
	 * @param message the message
	 */
	public void debug(String message) {
		logger.debug(message);
	}

	/**
	 * This method includes information about a Throwable t in the message. This method is needed
	 * because {@link PatternLayout} does not include the Throwable information and we do not want
	 * to use another more complex layout.
	 * 
	 * @param message
	 *            a message
	 * @param t
	 *            a Throwable
	 * @return Returns a string including the message and information about the Throwable.
	 */
	private String mergeThrowableWithMessage(String message, Throwable t) {
		StringBuffer strBuf = new StringBuffer();
		if (message != null) {
			strBuf.append(message);
		}
		if (message != null && t != null) {
			strBuf.append(": ");
		}
		if (t != null) {
			strBuf.append(t.getMessage());
		}
		return strBuf.toString();
	}

}
