package de.fesenmeyer.dbnormalizer.core.util;

import java.sql.DatabaseMetaData;
import java.util.Collection;
import java.util.LinkedHashSet;

import de.fesenmeyer.dbnormalizer.core.AttributeStringSet;
import de.fesenmeyer.dbnormalizer.core.DBNormalizerException;
import de.fesenmeyer.dbnormalizer.core.Table;
import de.fesenmeyer.dbnormalizer.core.TableContainer;
import de.fesenmeyer.dbnormalizer.core.TableContainerType;
import de.fesenmeyer.dbnormalizer.core.log.DBNormalizerLogger;

/**
 * "Miscellaneous" Utility class: Contains helper methods for various purposes.
 * @author DF
 *
 */
public final class MiscUtil {

	/**
	 * The logger for this class.
	 */
	private static DBNormalizerLogger logger = DBNormalizerLogger.getLogger(MiscUtil.class);

	/**
	 * Makes the constructor inaccessible.
	 */
	private MiscUtil() {
	}

	/**
	 * Parses a set of attribute names from a string. The attribute names in the string
	 * must be comma-separated. The method checks
	 * if the attributes exist in the specified table and throws an exception
	 * if they don't.
	 * @param attributeSetString the string
	 * @param table the table
	 * @return a set of attribute names
	 * @throws DBNormalizerException if some of the attribute names in the string
	 * do not exist in the specified table
	 */
	public static AttributeStringSet parseAttributeStringSet(String attributeSetString, Table table)
			throws DBNormalizerException {
		String[] attributeNames = StringUtil.parseStringArray(attributeSetString);
		AttributeStringSet attributeStrSet = new AttributeStringSet();
		LinkedHashSet<String> nonExistingAttrs = new LinkedHashSet<String>();
		for (String attributeName : attributeNames) {
			if (table.getAttributes().contains(attributeName)) {
				attributeStrSet.add(attributeName);
			} else {
				nonExistingAttrs.add(attributeName);
			}
		}
		if (nonExistingAttrs.size() > 0) {
			String nonExistingAttrsStr = MiscUtil.getCollectionAsString(nonExistingAttrs);
			String strMessage = "The following attributes do not exist: " + nonExistingAttrsStr;
			logger.error(strMessage);
			throw new DBNormalizerException(strMessage);
		}
		return attributeStrSet;
	}

	/**
	 * Checks if two tables are equal by comparing their name, catalog and schema.
	 * @param table1Name the first table's (simple) name
	 * @param table1Catalog the first table's catalog, may be <code>null</code>
	 * @param table1Schema the first table's schema, may be <code>null</code>
	 * @param table2Name the second table's (simple) name
	 * @param table2Catalog the second table's catalog, may be <code>null</code>
	 * @param table2Schema the second table's schema, may be <code>null</code>
	 * @return <code>true</code>, if the tables are equal;
	 * <code>false</code>, otherwise
	 */
	public static boolean checkTableEquality(String table1Name, String table1Catalog,
			String table1Schema, String table2Name, String table2Catalog, String table2Schema) {
		String t1qName = getQualifiedTableName(table1Name, table1Catalog, table1Schema);
		String t2qName = getQualifiedTableName(table2Name, table2Catalog, table2Schema);
		return t1qName.equals(t2qName);
	}

	/**
	 * Generates a qualified table name based on the arguments
	 * tableName, catalog and schema.
	 * @param tableName a (simple) table name
	 * @param catalog a catalog name, may be <code>null</code>
	 * @param schema a schema name, may be <code>null</code>
	 * @return the qualified name
	 */
	public static String getQualifiedTableName(String tableName, String catalog, String schema) {
		Table table = new Table();
		table.setName(tableName);
		table.setSchema(schema);
		table.setCatalog(catalog);
		return table.getQualifiedName();
	}

	/**
	 * Creates a TableContainer. At least one of the arguments must be <code>null</code>.
	 * @param catalog the catalog, may be <code>null</code>
	 * @param schema the schema, may be <code>null</code>
	 * @return Returns a TableContainer of type 
	 * <ul>
	 * 	<li>{@link TableContainerType#NONE}, if both arguments are <code>null</code>,</li>
	 * 	<li>{@link TableContainerType#CATALOG}, if catalog is not <code>null</code> or</li>
	 * 	<li>{@link TableContainerType#SCHEMA}, if schema is not  <code>null</code>.</li>
	 * </ul>
	 * @see TableContainer
	 * @see TableContainerType
	 */
	public static TableContainer createTableContainer(String catalog, String schema) {
		TableContainer tableContainer = TableContainer.NO_CONTAINER;
		
		if (catalog != null) {
			tableContainer = new TableContainer(catalog, TableContainerType.CATALOG);
		}
		if (schema != null) {
			tableContainer = new TableContainer(schema, TableContainerType.SCHEMA);
		}

		return tableContainer;
	}

	/**
	 * Gets a Collection of objects as a StringBuffer object. Builds a comma-separated string
	 * using the objects' toString-method.
	 * @param collection a collection of objects
	 * @return a StringBuffer object representing the collection
	 */
	public static StringBuffer getCollectionAsStringBuffer(Collection<? extends Object> collection) {
		StringBuffer strBuf = new StringBuffer();
		int i = 0;
		for (Object o : collection) {
			if (i != 0) {
				strBuf.append(", ");
			}
			strBuf.append(o.toString());
			i++;
		}
		return strBuf;
	}

	/**
	 * Gets a Collection of objects as a string. Builds a comma-separated string
	 * using the objects' toString-method.
	 * @param collection a collection of objects
	 * @return a string representing the collection
	 */
	public static String getCollectionAsString(Collection<? extends Object> collection) {
		StringBuffer strBuf = getCollectionAsStringBuffer(collection);
		return strBuf.toString();
	}
	
	/**
	 * Returns a string representation of the delete or update rule
	 * specified by argument rule. Examples are "CASCADE" or "RESTRICT".
	 * @param rule the {@link DatabaseMetaData}-Constant of the rule
	 * @return a string representation of the rule
	 */
	public static String getDeleteOrUpdateRuleAsString(int rule) {
		String str = "?";
		switch (rule) {
		case DatabaseMetaData.importedKeyCascade:
			str = "CASCADE";
			break;
		case DatabaseMetaData.importedKeySetNull:
			str = "SET NULL";
			break;
		case DatabaseMetaData.importedKeySetDefault:
			str = "SET DEFAULT";
			break;
		case DatabaseMetaData.importedKeyRestrict:
			str = "RESTRICT";
			break;
		case DatabaseMetaData.importedKeyNoAction:
			str = "NO ACTION";
			break;
		default:
			break;
		}
		return str;
	}
}
