package de.fesenmeyer.dbnormalizer.gui;

import java.awt.BorderLayout;
import java.awt.CardLayout;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Image;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;

import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSplitPane;
import javax.swing.JTextArea;
import javax.swing.SwingUtilities;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.TreePath;

import de.fesenmeyer.dbnormalizer.core.DBConfig;
import de.fesenmeyer.dbnormalizer.core.DBNormalizerException;
import de.fesenmeyer.dbnormalizer.core.DBNormalizerSession;
import de.fesenmeyer.dbnormalizer.core.Table;
import de.fesenmeyer.dbnormalizer.core.configuration.ConfigurationException;
import de.fesenmeyer.dbnormalizer.core.configuration.ConfigurationReader;
import de.fesenmeyer.dbnormalizer.core.configuration.ConfigurationWriter;
import de.fesenmeyer.dbnormalizer.core.configuration.SessionConfig;
import de.fesenmeyer.dbnormalizer.core.dbaccess.DBException;
import de.fesenmeyer.dbnormalizer.core.log.DBNormalizerLogger;
import de.fesenmeyer.dbnormalizer.core.log.TextAreaAppender;
import de.fesenmeyer.dbnormalizer.gui.util.FileFilters;

/**
 * @author DF
 *
 */
@SuppressWarnings("serial")
public class ApplicationFrame extends JFrame implements ActionListener {

	private static DBNormalizerLogger logger = DBNormalizerLogger.getLogger(ApplicationFrame.class);

	public static final String TABLE_CARD_NAME = "tableCard";

	public static final String NORM_PROP_CARD_NAME = "normPropCard";

	public static final String NORM_PROP_TABLE_CARD_NAME = "normPropTableCard";

	public static final String OTHER_CARD_NAME = "otherCard";

	private DBNormalizerSession session = null;

	private InformationPanel infoPanel = null;

	private JFileChooser fileChooser = null;

	private Table currentTable = null;

	private TableTreePanel tableTreePanel = null;

	private JPanel rightPanel = null;

	private JButton newDBSessionButton = null;

	private JButton openButton = null;
	
	private JButton openTaneButton = null;

	private JButton saveButton = null;

	private DBConfigFrame newDbConfigFrame = null;

	private DBConfigFrame showDbConfigFrame = null;

	private TableMainPanel tableMainPanel = null;

	private NormPropPanel normPropPanel = null;

	private NormPropTablePanel normPropTablePanel = null;

	private JSplitPane verticalSplit;

	private JSplitPane horizontalSplit;

	private JButton showDBConfigButton;

	private JButton newNoDBSessionButton;

	public ApplicationFrame() {
		super("DB Normalizer");
		setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
		// ask the user for confirmation before closing the application
		this.addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent e) {
				String strMessage = "Do you really want to close the application?";
				int option = JOptionPane.showOptionDialog(ApplicationFrame.this, strMessage,
						"Closing DBNormalizer", JOptionPane.YES_NO_OPTION,
						JOptionPane.WARNING_MESSAGE, null, new String[] { "Yes", "No" }, "No");
				if (option == JOptionPane.YES_OPTION) {
					System.exit(0);
				} else {
					// do nothing
				}
			}
		});

		// set an icon for the application
		Image iconImage = IconManager.getInstance().getImage(IconManager.DBNORMALIZER_ICON);
		setIconImage(iconImage);

		this.fileChooser = new JFileChooser();
		this.fileChooser.setFileFilter(FileFilters.getXMLFileFilter());

		this.setLayout(new BorderLayout());

		JPanel buttonPanel = createButtonPanel();
		this.add(buttonPanel, BorderLayout.PAGE_START);

		this.tableTreePanel = new TableTreePanel();

		this.verticalSplit = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
		this.verticalSplit.setOneTouchExpandable(true);
		this.add(this.verticalSplit, BorderLayout.CENTER);

		this.infoPanel = new InformationPanel();
		this.verticalSplit.setBottomComponent(infoPanel);

		this.horizontalSplit = new JSplitPane();
		this.horizontalSplit.setOneTouchExpandable(true);
		this.horizontalSplit.setLeftComponent(tableTreePanel);

		this.verticalSplit.setTopComponent(this.horizontalSplit);

		CardLayout cl = new CardLayout();
		this.rightPanel = new JPanel(cl);

		this.tableMainPanel = new TableMainPanel();
		this.tableMainPanel.setApplicationFrame(this);
		this.rightPanel.add(this.tableMainPanel, TABLE_CARD_NAME);

		this.normPropPanel = new NormPropPanel();
		this.rightPanel.add(this.normPropPanel, NORM_PROP_CARD_NAME);

		this.normPropTablePanel = new NormPropTablePanel();
		this.rightPanel.add(this.normPropTablePanel, NORM_PROP_TABLE_CARD_NAME);

		// add an empty dummy panel which will be shown if there
		// is no information to be displayed for the current selection
		this.rightPanel.add(new JPanel(), OTHER_CARD_NAME);

		cl.show(this.rightPanel, OTHER_CARD_NAME);

		this.horizontalSplit.setRightComponent(rightPanel);

		tableTreePanel.getTableTree().addTreeSelectionListener(new TreeSelectionListener() {

			public void valueChanged(TreeSelectionEvent e) {
				TreePath path = e.getNewLeadSelectionPath();
				CardLayout cl = (CardLayout) rightPanel.getLayout();

				String cardToShow = null;
				if (path == null) {
					cardToShow = OTHER_CARD_NAME;
				} else {
					DBNormalizerTreeNode node = ((DBNormalizerTreeNode) path.getLastPathComponent());
					DBNormalizerTreeNode.NodeType nodeType = node.getNodeType();
					Object o = node.getUserObject();
					switch (nodeType) {
					case TABLE:
						try {
							setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
							
							currentTable = (Table) o;
	
							tableMainPanel.setTable(currentTable);
	
							cardToShow = TABLE_CARD_NAME;
						} catch (Exception ex){
							logger.error("An unexpected exception occured", ex);
						} finally {
							setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
						}
						break;
					case NORM_PROPOSAL:
						DBNormalizerTreeNode orgTableNode = (DBNormalizerTreeNode) node.getParent();
						Table orgTable = (Table) orgTableNode.getUserObject();

						normPropPanel.setNormProposal(orgTable.getNormProposal());

						cardToShow = NORM_PROP_CARD_NAME;
						break;
					case NORM_PROPOSAL_TABLE:
						Table normPropTable = (Table) o;
						normPropTablePanel.setTable(normPropTable);

						cardToShow = NORM_PROP_TABLE_CARD_NAME;
						break;
					default:
						cardToShow = OTHER_CARD_NAME;
						break;
					}
				}

				cl.show(rightPanel, cardToShow);
			}

		});

	}

	private JPanel createButtonPanel() {
		JPanel buttonPanel = new JPanel(new FlowLayout());

		this.newDBSessionButton = new JButton("New DB Session...");
		this.newDBSessionButton.addActionListener(this);
		buttonPanel.add(newDBSessionButton);

		this.newNoDBSessionButton = new JButton("New No-DB Session...");
		this.newNoDBSessionButton.addActionListener(this);
		buttonPanel.add(this.newNoDBSessionButton);

		this.openButton = new JButton("Open...");
		openButton.addActionListener(this);
		buttonPanel.add(openButton);
		
		this.openTaneButton = new JButton("Open TANE output...");
		openTaneButton.addActionListener(this);
		buttonPanel.add(openTaneButton);

		this.saveButton = new JButton("Save...");
		this.saveButton.setEnabled(false);
		saveButton.addActionListener(this);
		buttonPanel.add(saveButton);

		this.showDBConfigButton = new JButton("Show DB Config...");
		this.showDBConfigButton.setEnabled(false);
		showDBConfigButton.addActionListener(this);
		buttonPanel.add(this.showDBConfigButton);

		return buttonPanel;
	}

	private DBNormalizerSession createSession(String fileName, boolean isTaneOutput) throws ConfigurationException {
		DBNormalizerSession newSession = null;

		ConfigurationReader configReader = new ConfigurationReader(fileName);
		SessionConfig sessionConfig = null;
		
		try {
			if (isTaneOutput) {
				sessionConfig = configReader.readTaneJavaConfig();
			} else {
				sessionConfig = configReader.read();
			}
		} catch (Exception e) {
			throw new ConfigurationException(e);
		}

		DBConfig dbConfig = sessionConfig.getDbConfig();
		if (dbConfig == null) {
			newSession = new DBNormalizerSession();
		} else {
			try {
				newSession = new DBNormalizerSession(dbConfig);
			} catch (DBException ex) {
				logger.error("Exception loading DB Configuration", ex);
				// if an error occurs, open DBConfigFrame to change DBConfig
				getNewDBConfigFrame().setDbConfig(dbConfig);
				newSession = tryCreateDBSession();
			}
		}

		// load the table information (currently only FDs) contained in the file
		if (newSession != null) {
			newSession.addTableInformation(sessionConfig.getTableContainerConfigs());
		}

		return newSession;
	}

	private static void createAndShowGUI() {
		ApplicationFrame frame = new ApplicationFrame();
		frame.pack();
		Dimension screen = frame.getToolkit().getScreenSize();
		frame.setSize(screen.width - 50, screen.height - 75);
		frame.setLocation(25, 25);
		// maximize the window
		frame.setVisible(true);

		frame.setDefaultDividerLocations();

		setupLog4JAppender(frame.getInfoPanel().getTextArea());
	}

	private static void setupLog4JAppender(JTextArea jTextArea) {
		// This code attaches the appender to the text area
		TextAreaAppender.setTextArea(jTextArea);
		// (log4j configuration is contained in the file log4j.properties
		// in the source folder)
	}

	public static void main(String[] args) {
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				createAndShowGUI();
			}
		});
	}

	public void actionPerformed(ActionEvent e) {
		if (e.getSource() == this.openButton) {
			int retVal = this.fileChooser.showOpenDialog(this);
			if (retVal == JFileChooser.APPROVE_OPTION) {
				String fileName = this.fileChooser.getSelectedFile().getAbsolutePath();
				DBNormalizerSession newSession = null;
				this.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
				try {
					newSession = createSession(fileName, false);
					if (newSession != null) {
						changeSession(newSession);
					}
				} catch (DBNormalizerException ex) {
					logger.error("Error loading configuration file '" + fileName + "'", ex);
				} finally {
					this.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
				}
			}
		} else if (e.getSource() == this.openTaneButton) {
			int retVal = this.fileChooser.showOpenDialog(this);
			if (retVal == JFileChooser.APPROVE_OPTION) {
				String fileName = this.fileChooser.getSelectedFile().getAbsolutePath();
				DBNormalizerSession newSession = null;
				this.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
				try {
					newSession = createSession(fileName, true);
					if (newSession != null) {
						changeSession(newSession);
					}
				} catch (DBNormalizerException ex) {
					logger.error("Error loading TANE output file '" + fileName + "'", ex);
				} finally {
					this.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
				}
			}
		} else if (e.getSource() == this.newNoDBSessionButton) {
			DBNormalizerSession session = new DBNormalizerSession();
			changeSession(session);
		} else if (e.getSource() == this.newDBSessionButton) {
			DBNormalizerSession newSession = tryCreateDBSession();
			if (newSession != null) {
				changeSession(newSession);
			}
		} else if (e.getSource() == this.saveButton) {
			int retVal = this.fileChooser.showSaveDialog(this);
			if (retVal == JFileChooser.APPROVE_OPTION) {				
				this.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
				try {
					String fileName = this.fileChooser.getSelectedFile().getAbsolutePath();
					//assure that the file is saved with the extension ".xml", if
					//the XMLFileFilter is selected
					if (this.fileChooser.getFileFilter().getDescription().equals(
							FileFilters.getXMLFileFilter().getDescription())) {
						String[] strSplit = fileName.split("[.]");
						String suffix = strSplit[strSplit.length - 1];
						if (!suffix.equalsIgnoreCase("xml")) {
							fileName += ".xml";
							this.fileChooser.setSelectedFile(new File(fileName));
						}
					}
					SessionConfig sessionConfig = this.session.createSessionConfig(fileName);
					ConfigurationWriter configWriter = new ConfigurationWriter(sessionConfig);
					configWriter.write();
				} catch (ConfigurationException ex) {
					logger.error("Exception while saving the session", ex);
				} finally {
					this.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
				}
			}

		} else if (e.getSource() == this.showDBConfigButton) {
			getShowDBConfigFrame().setVisible(true);
		}

	}

	private DBNormalizerSession tryCreateDBSession() {
		DBNormalizerSession newSession = null;

		//get the currently set cursor
		Cursor currentCursor = this.getCursor();
		
		boolean showFrame = true;
		while (showFrame) {
			this.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
			getNewDBConfigFrame().setVisible(true);
			if (getNewDBConfigFrame().isAccepted()) {
				try {
					this.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
					newSession = new DBNormalizerSession(getNewDBConfigFrame().getDbConfig());
					showFrame = false;
				} catch (DBException ex) {
					logger.error("Exception loading DB Configuration", ex);
				}
			} else {
				showFrame = false;
			}
		}

		this.setCursor(currentCursor);
		
		return newSession;
	}

	private void changeSession(DBNormalizerSession session) {
		//close the old session, if it is not null
		if (this.session != null) {
			try {
				this.session.closeSession();
			} catch (DBException e) {
				logger.warn("Exception while closing the old session", e);
			}
		}
		
		this.session = session;
		DBConfig dbConfig = this.session.getDbConfig();
		if (dbConfig == null) {
			this.showDBConfigButton.setEnabled(false);
		} else {
			this.showDBConfigButton.setEnabled(true);
			getShowDBConfigFrame().setDbConfig(dbConfig);
			getNewDBConfigFrame().setDbConfig(dbConfig);
		}
		this.saveButton.setEnabled(true);
		this.tableTreePanel.changeSession(session);
		this.tableMainPanel.setSession(session);
	}

	public InformationPanel getInfoPanel() {
		return infoPanel;
	}

	public TableTree getTableTree() {
		return tableTreePanel.getTableTree();
	}

	public void setDefaultDividerLocations() {
		this.horizontalSplit.setDividerLocation(0.3);
		this.verticalSplit.setDividerLocation(0.8);
	}

	private DBConfigFrame getShowDBConfigFrame() {
		if (this.showDbConfigFrame == null) {
			this.showDbConfigFrame = new DBConfigFrame(this, false);
		}
		return this.showDbConfigFrame;
	}

	private DBConfigFrame getNewDBConfigFrame() {
		if (this.newDbConfigFrame == null) {
			this.newDbConfigFrame = new DBConfigFrame(this, true);
		}
		return this.newDbConfigFrame;
	}

}
