package de.fesenmeyer.dbnormalizer.gui;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import javax.swing.JTree;
import javax.swing.event.TreeExpansionEvent;
import javax.swing.event.TreeExpansionListener;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;

import de.fesenmeyer.dbnormalizer.core.DBNormalizerSession;
import de.fesenmeyer.dbnormalizer.core.Table;
import de.fesenmeyer.dbnormalizer.core.TableColumn;
import de.fesenmeyer.dbnormalizer.core.TableContainer;
import de.fesenmeyer.dbnormalizer.core.dbaccess.DatabaseMetadataUtil;
import de.fesenmeyer.dbnormalizer.core.log.DBNormalizerLogger;

/**
 * @author DF
 *
 */
@SuppressWarnings("serial")
public class DBTree extends JTree {

	private static DBNormalizerLogger logger = DBNormalizerLogger.getLogger(DBTree.class);

	private DBNormalizerSession session = null;

	/**
	 * holds the tableContainers for which its table's attributes have already been loaded (needed
	 * for lazy loading)
	 */
	private Set<TableContainer> tableContainersWithAttrsLoaded = null;

	public DBTree() {
		super();

		DBNormalizerTreeNode top = new DBNormalizerTreeNode(
				DBNormalizerTreeNode.NodeType.ROOT_NODE, "Underlying DB");

		DefaultTreeModel model = (DefaultTreeModel) this.getModel();
		model.setRoot(top);

		// layout initialization
		this.setCellRenderer(new DBTreeCellRenderer());

		this.addTreeExpansionListener(new TreeExpansionListener() {

			public void treeCollapsed(TreeExpansionEvent arg0) {
				// do nothing
			}

			public void treeExpanded(TreeExpansionEvent evt) {
				TreePath path = evt.getPath();
				DBNormalizerTreeNode node = (DBNormalizerTreeNode) path.getLastPathComponent();
				if (DBNormalizerTreeNode.NodeType.TABLE_CONTAINER.equals(node.getNodeType())) {
					TableContainer tableContainer = (TableContainer) node.getUserObject();
					if (!tableContainersWithAttrsLoaded.contains(tableContainer)) {
						// lazy loading of the attributes of the tables contained in this container
						lazyLoadTableColumns(node);
						tableContainersWithAttrsLoaded.add(tableContainer);
					}
				}
			}

		});

	}

	private void lazyLoadTableColumns(DBNormalizerTreeNode tableContainerNode) {
		for (int i = 0; i < tableContainerNode.getChildCount(); i++) {
			DBNormalizerTreeNode tableNode = (DBNormalizerTreeNode) tableContainerNode
					.getChildAt(i);
			Table table = (Table) tableNode.getUserObject();
			try {
				Collection<TableColumn> tableColumns = DatabaseMetadataUtil.getTableColumnInfos(
						session, table.getCatalog(), table.getSchema(), table.getName());
				ArrayList<String> attrNames = new ArrayList<String>(tableColumns.size());
				for (TableColumn attr : tableColumns) {
					String attrName = attr.getName();
					attrNames.add(attrName);
				}
				createAttributeNodes(tableNode, attrNames);
				// inform the tree model about the change of the table node
				DefaultTreeModel model = (DefaultTreeModel) this.getModel();
				model.nodeStructureChanged(tableNode);
			} catch (SQLException e) {
				logger.warn("Attributes of table '" + table.getQualifiedName() + "' could not "
						+ "be loaded from DB", e);
			}
		}
	}

	private void createTopLevelNodes(DBNormalizerTreeNode root) throws SQLException {
		Collection<TableContainer> tableContainers = DatabaseMetadataUtil
				.getTableContainers(this.session);

		for (TableContainer tableContainer : tableContainers) {
			DBNormalizerTreeNode topLevelNode = new DBNormalizerTreeNode(
					DBNormalizerTreeNode.NodeType.TABLE_CONTAINER, tableContainer);
			root.add(topLevelNode);
			createTableNodes(topLevelNode, tableContainer);
		}
	}

	private void createTableNodes(DBNormalizerTreeNode topLevel, TableContainer tableContainer)
			throws SQLException {
		Collection<Table> tables = DatabaseMetadataUtil.getTableInfos(this.session, tableContainer,
				false);

		for (Table table : tables) {
			DBNormalizerTreeNode tableNode = new DBNormalizerTreeNode(
					DBNormalizerTreeNode.NodeType.TABLE, table);
			// we have not loaded the attributes yet, therefore we cannot
			// create attribute nodes here (we load it when table container
			// gets expanded)
			topLevel.add(tableNode);
		}
	}

	private void createAttributeNodes(DBNormalizerTreeNode tableNode, Collection<String> attributes) {
		for (String attr : attributes) {
			tableNode.add(new DBNormalizerTreeNode(DBNormalizerTreeNode.NodeType.ATTRIBUTE, attr));
		}
	}

	public void setSession(DBNormalizerSession session) {
		this.session = session;
		this.tableContainersWithAttrsLoaded = new HashSet<TableContainer>();

		DBNormalizerTreeNode top = (DBNormalizerTreeNode) this.getModel().getRoot();
		top.removeAllChildren();

		try {
			createTopLevelNodes(top);
		} catch (SQLException e) {
			logger.error("Exception loading database information", e);
		}
		((DefaultTreeModel) this.getModel()).nodeStructureChanged(top);
	}
}
