package de.fesenmeyer.dbnormalizer.gui;

import java.util.Collection;

import javax.swing.JTree;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;

import de.fesenmeyer.dbnormalizer.core.AttributeStringSet;
import de.fesenmeyer.dbnormalizer.core.DBNormalizerSession;
import de.fesenmeyer.dbnormalizer.core.Table;
import de.fesenmeyer.dbnormalizer.core.TableContainer;
import de.fesenmeyer.dbnormalizer.core.TableContainerType;
import de.fesenmeyer.dbnormalizer.core.algorithms.NormalizationProposal;
import de.fesenmeyer.dbnormalizer.core.log.DBNormalizerLogger;

/**
 * @author DF
 *
 */
@SuppressWarnings("serial")
public class TableTree extends JTree {

	private static DBNormalizerLogger logger = DBNormalizerLogger.getLogger(TableTree.class);

	private DBNormalizerSession session = null;

	private DBNormalizerTreeNode top = new DBNormalizerTreeNode(
			DBNormalizerTreeNode.NodeType.ROOT_NODE, "Tables to observe");

	public TableTree() {
		super();

		DefaultTreeModel model = (DefaultTreeModel) this.getModel();
		model.setRoot(top);

		// layout initialization
		this.setCellRenderer(new DBTreeCellRenderer());

	}

	private void createTopLevelNodes(DBNormalizerTreeNode root) {
		Collection<TableContainer> tableContainers = session.getTableContainers2Normalize();

		for (TableContainer tableContainer : tableContainers) {
			DBNormalizerTreeNode topLevelNode = new DBNormalizerTreeNode(
					DBNormalizerTreeNode.NodeType.TABLE_CONTAINER, tableContainer);
			root.add(topLevelNode);
			createTableNodes(topLevelNode, tableContainer);
		}
	}

	private void createTableNodes(DBNormalizerTreeNode topLevel, TableContainer tableContainer) {
		Collection<Table> tables = session.getTables2Normalize(tableContainer);

		for (Table table : tables) {
			DBNormalizerTreeNode tableNode = new DBNormalizerTreeNode(
					DBNormalizerTreeNode.NodeType.TABLE, table);
			createAttributeNodes(tableNode, table);
			// if a normalizationProposal exists, create a node for it
			NormalizationProposal normProposal = table.getNormProposal();
			if (normProposal != null) {
				createNormProposalNode(tableNode, normProposal);
			}
			topLevel.add(tableNode);
		}

	}

	private void createAttributeNodes(DBNormalizerTreeNode tableNode, Table table) {
		AttributeStringSet attributes = table.getAttributes();
		for (String attr : attributes) {
			tableNode.add(new DBNormalizerTreeNode(DBNormalizerTreeNode.NodeType.ATTRIBUTE, attr));
		}
	}

	private DBNormalizerTreeNode createNormProposalNode(DBNormalizerTreeNode tableNode,
			NormalizationProposal normProposal) {
		DBNormalizerTreeNode normProposalNode = new DBNormalizerTreeNode(
				DBNormalizerTreeNode.NodeType.NORM_PROPOSAL, "Normalization Proposal");
		createNormProposalTableNodes(normProposalNode, normProposal);
		tableNode.add(normProposalNode);
		return normProposalNode;
	}

	private void createNormProposalTableNodes(DBNormalizerTreeNode normProposalNode,
			NormalizationProposal normProposal) {
		Collection<Table> tables = normProposal.getCreatedTables();
		for (Table table : tables) {
			DBNormalizerTreeNode normProposalTableNode = new DBNormalizerTreeNode(
					DBNormalizerTreeNode.NodeType.NORM_PROPOSAL_TABLE, table);
			createNormProposalAttributeNodes(normProposalTableNode, table);
			normProposalNode.add(normProposalTableNode);
		}
	}

	private void createNormProposalAttributeNodes(DBNormalizerTreeNode normProposalTableNode,
			Table table) {
		AttributeStringSet attributes = table.getAttributes();
		for (String attr : attributes) {
			normProposalTableNode.add(new DBNormalizerTreeNode(
					DBNormalizerTreeNode.NodeType.NORM_PROPOSAL_ATTRIBUTE, attr));
		}
	}

	public void setSession(DBNormalizerSession session) {
		this.session = session;

		top.removeAllChildren();

		createTopLevelNodes(top);

		((DefaultTreeModel) this.getModel()).nodeStructureChanged(top);
	}

	public void addContainer(TableContainer tableContainer) {
		boolean added = this.session.addTableContainer2Normalize(tableContainer);

		if (added) {
			DBNormalizerTreeNode tableContainerNode = new DBNormalizerTreeNode(
					DBNormalizerTreeNode.NodeType.TABLE_CONTAINER, tableContainer);
			top.add(tableContainerNode);
			((DefaultTreeModel) getModel()).nodesWereInserted(top,
					new int[] { top.getChildCount() - 1 });
		} else {
			logger.warn("Table container'" + tableContainer.getName()
					+ "' could not be added, because it already exists");
		}

	}

	public void addTable(DBNormalizerTreeNode containerNode, String tableName) {
		TableContainer tableContainer = (TableContainer) containerNode.getUserObject();
		TableContainerType type = tableContainer.getType();
		String tableContainerName = tableContainer.getName();
		Table table = new Table();
		table.setContainer(tableContainer);
		table.setName(tableName);
		if (type == TableContainerType.SCHEMA) {
			table.setSchema(tableContainerName);
		} else if (type == TableContainerType.CATALOG) {
			table.setCatalog(tableContainerName);
		}

		boolean added = this.session.addTable2Normalize(tableContainer, table);

		if (added) {
			DBNormalizerTreeNode tableNode = new DBNormalizerTreeNode(
					DBNormalizerTreeNode.NodeType.TABLE, table);
			containerNode.add(tableNode);
			((DefaultTreeModel) getModel()).nodesWereInserted(containerNode,
					new int[] { containerNode.getChildCount() - 1 });
		} else {
			logger.warn("Table '" + table.getName()
					+ "' could not be added, because it already exists");
		}

	}

	public void addAttributes(DBNormalizerTreeNode currentNode, String[] attrNames) {
		Table table = null;

		DBNormalizerTreeNode.NodeType nodeType = currentNode.getNodeType();
		Object o = currentNode.getUserObject();
		DBNormalizerTreeNode tableNode = null;

		switch (nodeType) {
		case TABLE:
			tableNode = currentNode;
			table = (Table) o;
			break;
		case ATTRIBUTE:
			tableNode = (DBNormalizerTreeNode) currentNode.getParent();
			table = (Table) tableNode.getUserObject();
			break;
		}

		for (String attrName : attrNames) {
			boolean added = table.getAttributes().add(attrName);

			if (added) {
				DBNormalizerTreeNode attrNode = new DBNormalizerTreeNode(
						DBNormalizerTreeNode.NodeType.ATTRIBUTE, attrName);
				tableNode.add(attrNode);
				((DefaultTreeModel) getModel()).nodesWereInserted(tableNode, new int[] { tableNode
						.getChildCount() - 1 });
			} else {
				logger.warn("Attribute '" + attrName
						+ "' could not be added, because it already exists");
			}
		}

	}

	/**
	 * Removes a node not only from the tree, but also from the session. Currently it is only
	 * possible to remove one node at once, which is specified by the path argument
	 * 
	 * @param path
	 */
	public void remove(TreePath path) {
		DBNormalizerTreeNode node = (DBNormalizerTreeNode) path.getLastPathComponent();
		DBNormalizerTreeNode.NodeType nodeType = node.getNodeType();
		Object o = node.getUserObject();
		Table table = null;
		TableContainer container = null;

		boolean remove = false;

		switch (nodeType) {
		case TABLE_CONTAINER:
			container = (TableContainer) o;
			this.session.removeTableContainer2Normalize(container);
			remove = true;
			break;
		case TABLE:
			DBNormalizerTreeNode containerNode = (DBNormalizerTreeNode) node.getParent();
			container = (TableContainer) containerNode.getUserObject();

			table = (Table) o;

			this.session.getTables2Normalize(container).remove(table);
			remove = true;
			break;
		case NORM_PROPOSAL:
			// parent node contains the table to which the normalization
			// proposal belongs
			DBNormalizerTreeNode normTableNode = (DBNormalizerTreeNode) node.getParent();
			table = (Table) normTableNode.getUserObject();
			table.setNormProposal(null);

			remove = true;
			break;
		case ATTRIBUTE:
			DBNormalizerTreeNode tableNode = (DBNormalizerTreeNode) node.getParent();
			table = (Table) tableNode.getUserObject();
			if (table.isDbTable()) {
				remove = false;
			} else {
				String attrName = (String) o;
				table.getAttributes().remove(attrName);
				if (table.getFDs().size() > 0) {
					logger.warn("You have specified FDs for Table '" + table.getQualifiedName()
							+ "'." + " Please remove all occurencies of the removed attribute '"
							+ attrName + "' from these FDs.");
				}
				remove = true;
			}
			break;
		}

		if (remove == true) {
			((DefaultTreeModel) getModel()).removeNodeFromParent(node);
		}
	}

	private DBNormalizerTreeNode findTableContainerNode(TableContainer container) {
		DBNormalizerTreeNode foundNode = null;

		DBNormalizerTreeNode root = (DBNormalizerTreeNode) this.getModel().getRoot();
		for (int i = 0; i < root.getChildCount(); i++) {
			DBNormalizerTreeNode containerNode = (DBNormalizerTreeNode) root.getChildAt(i);
			TableContainer currTableContainer = (TableContainer) containerNode.getUserObject();
			if (currTableContainer.equals(container)) {
				foundNode = containerNode;
				break;
			}
		}

		return foundNode;
	}

	private DBNormalizerTreeNode findTableNode(Table table) {
		DBNormalizerTreeNode foundNode = null;

		DBNormalizerTreeNode containerNode = findTableContainerNode(table.getContainer());
		if (containerNode != null) {
			for (int i = 0; i < containerNode.getChildCount(); i++) {
				DBNormalizerTreeNode tableNode = (DBNormalizerTreeNode) containerNode.getChildAt(i);
				Table currTable = (Table) tableNode.getUserObject();
				if (table.equals(currTable)) {
					foundNode = tableNode;
					break;
				}
			}
		}

		return foundNode;
	}

	public void addNormalizationProposal(NormalizationProposal normProposal) {
		Table table = normProposal.getOldTable();
		DBNormalizerTreeNode tableNode = findTableNode(table);
		DBNormalizerTreeNode normProposalNode = null;

		// look if the normalization proposal node for the table already exists
		if (tableNode.getChildCount() > 0) {
			DBNormalizerTreeNode lastTableChildNode = (DBNormalizerTreeNode) tableNode
					.getLastChild();
			if (DBNormalizerTreeNode.NodeType.NORM_PROPOSAL
					.equals(lastTableChildNode.getNodeType())) {
				normProposalNode = lastTableChildNode;
			}
		}

		// if the normalization proposal node does not already exist
		if (normProposalNode == null) {
			// create a new normalization proposal node
			normProposalNode = createNormProposalNode(tableNode, normProposal);

			// inform the model about the changes
			DefaultTreeModel model = (DefaultTreeModel) this.getModel();
			model.nodesWereInserted(tableNode, new int[] { tableNode.getChildCount() - 1 });
		} else { // if the normalization proposal node already exists
			// remove the table nodes (children) of the old normalization proposal
			normProposalNode.removeAllChildren();

			// create new table nodes for the new normalization proposal
			createNormProposalTableNodes(normProposalNode, normProposal);

			// inform the model about the changes
			DefaultTreeModel model = (DefaultTreeModel) this.getModel();
			model.nodeStructureChanged(normProposalNode);
		}

		// select the normalization proposal node and expand all its child nodes
		TreePath normNodePath = new TreePath(normProposalNode.getPath());
		this.getSelectionModel().setSelectionPath(normNodePath);
		this.expandRecursively(normProposalNode);
		this.scrollPathToVisible(normNodePath);
	}

	public void expandRecursively(DefaultMutableTreeNode node) {
		TreePath treePath = new TreePath(node.getPath());
		this.expandPath(treePath);
		for (int i = 0; i < node.getChildCount(); i++) {
			DefaultMutableTreeNode childNode = (DefaultMutableTreeNode) node.getChildAt(i);
			this.expandRecursively(childNode);
		}
	}

}
